/* HickUP 1.2 -- Graphical interface to hick.
** Copyright (C) 1998 - 2001, 2008 Daniel Bergstrom <daniel@octocode.com>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2, or (at your option)
** any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <gtk/gtkvbox.h>
#include <gtk/gtkdialog.h>
#include <gtk/gtknotebook.h>

#include "gui.h"
#include "pref.h"
#include "report.h"
#include "gpixmap.h"
#include "availrep.h"

#define TAB_EXPIRED  0
#define TAB_REMOVED  1
#define TAB_EXPIRE   2
#define TAB_MODIFIED 3
#define TAB_CHANGED  4
#define TAB_CONFLICT 5

static int expire_check(Project *prj, time_t expire, time_t curtime, 
                        const char *name, const char *version, 
                        const char *type);

static void modified_check(Project *prj, time_t modified, userEntryList *ulist,
                           const char *name, const char *version, 
                           const char *type);

static int default_check(Project *prj, Packet *packet, 
                         userEntryList *installed, SpecVar *svar, 
                         int have_svar);

static void availrep_new(Project *prj);
static void cb_select_row(GtkWidget *widget, gint row, gint column, 
                          GdkEventButton *event, gpointer data);
static void cb_select_changed(GtkWidget *widget, gint row, gint column,
                              GdkEventButton *event, gpointer data);


void
availrep_clear(Project *prj)
{
  if(prj->report.window) {
    report_clear((Report *)prj->report.arg);
  }
}


void
availrep_show_report(Project *prj)
{
  int page;
  Report *r;

  if(prj->report.window) {
    r = ((Report *)prj->report.arg);

    report_show_pixmaps(r);
    page = report_display(r, ((_pref.report_expired  << TAB_EXPIRED) |
                              (_pref.report_expired  << TAB_REMOVED) |
                              (_pref.report_expire   << TAB_EXPIRE) |
                              (_pref.report_modified << TAB_MODIFIED) |
                              (_pref.report_modified << TAB_CHANGED) |
                              (_pref.report_conflict << TAB_CONFLICT)));

    if(page != -1) {
      gtk_notebook_set_page(GTK_NOTEBOOK(r->notebook), page);
      gtk_widget_show(prj->report.window);
    }
  }
}


void
availrep_show(Project *prj)
{
  Report *r;

  if(!prj->report.window) {
    availrep_new(prj);
  }
   
  r = ((Report *)prj->report.arg);

  report_show_pixmaps(r);
  gtk_notebook_set_page(GTK_NOTEBOOK(r->notebook), 0);
  gtk_widget_show(prj->report.window);
}


extern void
availrep_removed(Project *prj, const char *type, userEntryList *ulist)
{
  userEntry *ue;

  if(!ulist->ul_head) {
    return;
  }

  if(!prj->report.window) {
    availrep_new(prj);
  }

  for(ue = ulist->ul_head; ue; ue = ulist->ul_head) {
    ulist->ul_head = ue->ue_next;

    report_add_clist_item((Report *)prj->report.arg, TAB_REMOVED,
                          ue->ue_name, "<Whole Packet>", type);
    FreeMem(ue);
  }

  ulist->ul_head = ulist->ul_tail = NULL;
}


void 
availrep_conflict(Project *prj, const char *type, AvailPacket *apack, 
                  AvailVersion *aver, AvailPacket *pack, AvailVersion *ver)
{
  if(!prj->report.window) {
    availrep_new(prj);
  }

  report_add_clist_item((Report *)prj->report.arg, TAB_CONFLICT,
                        apack->packet->name, (aver ? aver->version->name :
                                              "<Whole Packet>"),
                        pack->packet->name, (ver ? ver->version->name :
                                             "<Whole Packet>"), type);
}


int
availrep_check(Project *prj, Packet *packet, time_t curtime, SpecVar *svar,
               int have_svar, userEntryList *installed)
{
  int isdef;
  Version *ver;
  userEntry *ue, *uetmp;

  if(!installed->ul_head) {
    return 0; /* no installed packets to check */
  }

  if(!expire_check(prj, packet->info.expire, curtime, packet->name, 
                   "<Whole Packet>", packet->type)) {
    return 0; /* expierd packet - nothing more to do */
  }

  modified_check(prj, packet->info.modified, installed, packet->name, NULL,
                 packet->type);

  isdef = default_check(prj, packet, installed, svar, have_svar);

  for(ue = installed->ul_head; ue; ue = uetmp) {
    uetmp = ue->ue_next;

    if(!(ver = version_find(&packet->vers, ue->ue_version))) { 
      if(!prj->report.window) { /* it's really removed */
        availrep_new(prj);
      }

      report_add_clist_item((Report *)prj->report.arg, TAB_REMOVED,
                            ue->ue_name, ue->ue_version, packet->type);

      FreeMem(_entryRemove(ue, installed));
      continue;
    }

    if(packet->info.forced) { /* is this a forced packet? */
      if(!ver->info.forced) { /* not a forced version installed - remove */
        if(!prj->report.window) {
          availrep_new(prj);
        }

        project_removed(prj, ue->ue_name, ue->ue_version, packet->type);

        report_add_clist_item((Report *)prj->report.arg, TAB_REMOVED,
                              ue->ue_name, ue->ue_version, packet->type);
      }
    } else { /* got a project spec. check with it. */
      if(have_svar && !(!svar && packet->required)) {
        if(!specvar_is_version(svar, ver->name)) {
          if(!prj->report.window) { /* Not in the project spec. */
            availrep_new(prj);
          }

          project_removed(prj, ue->ue_name, ue->ue_version, packet->type);

          report_add_clist_item((Report *)prj->report.arg, TAB_REMOVED,
                                ue->ue_name, ue->ue_version, packet->type);
        }
      }
    }

    expire_check(prj, ver->info.expire, curtime, packet->name, ver->name, 
                 packet->type);

    modified_check(prj, ver->info.modified, installed, packet->name,
                   ver->name, packet->type);
  }

  return isdef;
}


static int
expire_check(Project *prj, time_t expire, time_t curtime, const char *name, 
             const char *version, const char *type)
{
  char buf[64];

  if(expire) {
    if(!prj->report.window) {
      availrep_new(prj);
    }

    if(expire < curtime) {
      report_add_clist_item((Report *)prj->report.arg, TAB_EXPIRED, name, 
                            version, type);
      return 0;
    } else {
      strftime(buf, 64, "%b %e %Y", localtime(&expire));
      report_add_clist_item((Report *)prj->report.arg, TAB_EXPIRE, buf, name, 
                            version, type);
    }
  }

  return 1;
}


static void
modified_check(Project *prj, time_t modified, userEntryList *ulist,
               const char *name, const char *version, const char *type)
{
  char buf[64];
  userEntry *ue;

  if(modified) {
    for(ue = ulist->ul_head; ue; ue = ue->ue_next) {
      if(version && strcasecmp(version, ue->ue_version)) {
        continue;
      }

      if(modified > ue->ue_date) {
        if(!prj->report.window) {
          availrep_new(prj);
        }

        strftime(buf, 64, "%b %e %Y", localtime(&modified));
        report_add_clist_item((Report *)prj->report.arg, TAB_MODIFIED, buf, 
                              name, (version ? version : "<Whole Packet>"), 
                              type);
        return;
      }
    }
  }
}


static int
default_check(Project *prj, Packet *packet, userEntryList *installed,
              SpecVar *svar, int have_svar )
{
  size_t n;
  int defch;
  Version *ver;
  userEntry *ue;
  GString *gs_new, *gs_old;

  /* check if any of the installed versions have changed default flag */

  for(defch = -1, ue = installed->ul_head; ue; ue = ue->ue_next) {
    if(ue->ue_flags != UE_DEFAULT) {
      defch = FALSE;
      break;
    }

    if(ue->ue_flags == UE_DEFAULT) {
      if(!(ver = version_find(&packet->vers, ue->ue_version)) || 
         (have_svar && !specvar_is_version(svar, ver->name)) ||
         !ver->isdef) {
        defch = TRUE;
      }
    }
  }

  if(defch == -1) {
    /* check if any if the versions who has default flag isn't installed */

    for(n = 0; n < packet->vers.siz; n++) {
      ver = parr_get(&packet->vers, Version, n);

      if(ver->isdef) {
        if(have_svar && !specvar_is_version(svar, ver->name)) {
          defch = TRUE;
          break;
        }

        for(ue = installed->ul_head; ue; ue = ue->ue_next) {
          if(!strcasecmp(ue->ue_version, ver->name)) {
            break;
          }
        }

        if(!ue) {
          defch = TRUE;
          break;
        }
      }
    }
  }

  if(defch != TRUE) {
    return (defch == -1 ? 1 : 0);
  }

  if(have_svar && !svar) {
    return 0;
  }

  ue = installed->ul_head;
  gs_old = g_string_new(ue->ue_version);

  for(ue = ue->ue_next; ue; ue = ue->ue_next) {
    gs_old = g_string_append(gs_old, ", ");
    gs_old = g_string_append(gs_old, ue->ue_version);
  }

  gs_new = g_string_new("");

  for(n = 0; n < packet->vers.siz; n++) {
    ver = parr_get(&packet->vers, Version, n);

    if(have_svar && !specvar_is_version(svar, ver->name)) {
      defch = FALSE;
      break;
    }

    if(ver->isdef) {
      if(gs_new->str[0]) {
        gs_new = g_string_append(gs_new, ", ");
      }

      gs_new = g_string_append(gs_new, ver->name);
    }
  }

  if((defch == FALSE) || !gs_new->str[0]) {
    gs_new = g_string_assign(gs_new, "<No Default>");
  }

  if(!prj->report.window) {
    availrep_new(prj);
  }

  report_add_clist_item((Report *)prj->report.arg, TAB_CHANGED, packet->name, 
                        gs_old->str, gs_new->str, packet->type);

  g_string_free(gs_old, TRUE);
  g_string_free(gs_new, TRUE);
  return 1;
}


static void
availrep_new(Project *prj)
{
  Report *r;
  char title[128];
  GtkWidget *vbox, *w;

  hsh_create_str(title, 128, "Packet Load Report For ", prj->name, NULL);

  dialog_new(&prj->report, title,
             "OK", BUTTON_ISDEF, NULL, NULL, NULL, NULL);

  gtk_widget_set_usize(prj->report.window, 530, 300);

  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(prj->report.window)->vbox), vbox, 
                     TRUE, TRUE, 0);

  gui_pixmap_create_dot(&_gmain.pixmaps, prj->report.window);

  r = report_new(6, _gmain.pixmaps.dot.pix, _gmain.pixmaps.dot.mask);
  prj->report.arg = (void *)r;

  report_tab_set_clist(r, TAB_EXPIRED, "Expired", "Expired packets.",
                       cb_select_row, prj, 3);
  report_tab_clist_col_set(r, TAB_EXPIRED, 0, "Packet", 140);
  report_tab_clist_col_set(r, TAB_EXPIRED, 1, "Version", 130);
  report_tab_clist_col_set(r, TAB_EXPIRED, 2, "For Type", -1);

  report_tab_set_clist(r, TAB_REMOVED, "Removed", "Removed packets.",
                       cb_select_row, prj, 3);
  report_tab_clist_col_set(r, TAB_REMOVED, 0, "Packet", 140);
  report_tab_clist_col_set(r, TAB_REMOVED, 1, "Version", 130);
  report_tab_clist_col_set(r, TAB_REMOVED, 2, "For Type", -1);

  report_tab_set_clist(r, TAB_EXPIRE, "Expire", "Packets that will "
                       "expire.", cb_select_row, prj, 4);
  report_tab_clist_col_set(r, TAB_EXPIRE, 0, "Date", 80);
  report_tab_clist_col_set(r, TAB_EXPIRE, 1, "Packet", 140);
  report_tab_clist_col_set(r, TAB_EXPIRE, 2, "Version", 130);
  report_tab_clist_col_set(r, TAB_EXPIRE, 3, "For Type", -1);

  report_tab_set_clist(r, TAB_MODIFIED, "Modified", "Modified packets "
                       "since last update.", cb_select_row, prj, 4);
  report_tab_clist_col_set(r, TAB_MODIFIED, 0, "Date", 80);
  report_tab_clist_col_set(r, TAB_MODIFIED, 1, "Packet", 140);
  report_tab_clist_col_set(r, TAB_MODIFIED, 2, "Version", 130);
  report_tab_clist_col_set(r, TAB_MODIFIED, 3, "For Type", -1);

  report_tab_set_clist(r, TAB_CHANGED, "Defaults", "Default "
                       "versions that have changed.", cb_select_changed, prj,
                       4);
  report_tab_clist_col_set(r, TAB_CHANGED, 0, "Packet", 130);
  report_tab_clist_col_set(r, TAB_CHANGED, 1, "Old Versions", 130);
  report_tab_clist_col_set(r, TAB_CHANGED, 2, "New Versions", 130);
  report_tab_clist_col_set(r, TAB_CHANGED, 3, "For Type", -1);

  report_tab_set_clist(r, TAB_CONFLICT, "Conflicts", "Conflicting packets, "
                       "the first packet have been removed, if possible.", 
                       cb_select_row, prj, 5);
  report_tab_clist_col_set(r, TAB_CONFLICT, 0, "Packet", 100);
  report_tab_clist_col_set(r, TAB_CONFLICT, 1, "Version", 100);
  report_tab_clist_col_set(r, TAB_CONFLICT, 2, "With Packet", 100);
  report_tab_clist_col_set(r, TAB_CONFLICT, 3, "With Version", 100);
  report_tab_clist_col_set(r, TAB_CONFLICT, 4, "For Type", -1);

  w = report_create(r);
  gtk_box_pack_start(GTK_BOX(vbox), w, TRUE, TRUE, 0);
  gtk_widget_show_all(vbox);  
}


static void
cb_select_row(GtkWidget *widget, gint row, gint column, GdkEventButton *event,
              gpointer data)
{
  gchar *packet, *version, *type;
  ReportTabCList *rtc = (ReportTabCList *)data;
  Project *prj = (Project *)rtc->rtab.arg;

  if(event && (event->type == GDK_2BUTTON_PRESS)) {
    if(rtc->colSize == 5) {
      gtk_clist_get_text(GTK_CLIST(rtc->clist), row, 0, &packet);
      gtk_clist_get_text(GTK_CLIST(rtc->clist), row, 1, &version);
      gtk_clist_get_text(GTK_CLIST(rtc->clist), row, 4, &type);
    } else {
      gtk_clist_get_text(GTK_CLIST(rtc->clist), row, rtc->colSize-3, &packet);
      gtk_clist_get_text(GTK_CLIST(rtc->clist), row, rtc->colSize-2, &version);
      gtk_clist_get_text(GTK_CLIST(rtc->clist), row, rtc->colSize-1, &type);
    }

    if(!strcmp(version, "<Whole Packet>")) {
      gui_find_expand(packet, NULL, type, prj, 0);
    } else {
      gui_find_expand(packet, version, type, prj, 0);
    }
  }
}


static void
cb_select_changed(GtkWidget *widget, gint row, gint column,
                  GdkEventButton *event, gpointer data)
{
  Node *node;
  gchar *packet, *type;
  ReportTabCList *rtc = (ReportTabCList *)data;
  Project *prj = (Project *)rtc->rtab.arg;

  if(event && (event->type == GDK_2BUTTON_PRESS)) {
    gtk_clist_get_text(GTK_CLIST(rtc->clist), row, 0, &packet);
    gtk_clist_get_text(GTK_CLIST(rtc->clist), row, 3, &type);

    if((node = gui_find_expand(packet, NULL, type, prj, 0))) {
      hview_freeze(&prj->packets.hview);
      gtk_ctree_expand(hview_ctree_get(&prj->packets.hview), 
                       ((NodeTree *)node)->ctnode);
      hview_thaw(&prj->packets.hview);
       
      gtk_ctree_node_moveto(hview_ctree_get(&prj->packets.hview),
                            ((NodeTree *)node)->ctnode, 0, 0.5, 0.0);
    }
  }
}
