/* HickUP 1.2 -- Graphical interface to hick.
** Copyright (C) 1998 - 2001, 2008 Daniel Bergstrom <daniel@octocode.com>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2, or (at your option)
** any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <gtk/gtkentry.h>
#include <gtk/gtkspinbutton.h>
#include <gtk/gtkoptionmenu.h>
#include <gtk/gtkcheckbutton.h>

#include "refs.h"
#include "pref.h"
#include "filter.h"

#ifdef USE_REGEX
# ifdef HAVE_REGEXEC
#  include <regex.h>
#  define REGEXEC_FLAGS (REG_EXTENDED | REG_NOSUB)
# else /* not HAVE_REGEXEC */
#  include <libgen.h>
# endif /* not HAVE_REGEXEC */
#endif /* USE_REGEX */


#define FILTER_DAYS_OFF 30
#define FILTER_CHECK_FILE  ".checkdate"

#define set_selection(mask, flag, select) \
   gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(select), \
                                ((mask & flag) == flag));

#define get_selection(mask, flag, select) \
  (mask) |= (GTK_TOGGLE_BUTTON(select)->active ? flag : 0);

#define is_leap(y) ((!((y) % 4) && ((y) % 100)) || !((y) % 400));

static int _month_len[2][12] = {
  { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 }
};

static int get_time(Filter *filter, AvailPacket *apack);
static int get_match(Filter *filter, AvailPacket *apack);
static int get_select_hide(unsigned int flags, unsigned int show, 
                           AvailPacket *apack);

static void add_packet(AvailPacket *apack, NTList *ntlist);
static void create_subgroups(Node *sub, NTList *parent, AvailPacket *apack);

/* from search.c */
extern void str_ncasecpy(char *s1, const char *s2, size_t siz);


void
filter_init(Filter *filter)
{
  filter->only_ok        = 0;
  filter->match_str      = 0;
  filter->match_strcase  = 0;
  filter->match          = FILTER_NONE;
  filter->match_not      = 0;
  filter->match_versions = 1;
  filter->select_hide    = FILTER_SELECT_HIDE_NORMAL;
  filter->select_show    = FILTER_SELECT_SHOW_NORMAL;
  filter->time_show      = FILTER_ALL;
  filter->time_before    = 0;
  filter->time_versions  = 1;
  filter->time_check     = time(0);

#ifdef HAVE_REGEXEC
  filter->preg_alloc = 0;
#endif /* HAVE_REGEXEC */
}


void 
filter_clear(Filter *filter)
{
  if(filter->match_str) {
    free(filter->match_str);
  }

  if(filter->match_strcase) {
    free(filter->match_strcase);
  }

#ifdef HAVE_REGEXEC
  if(filter->preg_alloc) {
    regfree(&filter->preg);
  }
  filter->preg_alloc = 0;
#endif /* HAVE_REGEXEC */

  filter->only_ok        = 0;
  filter->match_str      = 0;
  filter->match_strcase  = 0;
  filter->match          = FILTER_NONE;
  filter->match_not      = 0;
  filter->match_versions = 1;
  filter->select_hide    = FILTER_SELECT_HIDE_NORMAL;
  filter->select_show    = FILTER_SELECT_SHOW_NORMAL;
  filter->time_show      = FILTER_ALL;
  filter->time_before    = 0;
  filter->time_versions  = 1;
  filter->time_check     = time(0);
}


void
filter_copy(Filter *to, Filter *from)
{
  filter_clear(to);

  to->only_ok = from->only_ok;
  to->match_not = from->match_not;
  to->match_versions = from->match_versions;
  to->match = from->match;
  to->select_hide = from->select_hide;
  to->select_show = from->select_show;
  to->time_show = from->time_show;
  to->time_before = from->time_before;
  to->time_versions = from->time_versions;
  to->time_check = from->time_check;

  if(from->match_str) {
    to->match_str = MallocArr(char, strlen(from->match_str) + 1);
    strcpy(to->match_str, from->match_str);
  }

  if(from->match_strcase) {
    to->match_strcase = MallocArr(char, strlen(from->match_strcase) + 1);
    strcpy(to->match_strcase, from->match_strcase);
  }

#ifdef HAVE_REGEXEC
  if((to->preg_alloc = from->preg_alloc)) {
    regcomp(&to->preg, from->match_str, REGEXEC_FLAGS);
  }
#endif /* HAVE_REGEXEC */
}


void
filter_get(Filter *filter, HView *hview)
{
  size_t n;
  char *name;
  NTList *ntlist;
  Node *node, *sub;
  AvailPacket *apack;

  for(node = hview->avail->head; node; node = node->next) {
    apack = (AvailPacket *)node;
    ntlist = refs_lookup_type(apack->packet->type, &hview->display);

    if(!get_select_hide(filter->select_hide, filter->select_show, apack)) {
      continue;
    }

    if((filter->time_show != FILTER_ALL) && !get_time(filter, apack)) {
      continue;
    }

    if((filter->match != FILTER_NONE) && !get_match(filter, apack)) {
      continue;
    }

    for(n = 0; n < apack->packet->groups.siz; n++) {
      name = parr_get(&apack->packet->groups, char, n);

      if(!(sub = node_find(name, &apack->packet->subgroups))) {
        add_packet(apack, refs_lookup_group(name, ntlist));
      } else {
        create_subgroups(sub, ntlist, apack);
      }
    }
  }
}


static int
get_time_check(unsigned int flags, time_t check_date, PVInfo *info)
{
  switch(flags) {
    case FILTER_CREATED: {
      return (info->created >= check_date);
    }
    case FILTER_MODIFIED: {
      return (info->modified >= check_date);
    }
    case FILTER_EXPIRE: {
      return (info->expire >= check_date);
    }
  }

  return 0;
}


static int
get_time(Filter *filter, AvailPacket *apack)
{
  size_t n;
  AvailVersion *aver;

  if(get_time_check(filter->time_show, filter->time_check, 
                    &apack->packet->info)) {
    return !filter->time_before;
  }

  if(filter->time_versions) {
    for(n = 0; n < apack->vers.siz; n++) {
      aver = parr_get(&apack->vers, AvailVersion, n);

      if(get_time_check(filter->time_show, filter->time_check, 
                        &aver->version->info)) {
        return !filter->time_before;
      }
    }
  }

  return filter->time_before;
}


static int
get_match_string(Filter *filter, const char *search_name)
{
  char name[PNAME_LEN + 1];

  switch(filter->match) {
    case FILTER_MATCH: {
      if(!filter->match_strcase) {
        return 0;
      }

      str_ncasecpy(name, search_name, PNAME_LEN);
      return (strstr(name, filter->match_strcase) ? 1 : 0);
    }
    case FILTER_CASE_MATCH: {
      return (strstr(search_name, filter->match_str) ? 1 : 0);
    }
#ifdef USE_REGEX
    case FILTER_REGEX_MATCH: {
#ifdef HAVE_REGEXEC
      if(!filter->preg_alloc) {
        return 0;
      }

      return (regexec(&filter->preg, search_name, 0, NULL, 0) ? 0 : 1);
#else  /* HAVE_REGEX */
      if(!filter->match_strcase) {
        return 0;
      }

      return (regex(filter->match_strcase, search_name, 0) ? 1 : 0);
#endif /* HAVE_REGEX */
    }
#endif /* USE_REGEX */
  }

  return 0;
}


static int
get_match(Filter *filter, AvailPacket *apack)
{
  size_t n;
  AvailVersion *aver;

  if(get_match_string(filter, apack->packet->name)) {
    return !filter->match_not;
  }

  if(filter->match_versions) {
    for(n = 0; n < apack->vers.siz; n++) {
      aver = parr_get(&apack->vers, AvailVersion, n);

      if(get_match_string(filter, aver->version->name)) {
        return !filter->match_not;
      }
    }
  }

  return filter->match_not;
}


static int
get_select_show(unsigned int flags, AvailPacket *apack)
{
  if((flags & FILTER_HIDDEN) && apack->packet->hidden) {
    return 1;
  }

  if((flags & FILTER_NOTSELECTED) && (apack->info.install == INSTALL_NOT)) {
    return 1;
  }

  if((flags & FILTER_SELECTED) && (apack->info.install == INSTALL_VERSION)) {
    return 1;
  }

  if((flags & FILTER_DEFAULT) && (apack->info.install == INSTALL_DEFAULT)) {
    return 1;
  }

  if((flags & FILTER_FORCED) && (apack->info.install == INSTALL_FORCE)) {
    return 1;
  }

  if((flags & FILTER_REQUIRED) && apack->packet->required) {
    return 1;
  }

  return 0;
}


static int
get_select_hide(unsigned int flags, unsigned int show, AvailPacket *apack)
{
  if((flags & FILTER_HIDDEN) && apack->packet->hidden &&
     !get_select_show(show, apack)) {
    return 0;
  }

  if((flags & FILTER_NOTSELECTED) && (apack->info.install == INSTALL_NOT) &&
     !get_select_show(show, apack)) {
    return 0;
  }

  if((flags & FILTER_SELECTED) && (apack->info.install == INSTALL_VERSION) &&
     !get_select_show(show, apack)) {
    return 0;
  }

  if((flags & FILTER_DEFAULT) && (apack->info.install == INSTALL_DEFAULT) &&
     !get_select_show(show, apack)) {
    return 0;
  }

  if((flags & FILTER_FORCED) && (apack->info.install == INSTALL_FORCE) &&
     !get_select_show(show, apack)) {
    return 0;
  }

  if((flags & FILTER_REQUIRED) && apack->packet->required &&
     !get_select_show(show, apack)) {
    return 0;
  }

  return 1;
}


static void
create_subgroups(Node *sub, NTList *parent, AvailPacket *apack)
{
  Node *node;
  NTList *ntlist;

  ntlist = refs_lookup_group(sub->name, parent);

  if(((SubGroup *)sub)->here) {
    add_packet(apack, ntlist);
  }

  for(node = ((SubGroup *)sub)->down.head; node; node = node->next) {
    create_subgroups(node, ntlist, apack);
  }
}


static void
add_packet(AvailPacket *apack, NTList *ntlist)
{
  size_t n;
  RefPacket *rp;
  AvailVersion *aver;

  rp = refs_packet_new(apack->packet->name, apack, ntlist);

  for(n = 0; n < apack->vers.siz; n++) {
    aver = parr_get(&apack->vers, AvailVersion, n);
    refs_version_new(aver->version->name, aver, rp);
  }
}


static void
filter_gui_set_selection(unsigned int mask, GtkWidget *select[])
{
  set_selection(mask, FILTER_HIDDEN,      select[0]);
  set_selection(mask, FILTER_NOTSELECTED, select[1]);
  set_selection(mask, FILTER_SELECTED,    select[2]);
  set_selection(mask, FILTER_DEFAULT,     select[3]);
  set_selection(mask, FILTER_FORCED,      select[4]);
  set_selection(mask, FILTER_REQUIRED,    select[5]);
}


void 
filter_gui_set(Filter *filter, TabFilter *ftab)
{
  struct tm *tm;

  gtk_option_menu_set_history(GTK_OPTION_MENU(ftab->match), 
                              filter->match + filter->match_not);
  gtk_entry_set_text(GTK_ENTRY(ftab->string), 
                     (filter->match_str ? filter->match_str : ""));

  gtk_option_menu_set_history(GTK_OPTION_MENU(ftab->dateset), 
                              filter->time_show + filter->time_before);

  tm = localtime(&filter->time_check);

  gtk_spin_button_set_value(GTK_SPIN_BUTTON(ftab->day), (gfloat)tm->tm_mday);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(ftab->month),
                            (gfloat)tm->tm_mon + 1);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(ftab->year),
                            (gfloat)tm->tm_year + 1900);

  cb_filter_update_date(NULL, ftab);

  filter_gui_set_selection(filter->select_hide, ftab->select_hide);
  filter_gui_set_selection(filter->select_show, ftab->select_show);
}


static void
filter_gui_get_selection(unsigned short *mask, GtkWidget *select[])
{
  *mask = 0;
  get_selection(*mask, FILTER_HIDDEN,      select[0]);
  get_selection(*mask, FILTER_NOTSELECTED, select[1]);
  get_selection(*mask, FILTER_SELECTED,    select[2]);
  get_selection(*mask, FILTER_DEFAULT,     select[3]);
  get_selection(*mask, FILTER_FORCED,      select[4]);
  get_selection(*mask, FILTER_REQUIRED,    select[5]);
}


static int 
filter_gui_get_menu_index(GtkWidget *optionmenu, GtkWidget *menus[], int siz)
{
  size_t n;

  for(n = 0; n < siz; n++) {
    if(menus[n] == GTK_OPTION_MENU(optionmenu)->menu_item) {
      return n;
    }
  }

  return 0;
}


void
filter_gui_get(Filter *filter, TabFilter *ftab)
{
  gchar *text;
  char *regerr;
  struct tm tm;
  int n, len, reg;
  GtkAdjustment *adjDay, *adjMonth, *adjYear;

  filter_clear(filter);

  text = gtk_entry_get_text(GTK_ENTRY(ftab->string));

  if(text && text[0]) {
    len = strlen(text);
    filter->match_str = MallocArr(char, len + 1);
    strcpy(filter->match_str, text);
  }

  n = filter_gui_get_menu_index(ftab->match, ftab->match_menus, 
                                MENU_MATCH_SIZE);

  switch(n) {
    case FILTER_NONE: {
      filter->match = FILTER_NONE;
      break;
    }
    case FILTER_MATCH + 1: {
      filter->match_not = 1;
    }
    case FILTER_MATCH: {
      filter->match = FILTER_MATCH;

      if(text && text[0]) {
        len = strlen(text);
        filter->match_strcase = MallocArr(char, len + 1);
        str_ncasecpy(filter->match_strcase, filter->match_str, len);
        filter->match_strcase[len] = (char)0;
      }
      break;
    }
    case FILTER_CASE_MATCH + 1: {
      filter->match_not = 1;
    }
    case FILTER_CASE_MATCH: {
      filter->match = FILTER_CASE_MATCH;
      break;
    }
#ifdef USE_REGEX
    case FILTER_REGEX_MATCH + 1: {
      filter->match_not = 1;
    }
    case FILTER_REGEX_MATCH: {
      filter->match = FILTER_REGEX_MATCH;

#ifdef HAVE_REGEXEC
      filter->match_strcase = 0;

      if(!(reg = regcomp(&filter->preg, filter->match_str, REGEXEC_FLAGS))) {
        filter->preg_alloc = 1;
      } else {
        filter->preg_alloc = 0;
        if((n = regerror(reg, &filter->preg, 0, 0)) && 
           (regerr = MallocArr(char, n + 1))) {
          regerror(reg, &filter->preg, regerr, n);
          dialog_err_return_new(0, "Regex Error", "Regex error: %s", regerr);
          FreeMem(regerr);
        }
      }
#else  /* HAVE_REGEX */
      filter->match_strcase = regcmp(filter->match_str, 0);
#endif /* HAVE_REGEX */
      break;
    }
#endif /* USE_REGEX */
  }

  
  n = filter_gui_get_menu_index(ftab->dateset, ftab->date_menus, 
                                MENU_DATE_SIZE);

  switch(n) {
    case FILTER_ALL: {
      filter->time_show = FILTER_ALL;
      break;
    }
    case FILTER_CREATED + 1: {
      filter->time_before = 1;
    }
    case FILTER_CREATED: {
      filter->time_show = FILTER_CREATED;
      break;
    }
    case FILTER_MODIFIED + 1: {
      filter->time_before = 1;
    }
    case FILTER_MODIFIED: {
      filter->time_show = FILTER_MODIFIED;
      break;
    }
    case FILTER_EXPIRE + 1: {
      filter->time_before = 1;
      break;
    }
    case FILTER_EXPIRE: {
      filter->time_show = FILTER_EXPIRE;
      break;
    }
  }

  adjDay   = gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(ftab->day));
  adjMonth = gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(ftab->month));
  adjYear  = gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(ftab->year));

  tm.tm_wday  = tm.tm_yday = tm.tm_sec = tm.tm_min = tm.tm_hour = 0;
  tm.tm_mday  = (int)adjDay->value;
  tm.tm_mon   = (int)adjMonth->value - 1;
  tm.tm_year  = (int)adjYear->value - 1900;
  tm.tm_isdst = -1;

  filter->time_check = mktime(&tm);

  filter_gui_get_selection(&filter->select_hide, ftab->select_hide);
  filter_gui_get_selection(&filter->select_show, ftab->select_show);
}


void
cb_filter_reset(GtkWidget *w, gpointer data)
{
  Filter filter;

  filter_init(&filter);
  filter_gui_set(&filter, (TabFilter *)data);
}


static time_t
get_timestamp(void)
{
  FILE *file;
  time_t t = 0;
  char fname[FNAME_LEN+1];

  hsh_create_str(fname, FNAME_LEN, _pref.path, "/", FILTER_CHECK_FILE, NULL);

  if((file = fopen(fname, "r"))) {
    if(fgets(fname, FNAME_LEN, file)) {
      t = atol(fname);
    }

    fclose(file);
  }

  if(!t) {
    time(&t);
    t -= FILTER_DAYS_OFF * 24 * 60 * 60;
  }

  return t;
}

static void
put_timestamp(void)
{
  FILE *file;
  char fname[FNAME_LEN+1];

  hsh_create_str(fname, FNAME_LEN, _pref.path, "/", FILTER_CHECK_FILE, NULL);

  if((file = fopen(fname, "w"))) {
    fprintf(file, "%ld\n", time(NULL));
    fclose(file);
  }
}


void
cb_filter_newpackets(GtkWidget *w, gpointer data)
{
  Filter filter;

  filter_init(&filter);

  filter.time_show     = FILTER_CREATED;
  filter.time_before   = 0;
  filter.time_versions = 1;
  filter.time_check    = get_timestamp();
  put_timestamp();

  filter_gui_set(&filter, (TabFilter *)data);
}


void
cb_filter_update_date(GtkWidget *w, gpointer data)
{
  int leap;
  TabFilter *ftab = (TabFilter *)data;
  GtkAdjustment *adjDay, *adjMonth, *adjYear;

  adjDay =   gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(ftab->day));
  adjMonth = gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(ftab->month));
  adjYear =  gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(ftab->year));

  leap = is_leap((int)adjYear->value);

  adjDay->upper = (gfloat)_month_len[leap][(int)adjMonth->value - 1];

  if(adjDay->value > adjDay->upper) {
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(ftab->day), adjDay->upper);
  }
}

