/* HickUP 1.2 -- Graphical interface to hick.
** Copyright (C) 1998 - 2001, 2008 Daniel Bergstrom <daniel@octocode.com>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2, or (at your option)
** any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <stdarg.h>

#include <gtk/gtkhbox.h>
#include <gtk/gtkmenu.h>
#include <gtk/gtkpixmap.h>
#include <gtk/gtkmenuitem.h>

#include "gui.h"
#include "refs.h"
#include "hview.h"
#include "wwrap.h"


static int hview_get_avail(NodeTree *ntree, AvailPacket **apack, 
                           AvailVersion **aver);


GtkWidget *
hview_new(HView *hview, enum hview_type type, gint columns, gchar **titles, 
          gint *cwidth, Project *prj)
{
  GtkWidget *w;

  hview->menu.menu = NULL;
  hview->select.packet = NULL;
  hview->select.version = NULL;
  parr_init(&hview->menu.items);
  hview->type    = type;
  hview->avail   = NULL;
  hview->project = prj;
  hsh_nodelist_init(&hview->display);
  hview->pixfunc = NULL;
  
  if(type == HVIEW_CTREE) {
    w = ww_ctree(&hview->view, columns, 1, titles, cwidth);
  } else {
    w = ww_clist(&hview->view, columns, titles, cwidth);
  }

  return w;
}


void 
hview_menu_create(HView *hview, ...)
{
  char *p;
  int type;
  va_list pa;
  GPixmap *pixi;
  GtkSignalFunc sfunc;
  HViewMenuItem *hitem;
  GtkWidget *item, *box;

  if(!hview->menu.menu) {
    hview->menu.menu = gtk_menu_new();
  }

  va_start(pa, hview);

  while((p = va_arg(pa, char *))) {
    pixi = va_arg(pa, GPixmap *);
    type = va_arg(pa, int);
    sfunc = va_arg(pa, GtkSignalFunc);

    if(*p == (char)0) {
      item = gtk_menu_item_new();      
    } else {
      hitem = MallocMem(HViewMenuItem);
      hitem->type = (enum hview_menu_type)type;

      if(!pixi) {
        pixi = &_gmain.pixmaps.empty;
      }

      box = gtk_hbox_new(FALSE, 3);
      hitem->pixmap = gtk_pixmap_new(pixi->pix, pixi->mask);
      gtk_box_pack_start(GTK_BOX(box), hitem->pixmap, FALSE, TRUE, 0);

      hitem->label = gtk_label_new(p);
      gtk_box_pack_start(GTK_BOX(box), hitem->label, FALSE, TRUE, 0);

      hitem->item = item = gtk_menu_item_new();
      gtk_container_add(GTK_CONTAINER(item), box);

      gtk_signal_connect(GTK_OBJECT(hitem->item), "activate", sfunc, hview);
      parr_add(&hview->menu.items, hitem);
    }

    gtk_menu_append(GTK_MENU(hview->menu.menu), item);
    gtk_widget_show_all(item);
  }
}


void
hview_menu_set_sensitive(HView *hview, enum hview_menu_type type, 
                         gboolean sensitive)
{
  size_t n;
  HViewMenuItem *hitem;

  for(n = 0; n < hview->menu.items.siz; n++) {
    hitem = parr_get(&hview->menu.items, HViewMenuItem, n);

    if((type == MENU_ALL) || (hitem->type == type)) {
      gtk_widget_set_sensitive(hitem->item, sensitive);
    }
  }
}


void
hview_menu_set_show(HView *hview, enum hview_menu_type type, gboolean show)
{
  size_t n;
  HViewMenuItem *hitem;

  for(n = 0; n < hview->menu.items.siz; n++) {
    hitem = parr_get(&hview->menu.items, HViewMenuItem, n);

    if((type == MENU_ALL) || (hitem->type == type)) {
      if(show) {
        gtk_widget_show(hitem->item);
        gtk_widget_set_sensitive(hitem->item, TRUE);
      } else {
        gtk_widget_hide(hitem->item);
      }
    }
  }
}


void 
hview_menu_set_label(HView *hview, enum hview_menu_type type, const char *p)
{
  size_t n;
  HViewMenuItem *hitem;

  for(n = 0; n < hview->menu.items.siz; n++) {
    hitem = parr_get(&hview->menu.items, HViewMenuItem, n);

    if((type == MENU_ALL) || (hitem->type == type)) {
      gtk_label_set_text(GTK_LABEL(hitem->label), p);
    }
  }
}


void
hview_menu_set_gpixmap(HView *hview, enum hview_menu_type type, 
                       GPixmap *pixi)
{
  size_t n;
  HViewMenuItem *hitem;

  for(n = 0; n < hview->menu.items.siz; n++) {
    hitem = parr_get(&hview->menu.items, HViewMenuItem, n);

    if(((type == MENU_ALL) || (hitem->type == type)) && hitem->pixmap) {
      gtk_pixmap_set(GTK_PIXMAP(hitem->pixmap), pixi->pix, pixi->mask);
    }
  }
}


void
hview_display_clear(HView *hview)
{
  gtk_clist_freeze(GTK_CLIST(hview->view));
  gtk_clist_clear(GTK_CLIST(hview->view));
  gtk_clist_thaw(GTK_CLIST(hview->view));
}


void 
hview_display_destroy(HView *hview)
{
  hview_display_clear(hview);
  refs_free(&hview->display);
}


int
hview_select_set(HView *hview, NodeTree *ntree)
{
  hview->select.ntree = ntree;

  return hview_get_avail(ntree, &hview->select.packet, &hview->select.version);
}


int
hview_selected_get_avail(HView  *hview, AvailPacket **apack, 
                         AvailVersion **aver)
{
  if(hview->select.ntree && hview->select.packet) {
    *apack = hview->select.packet;
    *aver = hview->select.version;
    return TRUE;
  }

  *apack = 0;
  *aver = 0;
  return FALSE;
}


int
hview_select_get_avail(HView *hview, GdkEventButton *event, 
                       AvailPacket **apack, AvailVersion **aver,
                       gint *row, gint *column)
{
  NodeTree *ntree;
  GtkCTreeNode *ctnode;

  *apack = 0;
  *aver = 0;

  if(!gtk_clist_get_selection_info(hview_clist_get(hview), (int)event->x, 
                                   (int)event->y, row, column)) {
    return FALSE;
  }

  if(hview_is_ctree(hview)) {
    if(!(ctnode = gtk_ctree_node_nth(hview_ctree_get(hview), *row))) {
      return FALSE;
    }

    ntree = (NodeTree *)gtk_ctree_node_get_row_data(hview_ctree_get(hview),
                                                    ctnode);
  } else {
    ntree = (NodeTree *)gtk_clist_get_row_data(hview_clist_get(hview), *row);
  }

  return hview_get_avail(ntree, apack, aver);
}


void
hview_select_row(HView *hview, gint row)
{
  GtkCTreeNode *ctnode;
     
  if(hview_is_ctree(hview)) {
    if((ctnode = gtk_ctree_node_nth(hview_ctree_get(hview), row))) {
      gtk_ctree_select(hview_ctree_get(hview), ctnode);
    }
  } else {
    gtk_clist_select_row(hview_clist_get(hview), row, -1);
  }
}


void
hview_pixmap_display(HView *hview, Node *node, GPixmap *pixi)
{
  if(hview_is_ctree(hview)) {
    if(((NodeTree *)node)->ctnode) {
      gtk_ctree_node_set_pixmap(hview_ctree_get(hview), 
                                ((NodeTree *)node)->ctnode,
                                0, pixi->pix, pixi->mask);
    }
  } else {
    if(((RefData *)node)->row != -1) {
      gtk_clist_set_pixmap(hview_clist_get(hview), ((RefData *)node)->row,
                           0, pixi->pix, pixi->mask);
    }
  }
}


static int
hview_get_avail(NodeTree *ntree, AvailPacket **apack, AvailVersion **aver)
{
  if(ntree) {
    switch(ntree->node.flag) {
      case NT_EMPTY: {
        *aver = ((RefData *)ntree)->version;
        *apack = (*aver ? (*aver)->packet : ((RefData *)ntree)->packet);
        return TRUE;
      }
      case NT_TYPE:
      case NT_GROUP: {
        *aver = 0;
        *apack = 0;
        return TRUE;
      }
      case NT_PACKET: {
        *aver = 0;
        *apack = ((RefPacket *)ntree)->packet;
        return TRUE;
      }
      case NT_VERSION: {
        *aver = ((RefVersion *)ntree)->version;
        *apack = (*aver ? (*aver)->packet : 0);
        return TRUE;
      }
    }
  }

  *aver = 0;
  *apack = 0;
  return FALSE;
}
