/* HickUP 1.2 -- Graphical interface to hick.
** Copyright (C) 1998 - 2001, 2008 Daniel Bergstrom <daniel@octocode.com>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2, or (at your option)
** any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <stdarg.h>
#include <unistd.h>

#include <gtk/gtktext.h>

#include "load.h"
#include "packet.h"


typedef struct {
  Packet *packet;
  Version *version;
} PackVer;

#define packver_get(parr, n)   ((PackVer *)(parr)->arr[(n)])
#define packver_first(parr)    ((PackVer *)(parr)->arr[0])
#define packver_last(parr)     ((PackVer *)(parr)->arr[(parr)->siz - 1])

static void conflict_test(Packet *packet, Version *version, PArr *cons);
static int packetPcmp(const void *d1, const void *d2);

static void require_test(PArr *reqlist, Require *head);
static int packver_add(PArr *rarr, Packet *packet, Version *version);
static void packver_remove_last(PArr *rarr);


void
loaderr_clear(void)
{
  if(_gmain.load_err_dialog.mdialog.dialog.window) {
    ww_text_clear(_gmain.load_err_dialog.text);
    _gmain.load_err_dialog.have_err = FALSE;
  }
}


#define loaderr_write(buf, n) \
       gtk_text_insert(GTK_TEXT(_gmain.load_err_dialog.text), NULL, NULL, \
                       NULL, (buf), (n));


#define loaderr_put(buf, n, c) \
     if((n) + 1 >= 1024) { \
       loaderr_write(buf, n); \
       n = 0; \
     } \
     (buf)[(n)++] = c; \


static void
loaderr_new(void)
{
  dialog_err_new(&_gmain.load_err_dialog, "Packet Errors",
                 "The followind errors occurd during the loading of packet. "
                 "These have to be fixed before HickUP will work correctly.",
                 "OK", BUTTON_ISDEF, NULL, NULL, NULL,
                 NULL);
}


void
loaderr_add(const char *fmt, ...)
{
  va_list ap;
  size_t n = 0;
  char buf[1024], *p;

  if(!_gmain.load_err_dialog.mdialog.dialog.window) {
    loaderr_new();
    gtk_text_freeze(GTK_TEXT(_gmain.load_err_dialog.text));
  }

  va_start(ap, fmt);

  while(*fmt) {
    if(*fmt != '%') {
      loaderr_put(buf, n, *(fmt++));
    } else {
      if((*(++fmt) == 's') && (p = va_arg(ap, char *))) {
        while(*p) {
          loaderr_put(buf, n, *(p++));
        }

        fmt++;
      }
    }
  }

  if(n) {
    loaderr_write(buf, n);
  }

  _gmain.load_err_dialog.have_err = TRUE;
}


void
loaderr_show(void)
{
  if(_gmain.load_err_dialog.have_err && _pref.show_errors) {
    gtk_text_thaw(GTK_TEXT(_gmain.load_err_dialog.text));
    gtk_widget_show(_gmain.load_err_dialog.mdialog.dialog.window);
  }
}


void
loaderr_do_show(void)
{
  if(!_gmain.load_err_dialog.mdialog.dialog.window) {
    loaderr_new();
  }

  gtk_widget_show(_gmain.load_err_dialog.mdialog.dialog.window);
}


void
loaderr_check_information(NodeList *packList)
{
  size_t n, defnr;
  Node *node;
  PArr reqlist;
  Packet *packet;
  Version *version;

  parr_init(&reqlist);

  for(node = packList->head; node; node = node->next) {
    packet = (Packet *)node;

    if(packet->info.conflicts.siz) {
      conflict_test(packet, NULL, &packet->info.conflicts);
    }

    for(defnr = 0, n = 0; n < packet->vers.siz; n++) {
      version = parr_get(&packet->vers, Version, n);

      if(version->isdef) {
        if(defnr++ && !packet->multiple) {
          version->isdef = FALSE;

          loaderr_add("default: %s %s %s: multiple defaults in single select "
                      "packet.\n", packet->type, packet->name, version->name);
        }
      }

      if(version->info.conflicts.siz) {
        conflict_test(packet, version, &version->info.conflicts);
      }
    }

    if(packet->info.require) {
      packver_add(&reqlist, packet, NULL);
      require_test(&reqlist, packet->info.require);
      packver_remove_last(&reqlist);
    }

    for(n = 0; n < packet->vers.siz; n++) {
      version = parr_get(&packet->vers, Version, n);

      if(version->info.require) {
        packver_add(&reqlist, packet, version);
        require_test(&reqlist, version->info.require);
        packver_remove_last(&reqlist);
      }
    }
  }
}

static void
conflicts_justify(Packet *packet, Version *version, const char *inpack,
                  const char *inver, PArr *cons)
{
  size_t n;

  if(!(n = parr_find(cons, packet, packetPcmp))) {
    loaderr_add("conflict: %s %s %s: no conflict back from %s %s.\n",
                packet->type, packet->name,
                (version ? version->name : "(whole)"), inpack,
                (inver ? inver : "(whole)"));
    return;
  }

  n--;

  if(parr_get(cons, Conflict, n)->vers.siz && !version) {
    loaderr_add("conflict: %s %s %s: missing whole packet conflict "
                "back from %s %s.\n", packet->type, packet->name,
                (version ? version->name : "(whole)"), inpack,
                (inver ? inver : "(whole)"));
  } else {
    if(parr_get(cons, Conflict, n)->vers.siz && version) {
      if(!version_find(&parr_get(cons, Conflict, n)->vers, version->name)) {
        loaderr_add("conflict: %s %s %s: missing single version conflict "
                    "back from %s %s.\n", packet->type, packet->name,
                    (version ? version->name : "(whole)"), inpack,
                    (inver ? inver : "(whole)"));
      }
    }
  }
}


static void
conflict_test(Packet *packet, Version *version, PArr *cons)
{
  Conflict *con;
  size_t n = 0, i;

  for(n = 0; n < cons->siz; n++) {
    con = parr_get(cons, Conflict, n);

    if(!con->vers.siz) {
      if(con->packet->node.flag == NT_EMPTY) {
        loaderr_add("conflict: %s %s %s: missing packet %s.\n",
                    packet->type, packet->name,
                    (version ? version->name : "(whole)"), con->packet->name);
      } else {
        if(packet->info.forced && con->packet->info.forced) {
          loaderr_add("force: %s %s %s: conflicting force to %s (whole).\n",
                      packet->type, packet->name,
                      (version ? version->name : "(whole)"), 
                      con->packet->name);
        }

        conflicts_justify(packet, version, con->packet->name, NULL,
                          &con->packet->info.conflicts);
      }

      continue;
    }

    for(i = 0; i < con->vers.siz; i++) {
      if(!parr_get(&con->vers, Version, i)->real) {
        loaderr_add("conflict: %s %s %s: missing version %s in %s.\n",
                    packet->type, packet->name,
                    (version ? version->name : "(whole)"),
                    parr_get(&con->vers, Version, i)->name,
                    con->packet->name);
        continue;
      }

      if(((version && version->info.forced) || packet->info.forced) &&
         parr_get(&con->vers, Version, i)->info.forced) {
        loaderr_add("force: %s %s %s: conflicting force to %s in %s.\n",
                    packet->type, packet->name,
                    (version ? version->name : "(whole)"),
                    parr_get(&con->vers, Version, i)->name,
                    con->packet->name);
      }


      conflicts_justify(packet, version, con->packet->name,
                        parr_get(&con->vers, Version, i)->name,
                        &parr_get(&con->vers, Version, i)->info.conflicts);
    }
  }
}


static int
packetPcmp(const void *d1, const void *d2) {
  return (((Conflict *)d1)->packet != d2);
}


static void
require_test(PArr *reqlist, Require *head)
{
  size_t n;
  Require *req;
  Packet *packet = packver_last(reqlist)->packet;
  Version *version = packver_last(reqlist)->version;

  for(req = head; req; req = req->next) { /* packet check */
    if((reqlist->siz == 1) && (req->info.packet->node.flag == NT_EMPTY)) {
      loaderr_add("require: %s %s %s: missing packet %s.\n",
                  packet->type, packet->name,
                  (version ? version->name : "(whole)"),
                  req->info.packet->name);
      continue;
    }

    if(!packver_add(reqlist, req->info.packet, NULL)) {
      continue;
    }

    require_test(reqlist, req->info.packet->info.require);
    packver_remove_last(reqlist);


    for(n = 0; n < req->info.vers.siz; n++) { /* version check */
      if(!parr_get(&req->info.vers, Version, n)->real) {
        loaderr_add("require: %s %s %s: missing version %s in %s.\n",
                    packet->type, packet->name,
                    (version ? version->name : "(whole)"),
                    parr_get(&req->info.vers, Version, n)->name,
                    req->info.packet->name);
        continue;
      }

      if(!packver_add(reqlist, req->info.packet, 
                      parr_get(&req->info.vers, Version, n))) {
        continue;
      }

      require_test(reqlist, 
                   parr_get(&req->info.vers, Version, n)->info.require);
      packver_remove_last(reqlist);
    }

    if(req->ror) { /* go down the or's */
      require_test(reqlist, req->ror);
    }
  }
}


static int
packver_add(PArr *rarr, Packet *packet, Version *version)
{
  size_t n, i;
  PackVer *pv;

  for(n = 0; n < rarr->siz; n++) {
    if((parr_get(rarr, PackVer, n)->packet == packet) &&
       (parr_get(rarr, PackVer, n)->version == version)) {

      if(n != 0) {
        return 0;
      }

      switch(rarr->siz) {
        case 1: {
          loaderr_add("require: %s %s %s: loop in itself.\n",
                      packver_first(rarr)->packet->type, 
                      packver_first(rarr)->packet->name,
                      (packver_first(rarr)->version ?
                       packver_first(rarr)->version->name : "(whole)"));
          return 0;
        }
        case 2: {
          loaderr_add("require: %s %s %s: loop back from %s %s.\n",
                      packver_first(rarr)->packet->type, 
                      packver_first(rarr)->packet->name,
                      (packver_first(rarr)->version ?
                       packver_first(rarr)->version->name : "(whole)"),
                      packver_last(rarr)->packet->name,
                      (packver_last(rarr)->version ?
                       packver_last(rarr)->version->name : "(whole)"));
          return 0;
        }
        default: {
          loaderr_add("require: %s %s %s: loop back from %s %s. Through",
                      packver_first(rarr)->packet->type, 
                      packver_first(rarr)->packet->name,
                      (packver_first(rarr)->version ?
                       packver_first(rarr)->version->name : "(whole)"),
                      packver_last(rarr)->packet->name,
                      (packver_last(rarr)->version ?
                       packver_last(rarr)->version->name : "(whole)"));

          for(i = 1; i < (rarr->siz - 1); i++) {
            if((i != 1) && (i + 1 >= (rarr->siz - 1))) {
              loaderr_add(" and");
            }

            loaderr_add(" %s %s", packver_get(rarr, i)->packet->name,
                        (packver_get(rarr, i)->version ?
                         packver_get(rarr, i)->version->name : "(whole)"));


            if(i + 2 < (rarr->siz - 1)) {
              loaderr_add(",");
            }
          }

          loaderr_add(".\n");
          return 0;
        }
      }
    }
  }

  pv = MallocMem(PackVer);
  pv->packet = packet;
  pv->version = version;

  parr_add(rarr, pv);
  return 1;
}


static void 
packver_remove_last(PArr *rarr)
{
  if(rarr->siz) {
    FreeMem(packver_last(rarr));
    rarr->siz--;
  }
}
