/* HickUP 1.2 -- Graphical interface to hick.
** Copyright (C) 1998 - 2001, 2008 Daniel Bergstrom <daniel@octocode.com>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2, or (at your option)
** any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "load.h"
#include "packet.h"

static void versions_clear(PArr *vers);
static void pvinfo_clear(PVInfo *info);
static void subgroups_clear(NodeList *list);
static void require_clear(Require *rhead);

static void create_subgroups(Packet *packet, HickSubGroup *head, PArr *groups);
static SubGroup *lookup_subgroup(NodeList *list, char *name);
static SubGroup *subgroup_new(char *name, NodeList *list);
static char *lookup_group(PArr *groups, const char *name);


Packet *
packet_new(HickPacket *hpac, const char *name, const char *type, 
           PArr *groups, NodeList *emptyList, NodeList *packList)
{
  char *p;
  size_t n;
  Packet *packet;

  if((packet = packet_find(hpac->pac_name, type, emptyList))) {
    node_move(&packet->node, emptyList, packList);
  } else {
    packet = MallocMem(Packet);

    parr_init(&packet->vers);
    parr_init(&packet->groups);
    parr_init(&packet->info.conflicts);
    hsh_nodelist_init(&packet->subgroups);

    packet->info.about   = NULL;
    packet->info.require = NULL;
    packet->info.forced  = 0;
    packet->node.name    = packet->name;
    packet->type         = (char *)type;

    hsh_node_prepend(&packet->node, packList);
  }

  strcpy(packet->name, name);

  if(hpac) {
    for(n = 0; n < hpac->pac_groups.siz; n++) { /* add groups */
      p = lookup_group(groups, hpac->pac_groups.arr[n]);
      parr_add(&packet->groups, p);
    }

    create_subgroups(packet, hpac->pac_subGroups, groups);

    packet->hidden        = hpac->pac_hidden;
    packet->required      = hpac->pac_required;
    packet->multiple      = hpac->pac_multiple;
    packet->info.about    = hpac->pac_about;
    packet->info.created  = hpac->pac_created;
    packet->info.modified = hpac->pac_modified;
    packet->info.expire   = hpac->pac_expire;

    if(!packet->info.created) {
      loaderr_add("created: %s %s: not set.\n", packet->type, packet->name);
    }
    
    packet->node.flag = NT_PACKET;
  } else {
    packet->node.flag = NT_EMPTY;
  }

  return packet;
}


static void
create_subgroups(Packet *packet, HickSubGroup *head, PArr *groups)
{
  SubGroup *grp;
  HickSubGroup *hsub, *dsub;

  for(hsub = head; hsub; hsub = hsub->hsg_next) {
    grp = lookup_subgroup(&packet->subgroups, 
                          lookup_group(groups, hsub->hsg_group));
    grp = lookup_subgroup(&grp->down, lookup_group(groups, hsub->hsg_name));

    for(dsub = hsub->hsg_down; dsub; dsub = dsub->hsg_down) {
      grp = lookup_subgroup(&grp->down, lookup_group(groups, dsub->hsg_name));
    }

    grp->here = 1;
  }
}


static SubGroup *
lookup_subgroup(NodeList *list, char *name)
{
  Node *node;

  if(!(node = node_find(name, list))) {
    return subgroup_new(name, list);
  }

  return (SubGroup *)node;
}


static SubGroup *
subgroup_new(char *name, NodeList *list)
{
  SubGroup *sub;

  sub = MallocMem(SubGroup);
  sub->here = 0;
  sub->node.flag = 0;
  sub->node.name = name;
  hsh_nodelist_init(&sub->down);

  hsh_node_prepend(&sub->node, list);

  return sub;
}


static char *
lookup_group(PArr *groups, const char *name)
{
  char *p;

  p = parr_find_value(groups, char, name, parr_strcasecmp);

  if(!p) {
    p = MallocArr(char, strlen(name) + 1);
    strcpy(p, name);
    parr_add(groups, p);
  }

  return p;
}


Packet *
packet_find(const char *name, const char *type, NodeList *list)
{
  Node *node;

  if(list) {
    for(node = list->head; node; node = node->next) {
      if((((Packet *)node)->type == type) && !strcasecmp(name, node->name)) {
        return (Packet *)node;
      }  
    }
  }

  return NULL;
}


Version *
version_new(HickVer *hver, const char *name, Packet *packet)
{
  Version *ver;

  if(!(ver = version_find(&packet->vers, name))) {
    ver = MallocMem(Version);

    parr_init(&ver->info.conflicts);

    ver->info.about   = NULL;
    ver->info.require = NULL;
    ver->info.forced  = 0;
    ver->packet       = packet;

    parr_add(&packet->vers, ver);
  }

  strcpy(ver->name, name);

  if(hver) {
    ver->info.about    = hver->ver_about;
    ver->info.created  = hver->ver_created;
    ver->info.modified = hver->ver_modified;
    ver->info.expire   = hver->ver_expire;
    ver->info.forced   = 0;
    ver->override      = hver->ver_override;
    ver->isdef         = hver->ver_default;
    ver->real          = 1;

    if(!ver->info.created) {
      loaderr_add("created: %s %s: not set.\n", packet->type, packet->name,
                  ver->name);
    }

  } else {
    ver->real = 0;
  }

  return ver;
}


int
_versioncmp(const void *d1, const void *d2) {
  return strcasecmp(((Version *)d1)->name, (char *)d2);
}


void
packets_clear(NodeList *list)
{
  Node *node;

  for(node = list->head; node; node = list->head) {
    list->head = node->next;

    versions_clear(&((Packet *)node)->vers);
    pvinfo_clear(&((Packet *)node)->info);
    subgroups_clear(&((Packet *)node)->subgroups);

    if(((Packet *)node)->groups.arr) {
      FreeMem(((Packet *)node)->groups.arr);
    }

    FreeMem(node);
  }

  list->head = NULL;
  list->tail = NULL;
}


static void
subgroups_clear(NodeList *list)
{
  Node *node;

  for(node = list->head; node; node = list->head) {
    list->head = node->next;

    if(((SubGroup *)node)->down.head) {
      subgroups_clear(&((SubGroup *)node)->down);
    }

    FreeMem(node);
  }
}


static void
versions_clear(PArr *vers)
{
  size_t n;

  for(n = 0; n < vers->siz; n++) {
    pvinfo_clear(&parr_get(vers, Version, n)->info);
    FreeMem(parr_get(vers, Version, n));
  }

  if(vers->arr) {
    FreeMem(vers->arr);
  }
}


static void
pvinfo_clear(PVInfo *info)
{
  size_t n;

  if(info->about) {
    FreeMem(info->about);
  }

  for(n = 0; n < info->conflicts.siz; n++) {
    if(parr_get(&info->conflicts, Conflict, n)->vers.arr) {
      FreeMem(parr_get(&info->conflicts, Conflict, n)->vers.arr);
    }

    FreeMem(parr_get(&info->conflicts, Conflict, n));
  }

  if(info->conflicts.arr) {
    FreeMem(info->conflicts.arr);
  }

  require_clear(info->require);
}


static void
require_clear(Require *rhead) 
{
  Require *req;

  for(req = rhead; req; req = rhead) {
    rhead = req->next;

    if(req->ror) {
      require_clear(req->ror);
    }

    if(req->info.vers.arr) {
      FreeMem(req->info.vers.arr);
    }

    FreeMem(req);
  }
}
