/* HickUP 1.2 -- Graphical interface to hick.
** Copyright (C) 1998 - 2001, 2008 Daniel Bergstrom <daniel@octocode.com>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2, or (at your option)
** any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <gtk/gtkvbox.h>
#include <gtk/gtkhbox.h>
#include <gtk/gtklabel.h>
#include <gtk/gtkframe.h>
#include <gtk/gtkclist.h>
#include <gtk/gtkwindow.h>
#include <gtk/gtksignal.h>
#include <gtk/gtktooltips.h>
#include <gtk/gtknotebook.h>
#include <gtk/gtkspinbutton.h>
#include <gtk/gtkcheckbutton.h>

#include "hickmk.h"

#include "pref.h"
#include "wwrap.h"
#include "dialog.h"
#include "hickup.h"
#include "fileenable.h"

#include "common.h" /* from hick lib 1.3 */
#include "getfile.h"
#include "getpacket.h"


typedef struct {
  GtkWidget *ok_button;   /* button */

  GtkWidget *arequire;    /* checkbox */
  GtkWidget *showerrs;    /* checkbox */
  GtkWidget *hickprj;     /* checkbox */

  GtkWidget *expired;     /* checkbox */
  GtkWidget *expire;      /* checkbox */
  GtkWidget *modified;    /* checkbox */
  GtkWidget *forced;      /* checkbox */

  GtkWidget *info;        /* checkbox */
  GtkWidget *info_width;  /* spinbutton */
  GtkWidget *html;        /* checkbox */

  GtkWidget *shells;      /* clist */
  GtkWidget *all_scripts; /* checkbox */
  ArrayInfo sarr;         /* array of scripts */

  GtkWidget *menus;       /* clist */
  GtkWidget *all_menus;   /* checkbox */
  GtkWidget *submenu;     /* entry */
  ArrayInfo marr;         /* array of menus */
} PrefWidget;


#define INFO_WIDTH_MIN 50
#define INFO_WIDTH_MAX 500

#define TOKEN_SCRIPTS      300
#define TOKEN_MENUS        301
#define TOKEN_HTML         302
#define TOKEN_INFO         303
#define TOKEN_A_REQUIRE    304
#define TOKEN_R_EXPIRED    305
#define TOKEN_R_EXPIRE     306
#define TOKEN_R_MODIFIED   307
#define TOKEN_R_FORCED     308
#define TOKEN_SHOW_ERRORS  309
#define TOKEN_SUB_MENU     310

#define HICKPRJ_DISABLE_FILE ".hickprj_disable"


static FileToken _pref_tokens[] = {
  { "scripts",         TOKEN_SCRIPTS     },
  { "menus",           TOKEN_MENUS       },
  { "html_page",       TOKEN_HTML        },
  { "info_page",       TOKEN_INFO        },
  { "auto_require",    TOKEN_A_REQUIRE   },
  { "report_expired",  TOKEN_R_EXPIRED   },
  { "report_expire",   TOKEN_R_EXPIRE    },
  { "report_modified", TOKEN_R_MODIFIED  },
  { "report_forced",   TOKEN_R_FORCED    },
  { "show_errors",     TOKEN_SHOW_ERRORS },
  { "sub_menu",        TOKEN_SUB_MENU   },
  { NULL, 0 }
};

#define GET_TOKEN_NAME(type)   _pref_tokens[(type) - 300].name

#define get_active(w, value)   (value) = GTK_TOGGLE_BUTTON((w))->active;
#define set_active(w, value)   \
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON((w)), (value))


static void cb_ok(GtkWidget *w, gpointer data);
static void cb_cancel(GtkWidget *w, gpointer data);

static void save_print_flag(int type, unsigned int value, Buffert *buf);
static void save_print_array(int type, PrefArr *parr, Buffert *buf);

static void create_window(GtkWidget **window);
static GtkWidget *create_tab(const char *name, GtkWidget *notebook);
static GtkWidget *create_check(const char *name, GtkTooltips *tooltips,
                               const char *tip, GtkWidget *box);
static GtkWidget *create_frame(const char *name, GtkWidget *box);

static void create_general(GtkWidget *notebook, GtkTooltips *tooltips);
static void create_reports(GtkWidget *notebook, GtkTooltips *tooltips);
static void create_help(GtkWidget *notebook, GtkTooltips *tooltips);
static void create_scripts(GtkWidget *notebook, GtkTooltips *tooltips);
static void create_menus(GtkWidget *notebook, GtkTooltips *tooltips);

static void add_array_to_clist(ArrayInfo *arr, GtkWidget *clist);
static void set_selection(ArrayInfo *selected, ArrayInfo *items,
                          GtkWidget *clist);
static void get_selection(ArrayInfo *selected, ArrayInfo *items,
                          GtkWidget *clist);

static void cb_active(GtkWidget *w, gpointer data);


Preferences _pref;

static PrefWidget _prefw;


void
pref_init(int value, unsigned short width)
{
  hsh_array_init(&_pref.scripts.arri);
  _pref.scripts.all     = value;

  hsh_array_init(&_pref.menus.arri);
  _pref.menus.all       = value;

  _pref.info_width      = width;
  _pref.html_page       = value;
  _pref.info_page       = value;
  _pref.auto_require    = value;
  _pref.report_expired  = FALSE;
  _pref.report_expire   = FALSE;
  _pref.report_modified = FALSE;
  _pref.report_conflict = TRUE;
  _pref.report_forced   = FALSE;
  _pref.show_errors     = FALSE;

  strcpy(_pref.menu_sub, "Applications");
}


void
pref_clear(void)
{
  hsh_array_remove(&_pref.scripts.arri);
  hsh_array_remove(&_pref.menus.arri);
}


int
pref_load(void)
{
  FILE *file;
  int c, token;
  PrefArr *parr;
  struct stat sbuf;
  char fname[FNAME_LEN+1], *p;

  hsh_create_str(fname, FNAME_LEN, _pref.path, "/", _pref.pfile, NULL);

  if(!(file = fopen(fname, "r"))) {
    if(stat(fname, &sbuf) != -1) {
      return dialog_err_return_new(1, "Preferences File", "Can't open "
                                   "preferences file '%s' for reading.",fname);
    }

    return 1;
  }

  pref_init(FALSE, HICKMK_DEFAULT_WIDTH);

  while((c = _getSpaces(file)) != EOF) {
    if(c == '#') {
      _getComment(file);
      continue;
    }

    ungetc(c, file);

    if(!_getName(file, fname, FNAME_LEN)) {
      break;
    }

    switch((token = _token_lookup(fname, _pref_tokens))) {
      case TOKEN_SCRIPTS: 
      case TOKEN_MENUS: {
        parr = (token == TOKEN_SCRIPTS ? &_pref.scripts : &_pref.menus);

        c = _getSpaces(file);
        ungetc(c, file);

        if(c != '{') {
          if(!_getName(file, fname, FNAME_LEN)) {
            break;
          }

          if(!_token_case_cmp(fname, "all")) {
            parr->all = TRUE;
          }
        }

        _getList(file, &parr->arri);
        break;
      }
      case TOKEN_HTML: {
        _pref.html_page = TRUE;
        break;
      }
      case TOKEN_INFO: {
        c = _getNumber(file);

        _pref.info_page = TRUE;
        _pref.info_width = (c < INFO_WIDTH_MIN ? INFO_WIDTH_MIN :
                            (c > INFO_WIDTH_MAX ? INFO_WIDTH_MAX : c));
        break;
      }
      case TOKEN_A_REQUIRE: {
        _pref.auto_require = TRUE;
        break;
      }
      case TOKEN_R_EXPIRED: {
        _pref.report_expired = TRUE;
        break;
      }
      case TOKEN_R_EXPIRE: {
        _pref.report_expire = TRUE;
        break;
      }
      case TOKEN_R_MODIFIED: {
        _pref.report_modified = TRUE;
        break;
      }
      case TOKEN_R_FORCED: {
        _pref.report_forced = TRUE;
        break;
      }
      case TOKEN_SHOW_ERRORS: {
        _pref.show_errors = TRUE;
        break;
      }
      case TOKEN_SUB_MENU: {
        p = _getString(file);
        strncpy(_pref.menu_sub, p, MENU_LEN);
        _pref.menu_sub[MENU_LEN] = (char)0;
        FreeMem(p);
        break;
      }
    }
  }

  fclose(file);
  return 0;
}


int
pref_save(void)
{
  Buffert buf;
  char fname[FNAME_LEN+1];

  hsh_create_str(fname, FNAME_LEN, _pref.path, "/", _pref.pfile, NULL);

  if(hsh_path_create(_pref.path, NULL) == IOW_FAIL) {
    return dialog_err_return_new(1, "Preferences File", "Can't create "
                                 "path '%s' for the preferences file.",
                                 _pref.path);
  }

  if(!(buf.file = fopen(fname, "w"))) {
    return dialog_err_return_new(2, "Preferences File", "Can't open "
                                 "preferences file '%s' for writting..",fname);
  }

  buf.pos = 0;

  bprint(&buf, "# %s %s, Preferences File -- auto generated.\n\n",
         PROGRAM, VERSION_STRING);

  save_print_array(TOKEN_SCRIPTS, &_pref.scripts, &buf);
  save_print_array(TOKEN_MENUS, &_pref.menus, &buf);

  if(_pref.info_page) {
    bprint(&buf, "%s %d\n", GET_TOKEN_NAME(TOKEN_INFO), _pref.info_width);
  }

  save_print_flag(TOKEN_HTML,        _pref.html_page, &buf);
  save_print_flag(TOKEN_A_REQUIRE,   _pref.auto_require, &buf);
  save_print_flag(TOKEN_R_EXPIRED,   _pref.report_expired, &buf);
  save_print_flag(TOKEN_R_EXPIRE,    _pref.report_expire, &buf);
  save_print_flag(TOKEN_R_MODIFIED,  _pref.report_modified, &buf);
  save_print_flag(TOKEN_R_FORCED,    _pref.report_forced, &buf);
  save_print_flag(TOKEN_SHOW_ERRORS, _pref.show_errors, &buf);

  if(_pref.menu_sub[0]) {
    bprint(&buf, "%s %S\n", GET_TOKEN_NAME(TOKEN_SUB_MENU), _pref.menu_sub);
  }

  bflush(&buf);
  fclose(buf.file);
  return 0;
}


static void
save_print_flag(int type, unsigned int value, Buffert *buf)
{
  if(value) {
    bprint((buf), "%s\n", GET_TOKEN_NAME(type)); 
  }
}


static void
save_print_array(int type, PrefArr *parr, Buffert *buf)
{
  size_t n;

  bprint(buf, "%s %s{", GET_TOKEN_NAME(type), (parr->all ? "all " : ""));

  for(n = 0; n < parr->arri.siz; n++) {
    bprint(buf, " %S", parr->arri.arr[n]);
  }

  _bputs(" }\n\n", 4, buf);
}


void
cb_pref(GtkWidget *w, gpointer data)
{
  char fname[FNAME_LEN + 1];
  static GtkWidget *window = NULL;

  if(!window) {
    create_window(&window);
  }

/*   set_active(_prefw.arequire, _pref.auto_require); */

  hsh_create_str(fname, FNAME_LEN, _pref.path, "/", HICKPRJ_DISABLE_FILE, NULL);
  set_active(_prefw.hickprj, file_status(fname));

  set_active(_prefw.showerrs, _pref.show_errors);

  set_active(_prefw.expired,  _pref.report_expired);
  set_active(_prefw.expire,   _pref.report_expire);
  set_active(_prefw.modified, _pref.report_modified);
  set_active(_prefw.forced,   _pref.report_forced);

  set_active(_prefw.info,     _pref.info_page);
  set_active(_prefw.html,     _pref.html_page);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(_prefw.info_width),
                            (gfloat)_pref.info_width);

  set_active(_prefw.all_scripts, _pref.scripts.all);
  set_selection(&_pref.scripts.arri, &_prefw.sarr, _prefw.shells);

  set_active(_prefw.all_menus, _pref.menus.all);
  set_selection(&_pref.menus.arri, &_prefw.marr, _prefw.menus);
  gtk_entry_set_text(GTK_ENTRY(_prefw.submenu), _pref.menu_sub);

  gtk_widget_show(window);
  gtk_widget_grab_focus(_prefw.ok_button);
}


static void
cb_ok(GtkWidget *w, gpointer data)
{
  unsigned int fflag;
  char fname[FNAME_LEN + 1];
/*   get_active(_prefw.arequire, _pref.auto_require); */

  get_active(_prefw.hickprj, fflag);
  hsh_create_str(fname, FNAME_LEN, _pref.path, "/", HICKPRJ_DISABLE_FILE, NULL);

  if(fflag) {
    file_enable(fname);
  } else {
    file_disable(fname);
  }

  get_active(_prefw.showerrs, _pref.show_errors);


  get_active(_prefw.expired,  _pref.report_expired);
  get_active(_prefw.expire,   _pref.report_expire);
  get_active(_prefw.modified, _pref.report_modified);
  get_active(_prefw.forced,   _pref.report_forced);

  get_active(_prefw.info,     _pref.info_page);
  get_active(_prefw.html,     _pref.html_page);
  _pref.info_width =
    gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(_prefw.info_width));

  get_active(_prefw.all_scripts, _pref.scripts.all);
  hsh_array_remove(&_pref.scripts.arri);
  get_selection(&_pref.scripts.arri, &_prefw.sarr, _prefw.shells);

  get_active(_prefw.all_menus, _pref.menus.all);
  hsh_array_remove(&_pref.menus.arri);
  get_selection(&_pref.menus.arri, &_prefw.marr, _prefw.menus);
  strncpy(_pref.menu_sub, gtk_entry_get_text(GTK_ENTRY(_prefw.submenu)),
          MENU_LEN);
  _pref.menu_sub[MENU_LEN] = (char)0;

  pref_save();

  gtk_widget_hide(GTK_WIDGET(data));
}


static void
cb_cancel(GtkWidget *w, gpointer data)
{
  gtk_widget_hide(GTK_WIDGET(data));
}


static void
create_window(GtkWidget **window)
{
  GtkTooltips *tooltips;
  GtkWidget *vbox, *notebook;

  *window = gtk_window_new(GTK_WINDOW_DIALOG);
  gtk_container_set_border_width(GTK_CONTAINER(*window), 5);
  gtk_window_set_title(GTK_WINDOW(*window), "HickUP Preferences");
  gtk_window_set_policy(GTK_WINDOW(*window), TRUE, TRUE, TRUE);

  tooltips = gtk_tooltips_new();

  gtk_signal_connect(GTK_OBJECT(*window), "destroy",
                     GTK_SIGNAL_FUNC(gtk_widget_destroyed), window);
  gtk_signal_connect(GTK_OBJECT(*window), "delete_event",
                     GTK_SIGNAL_FUNC(gtk_widget_hide), NULL);

  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(*window), vbox);

  notebook = gtk_notebook_new();
  gtk_notebook_set_tab_hborder(GTK_NOTEBOOK(notebook), 5);
  gtk_notebook_set_homogeneous_tabs(GTK_NOTEBOOK(notebook), TRUE);
  gtk_box_pack_start(GTK_BOX(vbox), notebook, TRUE, TRUE, 0);

  create_general(notebook, tooltips);
  create_reports(notebook, tooltips);
  create_help(notebook, tooltips);
  create_scripts(notebook, tooltips);
  create_menus(notebook, tooltips);

  gtk_widget_show(notebook);

  ww_buttons(vbox, WW_HBBOX, GTK_BUTTONBOX_END, 5,
             "OK", BUTTON_ISDEF, &_prefw.ok_button, cb_ok, *window,
             "Cancel", BUTTON_CANDEF, NULL, cb_cancel, *window,
             NULL);

  gtk_widget_show(vbox);
}


static GtkWidget *
create_tab(const char *name, GtkWidget *notebook)
{
  GtkWidget *vbox, *w;

  w = gtk_label_new(name);
  vbox = gtk_vbox_new(FALSE, 3);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), vbox, w);
  gtk_widget_show(vbox);

  return vbox;
}


static GtkWidget *
create_check(const char *name, GtkTooltips *tooltips, const char *tip,
             GtkWidget *box)
{
  GtkWidget *w;

  w = gtk_check_button_new_with_label(name);
  gtk_box_pack_start(GTK_BOX(box), w, FALSE, TRUE, 0);

  if(tooltips) {
    gtk_tooltips_set_tip(tooltips, w, tip, name);
  }

  gtk_widget_show(w);
  return w;
}


static GtkWidget *
create_frame(const char *name, GtkWidget *box)
{
  GtkWidget *frame, *vbox;

  frame = gtk_frame_new(name);
  gtk_box_pack_start(GTK_BOX(box), frame, TRUE, TRUE, 0);
  gtk_widget_show(frame);

  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
  gtk_container_add(GTK_CONTAINER(frame), vbox);
  gtk_widget_show(vbox);

  return vbox;
}


static void
create_general(GtkWidget *notebook, GtkTooltips *tooltips)
{
  GtkWidget *vbox;

  vbox = create_tab("General", notebook);

  _prefw.showerrs = create_check("Show Load Errors", tooltips, "When loading "
                                 "the packet information, show packet errors.",
                                 vbox);
  _prefw.hickprj = create_check("Disable Project Chooser", tooltips,
                                "Disable Hick Project Chooser at login", vbox);
}


static void
create_reports(GtkWidget *notebook, GtkTooltips *tooltips)
{
  GtkWidget *vbox, *vvbox;

  vbox = create_tab("Reports", notebook);
  vvbox = create_frame(" Load Report ", vbox);

  _prefw.expired = create_check("Expired or removed packets", tooltips,
                                "If checked will it show removed or expired "
                                "packets.", vvbox);
  _prefw.expire = create_check("Expiering packets", tooltips, "Reports "
                               "packets which will expire.", vvbox);
  _prefw.modified = create_check("Modified packets", tooltips, "Lets you "
                                 "know when packet information have been "
                                 "modified.", vvbox);

  vvbox = create_frame(" Update Report ", vbox);

  _prefw.forced = create_check("Forced installed", tooltips, "When running "
                               "update will it also show the update report"
                               "when forced packets have been installed.",
                               vvbox);
}


static void
create_help(GtkWidget *notebook, GtkTooltips *tooltips)
{
  GtkAdjustment *adj;
  GtkWidget *vbox, *hbox, *hhbox, *w;

  vbox = create_tab("Help", notebook);

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, TRUE, 0);
  gtk_widget_show(hbox);

  _prefw.info = create_check("GNU Info", tooltips, "Create a personal GNU "
                             "info menu over help pages.", hbox);

  hhbox = gtk_hbox_new(FALSE, 5);
  gtk_box_pack_end(GTK_BOX(hbox), hhbox, FALSE, FALSE, 0);
  gtk_widget_show(hhbox);

  w = gtk_label_new("Page Width");
  gtk_box_pack_start(GTK_BOX(hhbox), w, FALSE, FALSE, 0);
  gtk_widget_show(w);

  adj = GTK_ADJUSTMENT(gtk_adjustment_new((gfloat)HICKMK_DEFAULT_WIDTH,
                                          (gfloat)INFO_WIDTH_MIN,
                                          (gfloat)INFO_WIDTH_MAX, 1.0, 100.0,
                                          0.0));

  _prefw.info_width = gtk_spin_button_new(adj, 0, 0);
  gtk_tooltips_set_tip(tooltips, _prefw.info_width, "Set the width for the "
                       "generation of info page.", "info width");
  gtk_spin_button_set_wrap(GTK_SPIN_BUTTON(_prefw.info_width), TRUE);
  gtk_widget_set_usize(_prefw.info_width, 45, -1);
  gtk_box_pack_end(GTK_BOX(hhbox), _prefw.info_width, FALSE, TRUE, 0);
  gtk_widget_show(_prefw.info_width);

  _prefw.html = create_check("HTML Index", tooltips, "Create a personal html "
                             "index over help pages.", vbox);
}


static void
create_scripts(GtkWidget *notebook, GtkTooltips *tooltips)
{
  gchar *titles[1];
  GtkWidget *vbox, *hbox, *w;

  vbox = create_tab("Scripts", notebook);

  _prefw.all_scripts = create_check("All Scripts", tooltips,
                                    "Select all scripts.", vbox);
  gtk_signal_connect(GTK_OBJECT(_prefw.all_scripts), "toggled",
                     GTK_SIGNAL_FUNC(cb_active), (gpointer)1);

  hbox = gtk_hbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);
  gtk_widget_show(hbox);

  titles[0] = "Scripts";
  w = ww_clist(&_prefw.shells, 1, titles, NULL);
  gtk_clist_set_selection_mode(GTK_CLIST(_prefw.shells),
                               GTK_SELECTION_MULTIPLE);
  gtk_box_pack_start(GTK_BOX(hbox), w, TRUE, TRUE, 0);

  _prefw.sarr.arr = hickmk_scripts_get(SCRIPT_PATH, &_prefw.sarr.siz);
  _prefw.sarr.alloc = _prefw.sarr.siz;

  add_array_to_clist(&_prefw.sarr, _prefw.shells);
}


static void
create_menus(GtkWidget *notebook, GtkTooltips *tooltips)
{
  gchar *titles[1];
  GtkWidget *vbox, *w, *hbox;

  vbox = create_tab("Menus", notebook);

  _prefw.all_menus = create_check("All Window Managers", tooltips,
                                  "Select all window managers.", vbox);
  gtk_signal_connect(GTK_OBJECT(_prefw.all_menus), "toggled",
                     GTK_SIGNAL_FUNC(cb_active), (gpointer)0);

  titles[0] = "Window Managers";
  w = ww_clist(&_prefw.menus, 1, titles, NULL);
  gtk_clist_set_selection_mode(GTK_CLIST(_prefw.menus),GTK_SELECTION_MULTIPLE);
  gtk_box_pack_start(GTK_BOX(vbox), w, TRUE, TRUE, 0);

  _prefw.marr.arr = hickmk_menus_get(MENUS_PATH, &_prefw.marr.siz);
  _prefw.marr.alloc = _prefw.marr.siz;

  add_array_to_clist(&_prefw.marr, _prefw.menus);

  hbox = gtk_hbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, TRUE, 0);
  gtk_widget_show(hbox);

  w = gtk_label_new("Menu Name");
  gtk_box_pack_start(GTK_BOX(hbox), w, FALSE, TRUE, 0);
  gtk_widget_show(w);

  _prefw.submenu = gtk_entry_new_with_max_length(MENU_LEN);
  gtk_box_pack_start(GTK_BOX(hbox), _prefw.submenu, TRUE, TRUE, 0);
  gtk_widget_show(_prefw.submenu);
}


static void
add_array_to_clist(ArrayInfo *arr, GtkWidget *clist)
{
  size_t n;
  gchar *text[1];

  gtk_clist_freeze(GTK_CLIST(clist));

  for(n = 0; n < arr->siz; n++) {
    text[0] = arr->arr[n];
    gtk_clist_insert(GTK_CLIST(clist), n, text);
  }

  gtk_clist_thaw(GTK_CLIST(clist)); 
}


static void
set_selection(ArrayInfo *selected, ArrayInfo *items, GtkWidget *clist)
{
  size_t n;

  gtk_clist_freeze(GTK_CLIST(clist));
  gtk_clist_unselect_all(GTK_CLIST(clist));

  for(n = 0; n < items->siz; n++) {
    if(hsh_array_find(items->arr[n], selected)) {
      gtk_clist_select_row(GTK_CLIST(clist), n, 0);
    } else {
      gtk_clist_unselect_row(GTK_CLIST(clist), n, 0);
    }
  }

  gtk_clist_thaw(GTK_CLIST(clist)); 
}


static void
get_selection(ArrayInfo *selected, ArrayInfo *items, GtkWidget *clist)
{
  GList *list;

  for(list = GTK_CLIST(clist)->selection; list; list = list->next) {
    hsh_array_add(items->arr[GPOINTER_TO_INT(list->data)], 1, selected);
  }
}


static void
cb_active(GtkWidget *w, gpointer data)
{
  if(data == 0) { /* menu */
    gtk_widget_set_sensitive(_prefw.menus, !GTK_TOGGLE_BUTTON(w)->active);
  } else { /* script */
    gtk_widget_set_sensitive(_prefw.shells, !GTK_TOGGLE_BUTTON(w)->active);
  }
}
