/* HickUP 1.2 -- Graphical interface to hick.
** Copyright (C) 1998 - 2001, 2008 Daniel Bergstrom <daniel@octocode.com>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2, or (at your option)
** any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "gui.h"
#include "refs.h"


RefType *
refs_lookup_type(const char *name, NodeList *list)
{
  RefType *rt;

  if((rt = (RefType *)ntlist_find(name, list))) {
    return rt;
  }

  rt = MallocMem(RefType);
  ntlist_init((NTList *)rt);
  rt->ntree.node.flag = NT_TYPE;
  rt->ntree.node.name = (char *)name;

  node_insert(&rt->ntree.node, list);
  return rt;
}


RefGroup *
refs_lookup_group(char *name, NTList *parent)
{
  RefGroup *rg;

  if((rg = (RefGroup *)ntlist_find(name, &parent->nlist))) {
    return rg;
  }

  rg = MallocMem(RefGroup);
  ntlist_init((NTList *)rg);
  rg->ntree.node.flag = NT_GROUP;
  rg->ntree.node.name = name;
  rg->ntree.parent = &parent->ntree;

  node_insert(&rg->ntree.node, &parent->nlist);
  return rg;
}


RefPacket *
refs_packet_new(char *name, AvailPacket *packet, NTList *parent)
{
  RefPacket *rp;

  rp = MallocMem(RefPacket);
  ntlist_init(&rp->ntlist);
  rp->packet = packet;
  rp->hview = NULL;
  rp->ntlist.ntree.node.flag = NT_PACKET;
  rp->ntlist.ntree.node.name = name;
  rp->ntlist.ntree.parent = &parent->ntree;

  parr_add(&packet->refs, &rp->ntlist);
  node_insert(&rp->ntlist.ntree.node, &parent->nlist);
  return rp;
}


RefVersion *
refs_version_new(char *name, AvailVersion *ver, RefPacket *rp)
{
  RefVersion *rv;

  rv = MallocMem(RefVersion);
  nodetree_init(&rv->ntree);
  rv->version = ver;
  rv->ntree.node.flag = NT_VERSION;
  rv->ntree.node.name = name;
  rv->ntree.parent = &rp->ntlist.ntree;

  hsh_node_append(&rv->ntree.node, &rp->ntlist.nlist);
  return rv;
}


RefData *
refs_data_new(char *name, AvailPacket *packet, AvailVersion *version, 
              NodeList *list)
{
  RefData *rd;

  rd = MallocMem(RefData);
  ntlist_init(&rd->ntlist);
  rd->packet = packet;
  rd->version = version;
  rd->hview = NULL;
  rd->row = -1;
  rd->ntlist.ntree.node.flag = NT_EMPTY;
  rd->ntlist.ntree.node.name = name;

  hsh_node_append(&rd->ntlist.ntree.node, list);
  return rd;
}


void
refs_free(NodeList *list)
{
  size_t n;
  Node *node;

  for(node = list->head; node; node = list->head) {
    list->head = node->next;

    if(node->flag == NT_EMPTY) { /* RefData */
      if(node->name) {
        FreeMem(node->name);
      }

      if(((RefData *)node)->packet) {
        n = parr_find(&((RefData *)node)->packet->refs, (NTList *)node, 0);

        if(n) {
          parr_remove(&((RefData *)node)->packet->refs, n - 1);
        }
      }
    } else {
      if(node->flag == NT_PACKET) {
        if(((RefPacket *)node)->packet) {
          n = parr_find(&((RefPacket *)node)->packet->refs, (NTList *)node, 0);

          if(n) {
            parr_remove(&((RefPacket *)node)->packet->refs, n - 1);
          }
        }
      }
    }

    if((node->flag != NT_VERSION) && ((NTList *)node)->nlist.head) {
      refs_free(&((NTList *)node)->nlist);
    }

    FreeMem(node);
  }

  list->head = list->tail = NULL;
}


void
refs_display_tree(NodeList *list, HView *hview, GtkCTreeNode *parent,
                  int expandall)
{
  Node *node;
  GPixmap *pixi;
  NodeTree *ntree;
  gchar *text[2];

  if(!list->head) {
    return;
  }

  for(text[0] = "", node = list->head; node; node = node->next) {
    ntree = (NodeTree *)node;
    text[1] = node->name;

    if((node->flag != NT_EMPTY) && _gmain.pixmaps.pixmaps[node->flag - 1]) {
      ntree->ctnode =
        gtk_ctree_insert_node(hview_ctree_get(hview), parent, NULL, text, 3,
                              _gmain.pixmaps.pixmaps[node->flag-1]->close.pix,
                              _gmain.pixmaps.pixmaps[node->flag-1]->close.mask,
                              _gmain.pixmaps.pixmaps[node->flag-1]->open.pix,
                              _gmain.pixmaps.pixmaps[node->flag-1]->open.mask,
                              (node->flag == NT_VERSION), expandall);
    } else {
      ntree->ctnode = gtk_ctree_insert_node(hview_ctree_get(hview), parent, 
                                            NULL, text, 0, NULL, NULL, NULL,
                                            NULL, (node->flag == NT_VERSION), 
                                            expandall);
    }

    pixi = NULL;

    switch(node->flag) {
      case NT_EMPTY: { /* RefData */
        ((RefData *)node)->hview = hview;

        if(((RefData *)node)->packet && hview->pixfunc) {
          pixi = hview_pixmap_get(hview, ((RefData *)node)->packet,
                                  ((RefData *)node)->version, NULL);
        }
        break;
      }
      case NT_PACKET: { /* RefPacket */
        ((RefPacket *)node)->hview = hview;

        if(((RefPacket *)node)->packet && hview->pixfunc) {
          pixi = hview_pixmap_get(hview, ((RefPacket *)node)->packet, NULL,
                                  NULL);
        }
        break;
      }
      case NT_VERSION: { /* RefVersion */
        if(((RefVersion *)node)->version && hview->pixfunc) {
          pixi = hview_pixmap_get(hview, ((RefVersion *)node)->version->packet,
                                  ((RefVersion *)node)->version, NULL);
        }
      }
    }

    if(pixi) {
      gtk_ctree_node_set_pixmap(hview_ctree_get(hview), ntree->ctnode, 
                                0, pixi->pix, pixi->mask);
    }

    if(node->flag != NT_VERSION) {
      refs_display_tree(&((NTList *)node)->nlist, hview, ntree->ctnode,
                        expandall);
    }

    if((node->flag == NT_TYPE) || (node->flag == NT_GROUP)) {
      gtk_ctree_sort_node(hview_ctree_get(hview), ntree->ctnode);
    }

    gtk_ctree_node_set_row_data(hview_ctree_get(hview), ntree->ctnode, ntree);
  }
}
