/* HickUP 1.2 -- Graphical interface to hick.
** Copyright (C) 1998 - 2001, 2008 Daniel Bergstrom <daniel@octocode.com>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2, or (at your option)
** any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <stdarg.h>

#include <gtk/gtkhbox.h>
#include <gtk/gtkvbox.h>
#include <gtk/gtkclist.h>
#include <gtk/gtklabel.h>
#include <gtk/gtkpixmap.h>
#include <gtk/gtknotebook.h>

#include "memd.h"
#include "wwrap.h"
#include "report.h"

static void clist_tab_create(Report *r, ReportTab *rt, GtkWidget *vbox);
static void clist_tab_clear(Report *r, ReportTab *rt);


Report *
report_new(size_t tabs, GdkPixmap *pix, GdkBitmap *mask)
{
  Report *r;

  r = MallocMem(Report);
  r->tabs = MallocArr(ReportTab *, tabs);
  r->tabSize = tabs;
  r->pix = pix;
  r->mask = mask;

  return r;
}


void
report_tab_set(Report *r, size_t tabnr, const char *name, ReportCreate create,
               ReportClear clear, void *arg)
{
  ReportTab *rt;

  rt = MallocMem(ReportTab);
  rt->name = gtk_label_new(name);
  rt->pixmap = gtk_pixmap_new(r->pix, r->mask);
  rt->got_items = FALSE;
  rt->create = create;
  rt->clear = clear;
  rt->type = RTYPE_NONE;
  rt->arg = arg;

  r->tabs[tabnr] = (ReportTab *)rt;
}


void
report_tab_set_clist(Report *r, size_t tabnr, const char *name,
                     const char *help, ReportSignal func, void *arg, 
                     size_t columns)
{
  ReportTabCList *rtc;

  rtc = MallocMem(ReportTabCList);
  rtc->rtab.name = gtk_label_new(name);
  rtc->rtab.pixmap = gtk_pixmap_new(r->pix, r->mask);
  rtc->rtab.got_items = FALSE;
  rtc->rtab.create = clist_tab_create;
  rtc->rtab.clear = clist_tab_clear;
  rtc->rtab.type = RTYPE_CLIST;
  rtc->rtab.arg = arg;

  rtc->columns.length = MallocArr(gint, columns);
  rtc->columns.titles = MallocArr(gchar *, columns);
  rtc->colSize = columns;
  rtc->help = gtk_label_new(help);
  rtc->sig_func = func;

  r->tabs[tabnr] = &rtc->rtab;
}


void
report_show_pixmaps(Report *r)
{
  size_t n;

  if(r->pix && r->mask) {
    for(n = 0; n < r->tabSize; n++) {
      if(r->tabs[n]->got_items) {
        gtk_widget_show(GTK_WIDGET(r->tabs[n]->pixmap));
      }
    }
  }
}


int
report_display(Report *r, unsigned int mask)
{
  size_t n;

  for(n = 0; n < r->tabSize; n++) {
    if((mask & (1 << n)) && r->tabs[n]->got_items) {
      return n;
    }
  }

  return -1;
}


void
report_tab_clist_col_set(Report *r, size_t tabnr, size_t column,
                         const gchar *title, gint length)
{
  ((ReportTabCList *)r->tabs[tabnr])->columns.titles[column] = (gchar *)title;
  ((ReportTabCList *)r->tabs[tabnr])->columns.length[column] = length;
}


void
report_clear(Report *r)
{
  size_t n;

  for(n = 0; n < r->tabSize; n++) {
    r->tabs[n]->got_items = FALSE;
    gtk_widget_hide(GTK_WIDGET(r->tabs[n]->pixmap));

    if(r->tabs[n]->clear) {
     (r->tabs[n]->clear)(r, r->tabs[n]);
    }
  }
}


GtkWidget *
report_create(Report *r)
{
  size_t n;
  ReportTab *rt;
  GtkWidget *vbox, *hbox;

  r->notebook = gtk_notebook_new();
  gtk_notebook_set_tab_pos(GTK_NOTEBOOK(r->notebook), GTK_POS_TOP);
  gtk_notebook_set_tab_hborder(GTK_NOTEBOOK(r->notebook), 5);
  gtk_notebook_set_homogeneous_tabs(GTK_NOTEBOOK(r->notebook), 
                                    TRUE);

  for(n = 0; n < r->tabSize; n++) {
    rt = r->tabs[n];

    hbox = gtk_hbox_new(FALSE, 0); /* for name and pixmap in tab. */
    gtk_box_pack_start(GTK_BOX(hbox), rt->name,   FALSE, TRUE, 0);
    gtk_box_pack_start(GTK_BOX(hbox), rt->pixmap, FALSE, TRUE, 0);
    gtk_widget_show(rt->name);

    vbox = gtk_vbox_new(FALSE, 5); /* area in tab */
    gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);

    gtk_notebook_append_page(GTK_NOTEBOOK(r->notebook), vbox, hbox);

    if(r->tabs[n]->create) {
      (r->tabs[n]->create)(r, r->tabs[n], vbox);
    }

    gtk_widget_show(vbox);
  }

  return r->notebook;
}


gint
report_add_clist_item(Report *r, size_t tabnr, ...)
{
  size_t n;
  va_list pvar;
  ReportTabCList *rtc = (ReportTabCList *)r->tabs[tabnr];

  va_start(pvar, tabnr);

  for(n = 0; n < rtc->colSize; n++) {
    rtc->columns.titles[n] = va_arg(pvar, gchar *);
  }

  rtc->rtab.got_items = TRUE;
  return gtk_clist_append(GTK_CLIST(rtc->clist), rtc->columns.titles);
}


static void
clist_tab_create(Report *r, ReportTab *rt, GtkWidget *vbox)
{
  GtkWidget *w;
  ReportTabCList *rtc = (ReportTabCList *)rt;

  ww_frame(NULL, NULL, NULL, GTK_WIDGET(rtc->help), vbox);

  w = ww_clist(&rtc->clist, rtc->colSize, rtc->columns.titles,
               rtc->columns.length);
  gtk_box_pack_start(GTK_BOX(vbox), w, TRUE, TRUE, 0);

  gtk_signal_connect(GTK_OBJECT(rtc->clist), "select_row",
                     GTK_SIGNAL_FUNC(rtc->sig_func), rt);
}


static void
clist_tab_clear(Report *r, ReportTab *rt)
{
  gtk_clist_freeze(GTK_CLIST(((ReportTabCList *)rt)->clist));
  gtk_clist_clear(GTK_CLIST(((ReportTabCList *)rt)->clist));
  gtk_clist_thaw(GTK_CLIST(((ReportTabCList *)rt)->clist));
}
