/* HickUP 1.2 -- Graphical interface to hick.
** Copyright (C) 1998 - 2001, 2008 Daniel Bergstrom <daniel@octocode.com>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2, or (at your option)
** any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "common.h"             /* from hick lib 1.3 */
#include "getfile.h"
#include "getpacket.h"

#include "hickup.h"
#include "specvars.h"


#define TOKEN_ENABLE  300

static FileToken _spec_tokens[] = {
  { "enable",  TOKEN_ENABLE },
  { "force",   TOKEN_FORCE }, /* TOKEN_FORCE defined in getpacket.h */
  { NULL, 0 }
};

static void spec_free_list(NodeList *list);

static void print_list(const char *name, const char *type, NodeList *list, 
                       int mustver, Buffert *buf);
static void print_versions(PArr *parr, Buffert *buf);

static SpecType *get_spectype(const char *type, SpecVars *svars);
static int get_specvar(const char *filename, FILE *file, 
                       const char *type, int enable, SpecVars *svars);


int
spec_load(const char *filename, SpecVars *result, int print_errors)
{
  FILE *file;
  char buf[64], *type;
  int c, token, enable;

  spec_vars_init(result);

  if(!(file = fopen(filename, "r"))) {
    return 0;
  }

  while((c = _getSpaces(file)) != EOF) {
    if(c == '#') {
      _getComment(file);
      continue;
    }

    ungetc(c, file);

    if(!_getName(file, buf, 64)) {
      break;
    }

    token = _token_lookup(buf, _spec_tokens);
    
    if((token != TOKEN_ENABLE) && (token != TOKEN_FORCE)) {
      printf("%s: Unknown token '%s'\n", filename, buf);
      continue;
    }

    enable = (token == TOKEN_ENABLE ? SPEC_ENABLE : SPEC_FORCE);
    type = _getString(file);

    if((c = _getSpaces(file)) == '{') { /* array of items */
      do {
        if((c = _getSpaces(file)) == EOF) {
          printf("%s: Unexpected enf of file.\n", filename);
          break;
        }

        if(c == '#') {
          _getComment(file);
          continue;
        }

        if(c == '}') {
          break;
        }

        ungetc(c, file);

        get_specvar(filename, file, type, enable, result);
      } while(!feof(file));
    } else { /* one item */
      ungetc(c, file);
      get_specvar(filename, file, type, enable, result);
    }
  }

  fclose(file);
  return 1;
}


int
spec_save(const char *filename, SpecVars *svars)
{
  Node *node;
  Buffert buf;

  if(!(buf.file = fopen(filename, "w"))) {
    return 0;
  }

  buf.pos = 0;

  bprint(&buf, "# %s %s, Project Specification File -- auto generated.\n\n", 
         PROGRAM, VERSION_STRING);

  for(node = svars->types.head; node; node = node->next) {
    print_list("enable", node->name, &((SpecType *)node)->enable, 0, &buf);
    print_list("force", node->name, &((SpecType *)node)->force, 1, &buf);
  }

  bflush(&buf);
  fclose(buf.file);
  return 1;
}


SpecVar *
spec_find(const char *type, const char *name, int enable, SpecVars *svars)
{
  SpecType *stype;

  if((stype = spec_find_type(type, svars))) {
    return stype_find(name, enable, stype);
  }

  return NULL;
}


SpecVar *
stype_find(const char *name, int enable, SpecType *stype)
{
  Node *node;

  if(stype) {
    node = (enable ? stype->enable.head : stype->force.head);

    for( ; node; node = node->next) {
      if(!strcasecmp(name, node->name)) {
        return (SpecVar *)node;
      }
    }
  }

  return NULL;
}


int
spec_add(const char *type, const char *name, PArr *versions, int enable, 
         SpecVars *svars)
{
  size_t n;
  char *p, *np;
  SpecVar *svar;
  SpecType *stype;

  if(!enable && !versions->siz) {
    return 0;
  }

  stype = get_spectype(type, svars);

  svar = MallocMem(SpecVar);
  hsh_node_prepend(&svar->node, (enable ? &stype->enable : &stype->force));
  parr_init(&svar->versions);
  strncpy(svar->name, name, PNAME_LEN);
  svar->name[PNAME_LEN] = (char)0;

  if(versions) {
    for(n = 0; n < versions->siz; n++) {
      p = parr_get(versions, char, n);

      np = MallocArr(char, strlen(p) + 1);
      strcpy(np, p);

      parr_add(&svar->versions, np);
    }
  }

  return 1;
}


static void
print_list(const char *name, const char *type, NodeList *list, int mustver, 
           Buffert *buf)
{
  Node *node;

  if(list->head) {
    bprint(buf, "%s %S {\n", name, type);

    for(node = list->head; node; node = node->next) {
      if(mustver && !((SpecVar *)node)->versions.siz) {
        continue;
      }

      bprint(buf, "  %S", ((SpecVar *)node)->name);
      print_versions(&((SpecVar *)node)->versions, buf);
    }

    _bputs("}\n\n", 2, buf);
  }
}


static void
print_versions(PArr *parr, Buffert *buf)
{
  size_t n;

  switch(parr->siz) {
    case 0: {
      _bputs(" \"\"\n", 4, buf);
      break;
    }
    case 1: {
      bprint(buf, " %S\n", parr_get(parr, char, 0));
      break;
    }
    default: {
      _bputs(" {", 2, buf);

      for(n = 0; n < parr->siz; n++) {
        bprint(buf, " %S", parr_get(parr, char, n));
      }

      _bputs(" }\n", 3, buf);      
      break;
    }
  }
}


SpecVars *
spec_free(SpecVars *svars)
{
  Node *node;

  for(node = svars->types.head; node; node = svars->types.head) {
    svars->types.head = node->next;

    spec_free_list(&((SpecType *)node)->enable);
    spec_free_list(&((SpecType *)node)->force);

    FreeMem(node->name);
    FreeMem(node);
  }

  return svars;
}


int
specvar_is_version(SpecVar *svar, const char *version)
{
  return (svar && (!svar->versions.siz ||
                   parr_find_value(&svar->versions, char, version,
                                   parr_strcasecmp)));
}


static void
spec_free_list(NodeList *list)
{
  Node *node;

  for(node = list->head; node; node = list->head) {
    list->head = node->next;

    parr_clear(&((SpecVar *)node)->versions);
    FreeMem(node);
  }

  hsh_nodelist_init(list);
}


static SpecType *
get_spectype(const char *type, SpecVars *svars)
{
  SpecType *stype;

  if(!(stype = spec_find_type(type, svars))) {
    stype = MallocMem(SpecType);
    stype->node.name = MallocArr(char, strlen(type) + 1);
    strcpy(stype->node.name, type);
    hsh_nodelist_init(&stype->enable);
    hsh_nodelist_init(&stype->force);

    hsh_node_prepend(&stype->node, &svars->types);
  }

  return stype;
}


static int 
get_specvar(const char *filename, FILE *file, const char *type, int enable, 
            SpecVars *svars)
{
  size_t n;
  char *name;
  SpecVar *svar;
  SpecType *stype;
  ArrayInfo arri;

  if(!(name = _getString(file))) {
    return 0;
  }

  if(!_getList(file, &arri)) {
    FreeMem(name);
    return 0;
  }

  if(!enable && (arri.siz == 0)) {
    printf("%s: missing forced version on %s\n", filename, name);
    FreeMem(name);
    return 0;
  }

  stype = get_spectype(type, svars);

  if(stype_find(name, enable, stype)) {
    printf("%s: multiple declarations on %s\n", filename, name);
    FreeMem(name);
    return 0;
  }

  svar = MallocMem(SpecVar);
  strncpy(svar->name, name, PNAME_LEN);
  svar->name[PNAME_LEN] = (char)0;
  svar->node.name = svar->name;
  FreeMem(name);

  parr_init(&svar->versions);

  for(n = 0; n < arri.siz; n++) {
    parr_add(&svar->versions, arri.arr[n]);
  }

  if(arri.arr) {
    FreeMem(arri.arr);
  }

  hsh_node_prepend(&svar->node, (enable ? &stype->enable : &stype->force));
  return 1;
}
