/* HickUP 1.2 -- Graphical interface to hick.
** Copyright (C) 1998 - 2001, 2008 Daniel Bergstrom <daniel@octocode.com>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2, or (at your option)
** any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <stdarg.h>
#include <string.h>
#include <gtk/gtk.h>
#include "wwrap.h"
#include "dialog.h"
#include "hickup.h"

static gint clist_compare(GtkCList *clist, gconstpointer ptr1,
                          gconstpointer ptr2);

static GtkWidget *create_dialog(GUI_Dialog *gdialog, const char *title,
                                va_list pa);
static GtkWidget *create_buttons(GtkWidget *box, int type,
                                 GtkButtonBoxStyle layout, gint width,
                                 GtkSignalFunc stdfunc, gpointer data,
                                 va_list pa);

static void cb_hide(GtkWidget *w, gpointer data);
static void cb_exit(GtkWidget *w, gint data);
static void cb_return(GtkWidget *w, gpointer data);


GtkWidget *
ww_frame(GtkWidget **label, const char *title, const char *text,
         GtkWidget *textwidget, GtkWidget *box)
{
  GtkWidget *frame, *hbox;

  frame = gtk_frame_new(title);
  gtk_box_pack_start(GTK_BOX(box), frame, FALSE, TRUE, 0);
  gtk_widget_show(frame);

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(hbox), 5);
  gtk_container_add(GTK_CONTAINER(frame), hbox);

  if(!textwidget) {
    textwidget = gtk_label_new(text);
  }

  if(label) {
    *label = textwidget;
  }

  gtk_box_pack_start(GTK_BOX(hbox), textwidget, TRUE, FALSE, 0);
  gtk_label_set_line_wrap(GTK_LABEL(textwidget), TRUE);
  gtk_label_set_justify(GTK_LABEL(textwidget), GTK_JUSTIFY_FILL);
  gtk_widget_show(hbox);
  gtk_widget_show(textwidget);
  return frame;
}


GtkWidget *
ww_buttons(GtkWidget *box, int type, GtkButtonBoxStyle layout,
           gint width, ...)
{
  va_list pa;
  GtkWidget *w;

  va_start(pa, width);
  w = create_buttons(box, type, layout, width, NULL, NULL, pa);
  va_end(pa);

  return w;
}


GtkWidget *
ww_text_area(GtkWidget **text, int border)
{
  GtkWidget *table, *scroll;

  table = gtk_table_new(1, 2, FALSE);
  gtk_container_set_border_width(GTK_CONTAINER(table), border);

  *text = gtk_text_new(NULL, NULL);
  gtk_table_attach(GTK_TABLE(table), *text, 0, 1, 0, 1,
                   (GtkAttachOptions)(GTK_FILL | GTK_EXPAND | GTK_SHRINK),
                   (GtkAttachOptions)(GTK_FILL | GTK_EXPAND | GTK_SHRINK),
                   0, 0);

  gtk_text_set_word_wrap(GTK_TEXT(*text), TRUE);

  scroll = gtk_vscrollbar_new(GTK_TEXT(*text)->vadj);
  gtk_table_attach(GTK_TABLE(table), scroll, 1, 2, 0, 1,
                   GTK_FILL, 
                   (GtkAttachOptions)(GTK_EXPAND | GTK_FILL | GTK_SHRINK), 
                   0, 0);

  gtk_widget_show(scroll);
  gtk_widget_show(*text);
  gtk_widget_show(table);
  return table;
}


void
ww_text_clear(GtkWidget *widget)
{
  guint len;

  if((len = gtk_text_get_length(GTK_TEXT(widget)))) {
    gtk_text_freeze(GTK_TEXT(widget));
    gtk_text_backward_delete(GTK_TEXT(widget), len);
    gtk_text_thaw(GTK_TEXT(widget));
  }
}


GtkWidget *
ww_ctree(GtkWidget **ctree, gint columns, gint tree_column,
         gchar **titles, gint *cwidth)
{
  gint n;
  GtkWidget *vbox, *swin;

  vbox = gtk_vbox_new(FALSE, 5);

  swin = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(swin),
                                 GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_box_pack_start(GTK_BOX(vbox), swin, TRUE, TRUE, 0);

  *ctree = (titles ? gtk_ctree_new_with_titles(columns, tree_column, titles) :
            gtk_ctree_new(columns, tree_column));

  gtk_container_add(GTK_CONTAINER(swin), *ctree);

  gtk_clist_column_titles_passive(GTK_CLIST(*ctree));
  gtk_clist_set_compare_func(GTK_CLIST(*ctree), clist_compare);
  gtk_clist_set_selection_mode(GTK_CLIST(*ctree), GTK_SELECTION_BROWSE);
  gtk_ctree_set_line_style(GTK_CTREE(*ctree), GTK_CTREE_LINES_DOTTED);
  gtk_ctree_set_expander_style(GTK_CTREE(*ctree), GTK_CTREE_EXPANDER_SQUARE);

  for(n = 0; n < columns; n++) {
    if(cwidth && (cwidth[n] != -1)) {
      gtk_clist_set_column_min_width(GTK_CLIST(*ctree), n, cwidth[n]);
    }

    gtk_clist_set_column_auto_resize(GTK_CLIST(*ctree), n, TRUE);
  }

  gtk_widget_show(*ctree);
  gtk_widget_show(swin);
  gtk_widget_show(vbox);
  return vbox;
}


GtkWidget *
ww_clist(GtkWidget **clist, gint columns, gchar **titles, gint *cwidth)
{
  gint n;
  GtkWidget *swin;

  swin = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(swin),
                                 GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

  *clist = gtk_clist_new_with_titles(columns, titles);
  gtk_container_add(GTK_CONTAINER(swin), *clist);
  gtk_clist_set_selection_mode(GTK_CLIST(*clist), GTK_SELECTION_BROWSE);
  gtk_clist_column_titles_passive(GTK_CLIST(*clist));

  for(n = 0; n < columns; n++) {
    if(cwidth) {
      gtk_clist_set_column_min_width(GTK_CLIST(*clist), n, cwidth[n]);
    }

    gtk_clist_set_column_auto_resize(GTK_CLIST(*clist), n, TRUE);
  }

  gtk_widget_show(*clist);
  gtk_widget_show(swin);
  return swin;
}


GtkWidget *
ww_menu(const char **array, size_t siz, GtkSignalFunc func)
{
  size_t n;
  GtkWidget *menu, *mitem;

  menu = gtk_menu_new();

  for(n = 0; n < siz; n++) {
    mitem = gtk_menu_item_new_with_label(array[n]);
    gtk_menu_append(GTK_MENU(menu), mitem);

    gtk_signal_connect(GTK_OBJECT(mitem), "activate", GTK_SIGNAL_FUNC(func), 
                       (gpointer)n);
    gtk_widget_show(mitem);
  }

  gtk_widget_show(menu);
  return menu;
}


GtkWidget *
ww_array_option(const char **array, size_t siz, int select, GtkSignalFunc func)
{
  GtkWidget *omenu, *menu;

  omenu = gtk_option_menu_new();
  menu = ww_menu(array, siz, func);

  gtk_option_menu_set_menu(GTK_OPTION_MENU(omenu), menu);
  gtk_option_menu_set_history(GTK_OPTION_MENU(omenu), select);
  gtk_widget_show(omenu);
  return omenu;
}


GtkWidget *
dialog_new(GUI_Dialog *gdialog, const char *title, ...)
{
  va_list pa;
  GtkWidget *w;

  va_start(pa, title);
  w = create_dialog(gdialog, title, pa);
  va_end(pa);

  return w;
}


GtkWidget *
dialog_msg_new(GUI_Dialog_Msg *gmdialog, const char *title,
               const char *msg, ...)
{
  va_list pa;
  GtkWidget *hbox;

  va_start(pa, msg);

  create_dialog(&gmdialog->dialog, title, pa);

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(hbox), 20);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(gmdialog->dialog.window)->vbox), hbox,
                     TRUE, TRUE, 0);

  gmdialog->msg = gtk_label_new(msg);
  gtk_box_pack_start(GTK_BOX(hbox), gmdialog->msg, TRUE, TRUE, 0);
  gtk_widget_show(gmdialog->msg);
  gtk_widget_show(hbox);

  return gmdialog->dialog.window;
}


GtkWidget *
dialog_err_new(GUI_Dialog_Err *gtdialog, const char *title,
               const char *msg, ...)
{
  va_list pa;
  GtkWidget *window, *vbox, *w;

  va_start(pa, msg);

  window = create_dialog(&gtdialog->mdialog.dialog, title, pa);
  gtk_widget_set_usize(window, 500, 250);
  gtk_widget_realize(window);

  vbox = gtk_vbox_new(FALSE, 5);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(window)->vbox), vbox, TRUE, TRUE, 0);
  gtk_widget_show(vbox);

  ww_frame(&gtdialog->mdialog.msg, NULL, msg, NULL, vbox);

  w = ww_text_area(&gtdialog->text, 0);
  gtk_box_pack_start(GTK_BOX(vbox), w, TRUE, TRUE, 0);

  gtdialog->mdialog.dialog.arg = NULL;
  gtdialog->have_err = FALSE;

  return window;
}


int
dialog_err_exit_new(int exitcode, const char *title, const char *fmt, ...)
{
  va_list ap;
  GUI_Dialog_Msg mdialog;
  char buf[FNAME_LEN*2], real_title[FNAME_LEN+1];

  va_start(ap, fmt);

#ifdef HAVE_VSNPRINTF
  vsnprintf(buf, FNAME_LEN*2, fmt, ap);
#else
  vsprintf(buf, fmt, ap);
#endif

  hsh_create_str(real_title, FNAME_LEN, PROGRAM, " ", VERSION_STRING,
                 " Error: ", title, NULL);

  dialog_msg_new(&mdialog, real_title, buf,
                 "OK", BUTTON_ISDEF, NULL, cb_exit, exitcode,
                 NULL);

  gtk_window_set_position(GTK_WINDOW(mdialog.dialog.window), GTK_WIN_POS_MOUSE);
  gtk_widget_show(mdialog.dialog.window);

  if(!gtk_main_level()) {
    gtk_main();
  }

  gtk_exit(exitcode);
  return exitcode;
}


int
dialog_err_return_new(int retcode, const char *title, const char *fmt, ...)
{
  va_list ap;
  GUI_Dialog_Msg *mdialog;
  char buf[FNAME_LEN*2], real_title[FNAME_LEN+1];

  va_start(ap, fmt);

#ifdef HAVE_VSNPRINTF
  vsnprintf(buf, FNAME_LEN*2, fmt, ap);
#else
  vsprintf(buf, fmt, ap);
#endif

  hsh_create_str(real_title, FNAME_LEN, PROGRAM, " ", VERSION_STRING,
                 " Error: ", title, NULL);

  mdialog = MallocMem(GUI_Dialog_Msg);

  dialog_msg_new(mdialog, real_title, buf,
                 "OK", BUTTON_ISDEF, NULL, cb_return, mdialog,
                 NULL);

  gtk_window_set_position(GTK_WINDOW(mdialog->dialog.window),
                          GTK_WIN_POS_MOUSE);
  gtk_widget_show(mdialog->dialog.window);

  return retcode;
}


static GtkWidget *
create_dialog(GUI_Dialog *gdialog, const char *title, va_list pa)
{
  GtkWidget *w;

  gdialog->arg = NULL;
  gdialog->window = gtk_dialog_new();
  gtk_window_set_title(GTK_WINDOW(gdialog->window), title);
  gtk_window_set_policy(GTK_WINDOW(gdialog->window), TRUE, TRUE, TRUE);

  gtk_signal_connect(GTK_OBJECT(gdialog->window), "delete_event",
                     GTK_SIGNAL_FUNC(cb_hide), gdialog);
  gtk_signal_connect(GTK_OBJECT(gdialog->window), "destroy",
                     GTK_SIGNAL_FUNC(gtk_widget_destroyed), &gdialog->window);

  w = create_buttons(GTK_DIALOG(gdialog->window)->action_area, WW_HBBOX,
                     GTK_BUTTONBOX_SPREAD, 0, GTK_SIGNAL_FUNC(cb_hide), 
                     gdialog, pa);

  return gdialog->window;
}


static GtkWidget *
create_buttons(GtkWidget *box, int type, GtkButtonBoxStyle layout, gint width,
               GtkSignalFunc stdfunc, gpointer data, va_list pa)
{
  gint flag;
  gpointer sdata;
  const char *name;
  GtkSignalFunc sfunc;
  GtkWidget *bbox, **button, *b;

  bbox = (type == WW_HBBOX ? gtk_hbutton_box_new() : gtk_vbutton_box_new());
  gtk_box_pack_start(GTK_BOX(box), bbox, FALSE, TRUE, 0);

  gtk_container_set_border_width(GTK_CONTAINER(bbox), width);
  gtk_button_box_set_spacing(GTK_BUTTON_BOX(bbox), 10);
  gtk_button_box_set_layout(GTK_BUTTON_BOX(bbox), layout);

  while((name = va_arg(pa, const char *))) {
    b = gtk_button_new_with_label(name);
    gtk_box_pack_start(GTK_BOX(bbox), b, TRUE, TRUE, 0);

    flag = va_arg(pa, gint);

    if(flag != BUTTON_NODEF) {
      GTK_WIDGET_SET_FLAGS(b, GTK_CAN_DEFAULT);

      if(flag == BUTTON_ISDEF) {
        gtk_widget_grab_default(b);
      }
    }

    if((button = va_arg(pa, GtkWidget **))) {
      *button = b;
    }

    sfunc = va_arg(pa, GtkSignalFunc);
    sdata = va_arg(pa, gpointer);

    if(!sfunc) {
      gtk_signal_connect(GTK_OBJECT(b), "clicked", stdfunc, data);
    } else {
      gtk_signal_connect(GTK_OBJECT(b), "clicked", sfunc, sdata);
    }    
  }

  gtk_widget_show_all(bbox);
  return bbox;
}


/* modified copy of default_compare() in gtkclist.c from GTK+ 1.2.0 */

static gint
clist_compare(GtkCList *clist, gconstpointer ptr1, gconstpointer ptr2)
{
  char *text1;
  char *text2;
  GtkCListRow *row1 = (GtkCListRow *)ptr1;
  GtkCListRow *row2 = (GtkCListRow *)ptr2;

  switch(row1->cell[clist->sort_column].type) {
    case GTK_CELL_TEXT: {
      text1 = GTK_CELL_TEXT(row1->cell[clist->sort_column])->text;
      break;
    }
    case GTK_CELL_PIXTEXT: {
      text1 = GTK_CELL_PIXTEXT(row1->cell[clist->sort_column])->text;
      break;
    }
    default: {
      text1 = NULL;
    }
  }

  switch(row2->cell[clist->sort_column].type) {
    case GTK_CELL_TEXT: {
      text2 = GTK_CELL_TEXT(row2->cell[clist->sort_column])->text;
      break;
    }
    case GTK_CELL_PIXTEXT: {
      text2 = GTK_CELL_PIXTEXT(row2->cell[clist->sort_column])->text;
      break;
    }
    default: {
      return (text1 != NULL);
    }
  }

  return (text1 ? strcasecmp(text1, text2) : -1);
}


static void
cb_hide(GtkWidget *w, gpointer data)
{
  if(((GUI_Dialog *)data)->window) {
    gtk_widget_hide(((GUI_Dialog *)data)->window);
  } else {
    gtk_widget_hide(w);
  }
}


static void
cb_exit(GtkWidget *w, gint data)
{
  gtk_exit(data);
}


static void
cb_return(GtkWidget *w, gpointer data)
{
  GUI_Dialog_Msg *mdialog = (GUI_Dialog_Msg *)data;

  gtk_widget_destroy(mdialog->dialog.window);
  FreeMem(mdialog);
}
