/* HMod 1.0.1 -- Hick modules emulation.
** Copyright (C) 2000 - 2001 Daniel Bergstrom <daniel@octocode.com>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2, or (at your option)
** any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <stdio.h>
#include <ctype.h>
#include <strings.h>

#include "hstring.h"

#include "hmod.h"
#include "hmodenv.h"
#include "hmodshell.h"

static int hmod_env_value_find(HModEnvVar *hvar, const char *value);
static void hmod_env_value_print(FILE *file, HModEnvVar *hvar, 
                                 HModShell *shell);

static HModEnvVar *hmod_env_get(HModEnv *envlist, const char *name);
static char *hmod_find_env(const char *name, HickPacket *hpkg, int internal);


void
hmod_env_set(HModEnv *henv, const char *name, const char *value)
{
  HModEnvVar *hvar = hmod_env_get(henv, name);

  if(hvar) {
    hmod_env_value_clear(hvar);
    hmod_env_value_append(hvar, value);
  }
}


void
hmod_env_append(HModEnv *henv, const char *name, const char *value)
{
  HModEnvVar *hvar = hmod_env_get(henv, name);

  if(hvar) {
    hmod_env_value_append(hvar, value);
  }
}


void
hmod_env_prepend(HModEnv *henv, const char *name, const char *value)
{
  HModEnvVar *hvar = hmod_env_get(henv, name);

  if(hvar) {
    hmod_env_value_prepend(hvar, value);
  }
}


void
hmod_env_unset(HModEnv *henv, const char *name)
{
  HModEnvVar *hvar = hmod_env_get(henv, name);

  if(hvar) {
    hmod_env_value_clear(hvar);
  }
}


void
hmod_env_remove(HModEnv *henv, const char *name, const char *value)
{
  Node *node;
  HModEnvVar *hvar = hmod_env_get(henv, name);

  if(hvar) {
    if(!value) {
      hmod_env_value_clear(hvar);
      return;
    }

    for(node = hvar->values.head; node; node = node->next) {
      if(!strcmp(((HModEnvValue *)node)->value, value)) {
        hsh_node_remove(node, &hvar->values);
        return;
      }
    }
  }
}


void
hmod_env_print(FILE *file, HModEnv *henv, const char *name)
{
  Node *node;
  HModEnvVar *hvar;
  HModShell *shell = hmod_shell_get(hmod_shell_type_get(name));

  for(node = henv->vars.head; node; node = node->next) { /* set */
    hvar = (HModEnvVar *)node;

    if(hvar->values.head && !hvar->values.head->next) {
      hmod_env_value_print(file, hvar, shell);
    }
  }

  for(node = henv->vars.head; node; node = node->next) { /* prepend, append */
    hvar = (HModEnvVar *)node;

    if(hvar->values.head && hvar->values.head->next) {
      hmod_env_value_print(file, hvar, shell);
    }
  }

  for(node = henv->vars.head; node; node = node->next) { /* unset */
    hvar = (HModEnvVar *)node;

    if(!hvar->values.head) {
      hmod_env_value_print(file, hvar, shell);
    }
  }
}


static void
hmod_env_value_print(FILE *file, HModEnvVar *hvar, HModShell *shell)
{
  Node *node;
  HString str;

  if(!hvar->values.head) {
    fprintf(file, shell->unset, hvar->name, hvar->name);
    return;
  }

  hstring_init(&str);

  for(node = hvar->values.head; node; node = node->next) {
    if(hstring_length(&str)) {
      hstring_add_char(&str, ':');
    }

    hstring_add_string(&str, ((HModEnvValue *)node)->value, 
                       strlen(((HModEnvValue *)node)->value));
  }

  hstring_eos(&str);
  fprintf(file, shell->set, hvar->name, hstring_data(&str), hvar->name);
  hstring_free(&str);
}


static HModEnvVar *
hmod_env_get(HModEnv *henv, const char *name)
{
  char *p;
  size_t n;
  Node *node;
  HString str;
  HModEnvVar *hvar;

  for(node = henv->vars.head; node; node = node->next) {
    if(!strcmp(((HModEnvVar *)node)->name, name)) {
      return (HModEnvVar *)node;
    }
  }

  hmod_out_of_memory((hvar = MallocMem(HModEnvVar)));

  hsh_nodelist_init(&hvar->values);
  hvar->name = hmod_strdup(name);

  hsh_node_append(&hvar->node, &henv->vars);

  if((p = getenv(name))) {
    hstring_init(&str);

    for(n = 0; p[n]; n++) {
      if(p[n] == ' ') {
        hstring_add_char(&str, '\\');
      }

      hstring_add_char(&str, p[n]);
    }

    hstring_eos(&str);
    p = strtok(hstring_data(&str), ":");

    while(p) {
      hmod_env_value_append(hvar, p);
      p = strtok(0, ":");
    }

    hstring_free(&str);
  }

  return hvar;
}


static int
hmod_env_value_find(HModEnvVar *hvar, const char *value)
{
  Node *node;

  for(node = hvar->values.head; node; node = node->next) {
    if(!strcasecmp(((HModEnvValue *)node)->value, value)) {
      return 1;
    }
  }

  return 0;
}


void
hmod_env_value_append(HModEnvVar *hvar, const char *value)
{
  HModEnvValue *hvalue;

  if(!hmod_env_value_find(hvar, value)) {
    hmod_out_of_memory((hvalue = MallocMem(HModEnvValue)));
    hvalue->value = hmod_strdup(value);

    hsh_node_append(&hvalue->node, &hvar->values);
  }
}


void
hmod_env_value_prepend(HModEnvVar *hvar, const char *value)
{
  HModEnvValue *hvalue;

  if(!hmod_env_value_find(hvar, value)) {
    hmod_out_of_memory((hvalue = MallocMem(HModEnvValue)));
    hvalue->value = hmod_strdup(value);

    hsh_node_prepend(&hvalue->node, &hvar->values);
  }
}


void
hmod_env_value_clear(HModEnvVar *hvar)
{
  Node *node;

  for(node = hvar->values.head; node; node = hvar->values.head) {
    hvar->values.head = node->next;
    FreeMem(((HModEnvValue *)node)->value);
    FreeMem(((HModEnvValue *)node));
  }

  hsh_nodelist_init(&hvar->values);
}


static char *
hmod_env_parse_value(const char *name, HickEnv *env, HickPacket *hpkg,
                     HModEnv *henv)
{
  HString res, var;
  char *p, *value = env->env_value;
  size_t n, gotvar, curly, expanded;

  hstring_init(&var);
  hstring_init(&res);

  do {
    expanded = 0;

    for(gotvar = n = 0; value[n]; n++) {
      if(!gotvar) {
        if(value[n] == '$') {
          if(value[n + 1] == '$') {
            hstring_add_char(&res, value[++n]);
            continue;
          }

          if(value[n + 1] == '{') {
            n++;
            curly = 1;
          } else {
            curly = 0;
          }

          gotvar = 1;
          expanded = 1;
          hstring_clear(&var);
          continue;
        }
        
        hstring_add_char(&res, value[n]);
        continue;
      }

      if((curly && value[n] == '}') || 
         (!isalnum((int)value[n]) && (value[n] != '_'))) {
        gotvar = 0;
        n -= !curly;
      }

      if(!gotvar) {
        hstring_eos(&var);
        p = hmod_find_env(hstring_data(&var), hpkg, 
                          strcmp(env->env_name, hstring_data(&var)));

        if(p) {
          hstring_add_string(&res, p, strlen(p));
        } else {
          fprintf(stderr, "module: unset variable '%s' in '%s'.\n", 
                  hstring_data(&var), name);
          return 0;
        }

        continue;
      }
    
      hstring_add_char(&var, value[n]);
    }

    if(gotvar) {
      hstring_eos(&var);
      p = hmod_find_env(hstring_data(&var), hpkg, 
                        strcmp(env->env_name, hstring_data(&var)));

      if(p) {
        hstring_add_string(&res, p, strlen(p));
      } else {
        fprintf(stderr, "module: unset variable '%s' in '%s'.\n", 
                hstring_data(&var), name);
        return 0;
      }
    }

    hstring_free(&var);
    hstring_eos(&res);

    FreeMem(value);
    value = hstring_data(&res);
    hstring_init(&res);
  } while(expanded);

  return (env->env_value = value);
}


static char *
hmod_find_env(const char *name, HickPacket *hpkg, int internal)
{
  char *p = 0;
  HickEnv *env;
  HickVer *hver;

  if((p = getenv(name))) {
    return p;
  }

  if(internal) {
    for(env = hpkg->pac_env.el_head; env; env = env->env_next) {
      if(!strcmp(env->env_name, name)) {
        return env->env_value;
      }
    }

    for(hver = hpkg->pac_versions.vl_head; hver; hver = hver->ver_next) {
      for(env = HVER_ENV(hver).el_head; env; env = env->env_next) {
        if(!strcmp(env->env_name, name)) {
          return env->env_value;
        }
      }
    }
  }

  return 0;
}


int
hmod_env_parse(HickPacket *hpkg, HModEnv *henv)
{
  HickEnv *env;
  HickVer *hver;

  for(env = hpkg->pac_env.el_head; env; env = env->env_next) {
    if(!hmod_env_parse_value(hpkg->pac_name, env, hpkg, henv)) {
      return 0;
    }
  }
  
  for(hver = hpkg->pac_versions.vl_head; hver; hver = hver->ver_next) {
    for(env = HVER_ENV(hver).el_head; env; env = env->env_next) {
      if(!hmod_env_parse_value(hpkg->pac_name, env, hpkg, henv)) {
        return 0;
      }
    }
  }

  return 1;
}
