/* HMod 1.0.1 -- Hick modules emulation.
** Copyright (C) 2000 - 2001 Daniel Bergstrom <daniel@octocode.com>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2, or (at your option)
** any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "hmod.h"
#include "hick.h"
#include "entry.h"
#include "linker.h"
#include "hickenv.h"
#include "dofuncs.h"
#include "version.h"
#include "hmodadd.h"
#include "hmodload.h"
#include "hmodlist.h"

typedef struct {
  const char *name;
  const char *env;
} HModLinkConv;


static HModLinkConv link_conv[] = {
  { "bin",          "PATH" },
  { "sbin",         "PATH" },
  { "lib",          "LD_LIBRARY_PATH" },
  { "man",          "MANPATH" },
  { "include",      "C_INCLUDE_PATH" },
  { "info",         "INFOPATH" },
  { "help",         "HELPPATH" },
  { "aclocal",      "ACLOCAL_PATH" },
  { "app-defaults", "XFILESEARCHPATH" },
  { "plug_ins",     "NPX_PLUGIN_PATH" },
  { "plugins",      "NPX_PLUGIN_PATH" },
  { "plugin",       "NPX_PLUGIN_PATH" },
  { 0,              0 }
};

static void hmod_load_check_require(PArr *loaded, HickPackets *pkg,
                                    HickPacket *hpkg, HickVerList *inst,
                                    HickVerList *insted, HickRequire **head);
static HickEnv *hmod_link_convert(HickEnv *env, HickEnvList *list);
static const char *hmod_get_env(const char *name);
static int hmod_load_version_check(HickPacket *hpkg, PArr *vers, int tryadd);
static void hmod_load_env_dup(HickPacket *hpkg, HickVer *hver,
                              HickEnvList *list);

extern HickRequire *checkRequire(HickRequire *reqList, userEntryList *install,
                                 userEntryList *installed, const char *id, 
                                 FILE *logFile);

void
hmod_load_update(HModLoadUpdate *update, HModEnv *henv, HickPackets *pkg,
                 PArr *loaded, int tryadd, char **packets, size_t siz)
{ 
  size_t n, pos;
  HickEnv *env;
  HickVer *hver;
  HickPacket hpkg;
  char *name, *ver;
  HickVerList nouse;
  PArr install, vers;
  HickEnvList environ;
  HickRequire *reqList = 0;

  parr_init(&install);

  for(pos = n = 0; n < siz; n++, pos = 0) {
    if(hmod_hickloaded_get(packets[n], &pos, &name, &ver)) {
      hmod_list_add(&install, name, ver);
    }
  }

  initHickPacket(&hpkg);
  initHickVerList(&nouse);
  initHickEnvList(&environ);

  while(nextPacketFull(pkg, &hpkg)) {
    parr_init(&vers);
    freeHickVerList(&nouse, 1);

    for(n = 0; n < install.siz; n++) {
      if(strcasecmp(parr_get(&install, HModLoaded, n)->name, hpkg.pac_name)) {
        continue;
      }

      parr_add(&vers, parr_remove(&install, n--));
    }

    if(!vers.siz || !(n = hmod_load_version_check(&hpkg, &vers, tryadd))) {
      hmod_load_check_require(loaded, pkg, &hpkg, 0, &hpkg.pac_versions, 
                              &reqList);
      continue;
    }

    if(update->func_ver && !(update->func_ver)(&hpkg, loaded, tryadd, &vers, 
                                               (n == 1 ? 0 : 1), &nouse)) {
      hmod_load_check_require(loaded, pkg, &hpkg, 0, &nouse, &reqList);
      continue;
    }

    if(update->func_insted && (update->func_insted)(&hpkg, loaded, tryadd)) {
      hmod_load_check_require(loaded, pkg, &hpkg, &hpkg.pac_versions, &nouse,
                              &reqList);
      continue;
    }

    if(update->func_conflict && (update->func_conflict)(&hpkg, pkg, loaded, 
                                                        tryadd,
                                                        update->conflict_chk)){
      hmod_load_check_require(loaded, pkg, &hpkg, &hpkg.pac_versions, &nouse,
                              &reqList);
      continue;
    }

    if(!hmod_env_parse(&hpkg, henv)) {
      hmod_load_check_require(loaded, pkg, &hpkg, &hpkg.pac_versions, &nouse,
                              &reqList);
      continue;
    }

    if(update->environ_rescue && hpkg.pac_versions.vl_head) {
      for(hver = hpkg.pac_versions.vl_head; hver; hver = hver->ver_next) {
        if(_entryFindTwo(hpkg.pac_name, hver->ver_number, &pkg->hp_installed)){
          hmod_load_env_dup(&hpkg, hver, &environ);
        }
      }

      for(hver = nouse.vl_head; hver; hver = hver->ver_next) {
        if(_entryFindTwo(hpkg.pac_name, hver->ver_number, &pkg->hp_installed)){
          hmod_load_env_dup(&hpkg, hver, &environ);
        }
      }
    }

    hmod_load_link_convert(&hpkg); 

    if(update->func_env) {
      for(env = hpkg.pac_env.el_head; env; env = env->env_next) {
        (update->func_env)(env, henv);
      }

      for(hver = hpkg.pac_versions.vl_head; hver; hver = hver->ver_next) {
        if(update->func_loaded) {
          (update->func_loaded)(loaded, hpkg.pac_name, hver->ver_number);
        }

        for(env = HVER_ENV(hver).el_head; env; env = env->env_next) {
          (update->func_env)(env, henv);
        }
      }
    }

    if(environ.el_head) {
      for(env = environ.el_head; env; env = env->env_next) {
        hmod_add_environ(env, henv);
      }

      initHickEnvList(&environ);
    }

    hmod_load_check_require(loaded, pkg, &hpkg, &hpkg.pac_versions, &nouse,
                            &reqList);
  }

  for(n = 0; n < loaded->siz; n++) {
    _entryNew(parr_get(loaded, HModLoaded, n)->name, 
              parr_get(loaded, HModLoaded, n)->version, time(0), 0, 
              &pkg->hp_installed);
  }

  checkRequire(reqList, &pkg->hp_install, &pkg->hp_installed, "module:",
               stderr);

  if(!tryadd) {
    for(n = 0; n < install.siz; n++) {
      fprintf(stderr, "module: couldn't find packet '%s/%s'\n." ,
              parr_get(&install, HModLoaded, n)->name,
              parr_get(&install, HModLoaded, n)->version);
    }
  }
}

static void
hmod_load_env_dup(HickPacket *hpkg, HickVer *hver, HickEnvList *list)
{
  HickEnv *env, *nenv;

  for(env = hpkg->pac_env.el_head; env; env = env->env_next) {
    hmod_out_of_memory((nenv = MallocMem(HickEnv)));
    nenv->env_flags    = env->env_flags;
    nenv->env_priority = env->env_priority;
    nenv->env_name     = hmod_strdup(env->env_name);
    nenv->env_value    = hmod_strdup(env->env_value);

    nenv->env_prev = list->el_tail;
    list->el_tail = (list->el_tail ? (list->el_tail->env_next = nenv) :
                     (list->el_head = nenv));
  }

  for(env = HVER_ENV(hver).el_head; env; env = env->env_next) {
    hmod_out_of_memory((nenv = MallocMem(HickEnv)));
    nenv->env_flags    = env->env_flags;
    nenv->env_priority = env->env_priority;
    nenv->env_name     = hmod_strdup(env->env_name);
    nenv->env_value    = hmod_strdup(env->env_value);

    nenv->env_next = 0;
    nenv->env_prev = list->el_tail;
    list->el_tail = (list->el_tail ? (list->el_tail->env_next = nenv) :
                     (list->el_head = nenv));
  }
}


static void
hmod_load_check_require(PArr *loaded, HickPackets *pkg, HickPacket *hpkg, 
                        HickVerList *inst, HickVerList *insted, 
                        HickRequire **head)
{
  HickVer *ver;
  HickEnv *env;
  HModLoaded *lnode;
  HickRequirePacket *glist = 0, *rpac;

  for(env = hpkg->pac_require.el_head; env; env = env->env_next) {
    if((rpac = env->u.env_rpacs)) {
      rpac->rpac_next = glist;
      glist = rpac;
      env->u.env_rpacs = NULL;
    }
  }

  if(inst) {
    for(ver = inst->vl_head; ver; ver = ver->ver_next) {
      if(!(lnode = hmod_list_find(loaded, hpkg->pac_name, ver->ver_number)) &&
         !_entryFindTwo(hpkg->pac_name, ver->ver_number, &pkg->hp_installed)) {
        continue;
      }

      parseRequire(hpkg->pac_name, ver->ver_number, HVER_REQUIRE(ver).el_head,
                   glist, head);
    }
  }

  if(insted) {
    for(ver = insted->vl_head; ver; ver = ver->ver_next) {
      if(!(lnode = hmod_list_find(loaded, hpkg->pac_name, ver->ver_number)) &&
         !_entryFindTwo(hpkg->pac_name, ver->ver_number, &pkg->hp_installed)) {
        continue;
      }

      parseRequire(hpkg->pac_name, ver->ver_number, HVER_REQUIRE(ver).el_head,
                   glist, head);
    }
  }

  for(rpac = glist; rpac; rpac = glist) {
    glist = rpac->rpac_next;
    _freeHickRequirePacket(rpac);
  }
}


void
hmod_load_link_convert(HickPacket *hpkg)
{
  HickEnv *env;
  HickVer *hver;

  for(env = hpkg->pac_link.el_head; env; env = env->env_next) {
    hmod_link_convert(env, &hpkg->pac_env);
  }

  for(hver = hpkg->pac_versions.vl_head; hver; hver = hver->ver_next) {
    for(env = HVER_LINK(hver).el_head; env; env = env->env_next) {
      hmod_link_convert(env, &hpkg->pac_env);
    }
  }
}


static HickEnv *
hmod_link_convert(HickEnv *env, HickEnvList *list)
{
  size_t len;
  const char *p;
  char filename[FNAME_LEN + 1], path[FNAME_LEN + 1];
  char tofile[FNAME_LEN + 1], topath[FNAME_LEN + 1];

  if((env->env_name[0] == '/') || (env->env_name[0] == '~')) {
    return 0;
  }

  if(env->env_flags == HFLAG_LINK_RENAME) {
    _getFilename(0, topath, env->env_name, strlen(env->env_name));
  } else {
    strcpy(topath, env->env_name);
  }

  _getFilename(filename, path, env->env_value, strlen(env->env_value));

  do {
    _getFilename(tofile, topath, topath, strlen(topath));
    _getFilename(filename, path, path, strlen(path));
  } while(topath[0] && path[0]);

  if(path[0] && (p = hmod_get_env(tofile))) {

    len = strlen(path) + strlen(filename) + 2;
    hmod_out_of_memory((env = MallocMem(HickEnv)));
    env->env_next      = env->env_prev = 0;
    env->env_flags     = HFLAG_ENV_PREPEND;
    env->env_priority  = HICKPRI_ENVIRON;
    env->env_name      = hmod_strdup(p);
    env->env_value     = MallocArr(char, len);
    env->env_itemTitle = env->env_about = 0;
    hsh_array_init(&env->env_types);

    hmod_out_of_memory(env->env_value);
    hsh_create_str(env->env_value, len, path, "/", filename, NULL);

    if(list) {
      env->env_prev = list->el_tail;
      list->el_tail = (list->el_tail ? (list->el_tail->env_next = env) :
                       (list->el_head = env));
    }

    return env;
  }
  
  return 0;
}


static const char *
hmod_get_env(const char *name)
{
  size_t n;

  for(n = 0; link_conv[n].name; n++) {
    if(!strcasecmp(link_conv[n].name, name)) {
      return link_conv[n].env;
    }
  }

  return 0;
}


static int
hmod_load_version_check(HickPacket *hpkg, PArr *vers, int tryadd)
{
  size_t n, isdef = 1;
  HModLoaded *lnode;

  if(!hpkg->pac_versions.vl_head) {
    if(!tryadd) {
      for(n = 0; n < vers->siz; n++) {
        fprintf(stderr, "module: failed to find '%s/%s'.\n", 
                parr_get(vers, HModLoaded, n)->name, 
                parr_get(vers, HModLoaded, n)->version);
      }
    }

    return 0;
  }

  for(n = 0; n < vers->siz; n++) { /* named version overrides non entered. */
    lnode = parr_get(vers, HModLoaded, n);

    if(lnode->version[0]) {
      if(!_versionFind(lnode->version, &hpkg->pac_versions)) {
        parr_remove(vers, n--);

        if(!tryadd) {
          fprintf(stderr, "module: failed to find '%s/%s'.\n", 
                  lnode->name, lnode->version);
        }

        continue;
      }

      isdef = 0;
    }
  }

  if(!vers->siz) {
    return 0;
  }

  return (isdef ? 2 : 1);
}
