/* HMod 1.0.1 -- Hick modules emulation.
** Copyright (C) 2000 - 2001 Daniel Bergstrom <daniel@octocode.com>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2, or (at your option)
** any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <pwd.h>

#include "hmod.h"
#include "hmodadd.h"
#include "hmodrem.h"
#include "hmodenv.h"
#include "hmodcmd.h"
#include "hmodlist.h"
#include "hmodavail.h"
#include "hmodshell.h"

extern int optind;

static struct passwd *get_user(uid_t uid);
static char *get_hicktype(const char *path);


static int
usage(void)
{
  fputs("Usage: module [ OPTIONS ] COMMAND [ args ... ]\n\n"
        "  Commands:\n"
        "\tadd | load packet [ packet ... ]     Add packet(s).\n"
        "\ttry-add packet [ packet ... ]\t     Add without complaining.\n"
        "\trm | unload packet [ packet ... ]    Remove module(s).\n"
        "\tlist\t\t\t\t     List loaded packets.\n"
        "\tavail [ match ... ]\t\t     List available packets.\n"
        "\thelp\t\t\t\t     Prints this help text.\n", stderr);
  fputs("\tversion\t\t\t\t     Show version information.\n\n"
        "  Options:\n"
        "\t-c\t\t\t\t     Conflict checking with module.\n"
        "\t\t\t\t\t     install packets only.\n"
        "\t-P STRING\t\t\t     Other user hick home then ~/.hick.\n"
        "\t-h\t\t\t\t     Prints this help text.\n"
        "\t-v\t\t\t\t     Show version information.\n\n", stderr);

  return 0;
}


int
main(int argc, char **argv)
{
  PArr loaded;
  HModCmd cmd;
  HModEnv henv;
  HickPackets pkg;
  NodeList scripts;
  struct passwd *pwd;
  int n, confchk = 1;
  char *hicktype, *path = 0, pwdpath[FNAME_LEN + 1];

  if(argc <= 1) {
    fprintf(stderr, "module: missing shell type.\n");
    return usage();
  }

  while((n = getopt(argc - 1, argv + 1, "chvP:")) != EOF) {
    switch(n) {
      case 'c': {               /* version information */
        confchk = 0;
        break;
      }
      case 'h': {
        return usage();
      }
      case 'v': {
        fprintf(stderr, "%s %s\n", HMOD_PRG, HMOD_VER);
        return 0;
      }
      case 'P': {
        path = optarg;
        break;
      }
    }
  }

  if(argc - (optind - 1) <= 2) {
    fprintf(stderr, "module: missing command.\n");
    return usage();
  }

  if(hmod_shell_type_get(argv[1]) == SHELL_TYPE_NONE) {
    fprintf(stderr, "module: unknown shell '%s'.\n", argv[1]);
    return usage();
  }

  if((cmd = hmod_command_get(argv[optind + 1])) == HMOD_NONE) {
    fprintf(stderr, "module: unknown command '%s'.\n", argv[optind + 1]);
    return usage();
  }

  if(!(pwd = get_user(geteuid())) || !(hicktype = get_hicktype(PACKET_PATH))) {
    return usage();
  }

  if(!path) {
    hsh_create_str(pwdpath, FNAME_LEN, pwd->pw_dir, "/.hick", NULL);
    path = pwdpath;
  }

  if((n = openPackets(PACKET_PATH, path, hicktype, 0, &pkg))) {
    switch(n) {
      case HICKERR_PACKDIR: {
        fprintf(stderr, "module failed to load packets from '%s'.\n",
                PACKET_PATH);
        return 0;
      }
      case HICKERR_OUTOFMEMORY: {
        fprintf(stderr, "module: Out of memory -- exiting.\n");
        return 0;
      }
      default: {
        fprintf(stderr, "module unknown packet error.\n");
        return 0;
      }
    }
  }

  hmod_env_init(&henv);

  switch(cmd) {
    case HMOD_LIST: {
      hmod_list_print(stderr);
      break;
    }
    case HMOD_AVAIL: {
      if(argc - (optind - 1) >= 4) {
        hmod_avail_print(stderr, &pkg, argv + 3, argc - 3);
      } else {
        hmod_avail_print(stderr, &pkg, 0, 0);
      }
      break;
    }
    case HMOD_HELP: {
      return usage();
    }
    case HMOD_VERSION: {
      fprintf(stderr, "%s %s\n", HMOD_PRG, HMOD_VER);
      break;
    }
    case HMOD_ADD:
    case HMOD_TRYADD: {
      if(argc - (optind - 1) < 4) {
        fprintf(stderr, "module: missing argument to command.\n");
        return usage();
      }

      hsh_nodelist_init(&scripts);
      hmod_list_get(&loaded);

      hmod_add_get(&henv, &scripts, &pkg, &loaded, argv[1], 
                   (cmd == HMOD_TRYADD), argv + optind + 2, 
                   argc - (optind + 2), confchk);

      hmod_env_print(stdout, &henv, argv[1]);
      hmod_list_put(stdout, &loaded, argv[1]);
      break;
    }
    case HMOD_REMOVE: {
      if(argc < 4) {
        fprintf(stderr, "module: missing argument to command.\n");
        return usage();
      }

      hmod_list_get(&loaded);

      hmod_rem_get(&henv, &pkg, &loaded, argv[1], argv + optind + 2, 
                   argc - (optind + 2));

      hmod_env_print(stdout, &henv, argv[1]);
      hmod_list_put(stdout, &loaded, argv[1]);
      break;
    }
    default: {
      break;
    }
  }

  return 0;
}


static char *
get_hicktype(const char *path)
{
  size_t n;
  char *type;
  ArrayInfo types;

  if(!(type = getenv(HICKUPTYPE))) {
    fprintf(stderr, "$%s is not set.\n", HICKUPTYPE);
    return 0;
  }

  switch(getHickTypes(path, &types)) {
    case HICKERR_TYPEDIR: {
      fprintf(stderr, "Can't access hick type directory.\n");
      return 0;
    }
    case HICKERR_OUTOFMEMORY: {
      fprintf(stderr, "Out of memory!\n");
      exit(1);
    }
  }

  if(!(n = hsh_array_find(type, &types))) {
    printf("Non existing $%s '%s' has been specified.\n", HICKUPTYPE, type);
    return 0;
  }

  return hmod_strdup(types.arr[n - 1]);
}


static struct passwd *
get_user(uid_t uid)
{
  struct passwd *pwd;

  if(!(pwd = getpwuid(uid))) {
    fprintf(stderr, "Failed to obtain your user entry (UID #%d).", uid);
    return 0;
  }

  if(!pwd->pw_name || !*pwd->pw_name) {
    fprintf(stderr, "User name is missing in the password entry (UID #%d).",
            uid);
    return 0;
  }

  if(!pwd->pw_dir || !*pwd->pw_dir) {
    fprintf(stderr, "No home directory specified for user '%s'.",
            pwd->pw_name);
    return 0;
  }

  return pwd;
}
