/* HICK Lib 1.3.2 -- User Profiler Library.
** Copyright (C) 1997 - 2000 Daniel Bergstrom <daniel@octocode.com>
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the Free
** Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "entry.h"
#include "common.h"
#include "hickchk.h"
#include "getfile.h"
#include "version.h"

static void checkPrepend(HickCheckEntry *chk, HickCheckEntry **list);

static HickCheckEntry *checkNew(const char *name, const char *version,
                                const char *nVersion, enum hickCheck flag,
                                time_t etime);


int
checkPackets(const char *packpath, const char *homepath, const char *hicktype,
             unsigned int mask, time_t newsince, HickCheck *result)
{
  FILE *file;
  time_t curTime;
  HickEntry entry;
  HickPackets pack;
  userEntryList insted;
  userEntry *ue, *uetmp;
  HickVer *ver, *verdef;
  char fname[FNAME_LEN+1];
  HickCheckEntry *defent, *chkent;
  int storeTime = 0, err, freeUE, expFound;

  result->hc_expire = NULL;
  result->hc_new = result->hc_changed = result->hc_removed = NULL;

  /* Open a HickPackets, and start reading packets */

  if((err = openPackets(packpath, homepath, hicktype, 0, &pack))) {
    return err;
  }

  if((mask & CHK_NEW) && !newsince) {
    /* no check time was set, try to get one from the stored time stamp. */
    storeTime = 1;

    hsh_create_str(fname, FNAME_LEN, homepath, "/.", hicktype, "/",
                   CHECK_FILE, NULL);

    if((file = fopen(fname, "r"))) { /* load the time stamp, if any. */
      newsince = _getNumber(file);
      fclose(file);
    }
  }

  time(&curTime);

  initHickEntry(&entry);
  initUserEntryList(&insted);

  while(nextPacket(&pack, &entry, 0)) {
    defent = NULL;

    /* Get the default version, if any. */
    for(verdef = entry.he_versions.vl_head; verdef && !verdef->ver_default;
        verdef = verdef->ver_next);

    if(mask & CHK_NEW) { /* Check for new packets and versions */
      if(entry.he_created && (entry.he_created > newsince)) {
        chkent = checkNew(entry.he_name, (verdef ? verdef->ver_number : ""),
                          NULL, HICKCHK_NEWPACK, entry.he_created);
        checkPrepend(chkent, &result->hc_new);
      } else { /* if not a new packet, check for new versions */
        for(ver = entry.he_versions.vl_head; ver; ver = ver->ver_next) {
          if(ver->ver_created && (ver->ver_created > newsince)) {
            chkent = checkNew(entry.he_name, ver->ver_number, NULL,
                              HICKCHK_NEWVER, ver->ver_created),
              checkPrepend(chkent, &result->hc_new);
          }
        }
      }
    }

    if(!(mask & CHK_CHANGED) && !(mask & CHK_REMOVED) && !(mask & CHK_EXPIRE)){
      freeHickEntry(&entry);
      continue;
    }

    /* Get out the userEntry for this packet from the installed list. */
    for(ue = pack.hp_installed.ul_head; ue; ue = uetmp) {
      uetmp = ue->ue_next;

      if(!strcasecmp(ue->ue_name, entry.he_name)) {
        _entryMove(ue, &pack.hp_installed, &insted); 
      }
    }

    if(!insted.ul_head) { /* Nothing to check? */
      freeHickEntry(&entry);
      continue;
    }

    /* We first check packet changes */

    for(expFound = freeUE = 0, ue = insted.ul_head; ue; ue = uetmp){
      uetmp = ue->ue_next;

      if((mask & CHK_REMOVED) &&
         !(ver = _versionFind(ue->ue_version, &entry.he_versions))) {

        checkPrepend(checkNew(entry.he_name, ue->ue_version,
                              (verdef->ver_number ? verdef->ver_number : ""),
                              HICKCHK_REMVER, 0), &result->hc_removed);
        FreeMem(_entryRemove(ue, &insted));
        continue;
      }

      if((mask & CHK_EXPIRE) && entry.he_expire && !expFound) {
        if(entry.he_expire < curTime) {
          chkent = checkNew(ue->ue_name, ue->ue_version, 0, HICKCHK_EXPPACK,
                            entry.he_expire);
          checkPrepend(chkent, &result->hc_expire);
          expFound = 1;
        } else {
          chkent = checkNew(ue->ue_name, ue->ue_version, 0, HICKCHK_ISEXPPACK,
                            entry.he_expire);
          checkPrepend(chkent, &result->hc_expire);
        }

        freeUE = 1;
      }

      if(mask & CHK_CHANGED) {
        if(!defent && (ue->ue_flags == UE_DEFAULT) &&
           strcasecmp(ue->ue_version, verdef->ver_number)) {
          defent = checkNew(entry.he_name, ue->ue_version, verdef->ver_number,
                            HICKCHK_CHDEF, 0);
        } else {
          if(entry.he_modified && (entry.he_modified > ue->ue_date)) {
            checkPrepend(checkNew(entry.he_name, ue->ue_version, 0,
                                  HICKCHK_MODPACK, entry.he_modified),
                         &result->hc_changed);
          }
        }
      }

      if(freeUE) {
        FreeMem(_entryRemove(ue, &insted));
      }
    }

    /* Now check version changes */

    for(ver = entry.he_versions.vl_head; ver; ver = ver->ver_next) {
      for(ue = insted.ul_head; ue; ue = uetmp) {
        uetmp = ue->ue_next;

        if(strcasecmp(ue->ue_version, ver->ver_number)) {
          continue;
        }

        if((mask & CHK_EXPIRE) && ver->ver_expire) {
          if(ver->ver_expire < curTime) {
            checkPrepend(checkNew(entry.he_name, ue->ue_version, NULL,
                                  HICKCHK_EXPVER, ver->ver_expire),
                         &result->hc_expire);
          } else {
            checkPrepend(checkNew(entry.he_name, ue->ue_version, NULL,
                                  HICKCHK_ISEXPVER, ver->ver_expire),
                         &result->hc_expire);
          }
          continue;
        }

        if(mask & CHK_CHANGED) {
          if(defent && strcasecmp(defent->chk_version, ue->ue_version)) {
            continue;
          }

          if(ver->ver_modified && (ver->ver_modified > ue->ue_date)) {
            checkPrepend(checkNew(entry.he_name, ue->ue_version, NULL,
                                  HICKCHK_MODVER, ver->ver_modified),
                         &result->hc_changed);
          }
        }
      }
    }

    if(defent) {
      checkPrepend(defent, &result->hc_changed);
    }

    freeHickEntry(&entry);
    _freeUserEntryList(&insted);
  }

  if(mask & CHK_REMOVED) {
    while(pack.hp_installed.ul_head) {
      _freeUserEntryList(&insted);

      for(ue = pack.hp_installed.ul_head; ue; ue = uetmp) {
        uetmp = ue->ue_next;

        if(!insted.ul_head) {
          _entryMove(ue, &pack.hp_installed, &insted);
          continue;
        }

        if(!strcasecmp(insted.ul_head->ue_name, ue->ue_name)) {
          FreeMem(_entryRemove(ue, &pack.hp_installed));
        }
      }

      if((ue = insted.ul_head)) {
        checkPrepend(checkNew(ue->ue_name, ue->ue_version, NULL,
                              HICKCHK_REMPACK, 0), &result->hc_removed);
      }
    }
  }

  closePackets(&pack);
  freeHickEntry(&entry);
  _freeUserEntryList(&insted);

  if((mask & CHK_NEW) && storeTime) { /* write time stamp file */
    if(!(file = fopen(fname, "w"))) {
      return 0;
    }

    fprintf(file, "%ld\n", time(NULL));
    fclose(file);
  }

  return 0;
}


HickCheck *
freeHickCheck(HickCheck *hc)
{
  HickCheckEntry *chk;

  for(chk = hc->hc_new; chk; chk = hc->hc_new) {
    hc->hc_new = chk->chk_next;
    FreeMem(chk);
  }

  for(chk = hc->hc_changed; chk; chk = hc->hc_changed) {
    hc->hc_changed = chk->chk_next;
    FreeMem(chk);
  }

  for(chk = hc->hc_removed; chk; chk = hc->hc_removed) {
    hc->hc_removed = chk->chk_next;
    FreeMem(chk);
  }

  for(chk = hc->hc_expire; chk; chk = hc->hc_expire) {
    hc->hc_expire = chk->chk_next;
    FreeMem(chk);
  }

  return hc;
}


static void
checkPrepend(HickCheckEntry *chk, HickCheckEntry **list)
{
  if(chk) {
    chk->chk_next = *list;
    *list = chk;
  }
}


static HickCheckEntry *
checkNew(const char *name, const char *version, const char *nVersion,
         enum hickCheck flag, time_t etime)
{
  HickCheckEntry *chk;

  if(!(chk = MallocMem(HickCheckEntry))) {
    return NULL; /* out of memory. */
  }

  chk->chk_next = NULL;
  chk->chk_flag = flag;
  chk->chk_time = etime;
  strcpy(chk->chk_name, name);
  strcpy(chk->chk_version, (version ? version : ""));
  strcpy(chk->chk_nVersion, (nVersion ? nVersion : ""));

  return chk;
}
