/* HICK Lib 1.3,2 - User Profiler Library.
** Copyright (C) 1997 - 2000 Daniel Bergstrom <daniel@octocode.com>
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the Free
** Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "memd.h"
#include "entry.h"


userEntry *
_entryNew(const char *name, const char *version, time_t datet,
          unsigned int flag, userEntryList *to)
{
  userEntry *ue;

  if((ue = MallocMem(userEntry))) {
    ue->ue_date = datet;
    ue->ue_flags = flag;

    strncpy(ue->ue_name, name, PNAME_LEN);
    ue->ue_name[PNAME_LEN] = (char)0;

    strncpy(ue->ue_version, version, PVER_LEN);
    ue->ue_version[PVER_LEN] = (char)0;

    _entryPrepend(ue, to);
  }

  return ue;
}


userEntry *
_entryPrepend(userEntry *ue, userEntryList *to)
{
  if(ue) {
    ue->ue_prev = NULL;
    ue->ue_next = to->ul_head;
    to->ul_head = (!to->ul_head ? (to->ul_tail = ue) :
                   (to->ul_head->ue_prev = ue));
  }

  return ue;
}


userEntry *
_entryFind(const char *name, userEntryList *list)
{
  userEntry *ue;

  for(ue = list->ul_head; ue; ue = ue->ue_next) {
    if(!strcasecmp(ue->ue_name, name)) {
      return ue;
    }
  }

  return NULL;
}


userEntry *
_entryFindVersion(const char *version, userEntryList *list)
{
  userEntry *ue;

  for(ue = list->ul_head; ue; ue = ue->ue_next) {
    if(!strcasecmp(ue->ue_version, version)) {
      return ue;
    }
  }

  return NULL;
}


userEntry *
_entryFindTwo(const char *name, const char *version, userEntryList *list)
{
  userEntry *ue;

  for(ue = list->ul_head; ue; ue = ue->ue_next) {
    if(!strcasecmp(name, ue->ue_name) && !strcasecmp(version, ue->ue_version)){
      return ue;
    }
  }

  return NULL;
}


userEntry *
_entryMove(userEntry *ue, userEntryList *from, userEntryList *to)
{
  if(_entryRemove(ue, from)) {
    return _entryPrepend(ue, to);
  }

  return NULL;
}


void
_entryMoveName(const char *name, userEntryList *from, userEntryList *to)
{
  userEntry *ue, *uetmp;

  for(ue = from->ul_head; ue; ) {
    if(!strcasecmp(ue->ue_name, name)) {
      uetmp = ue->ue_next;
      _entryMove(ue, from, to);
      ue = uetmp;
    } else {
      ue = ue->ue_next;
    }
  }
}


void
_entryMoveList(userEntryList *from, userEntryList *to)
{
  if(to->ul_head) {
    to->ul_head->ue_prev = from->ul_tail;
  }

  if(from->ul_tail) {
    from->ul_tail->ue_next = to->ul_head;
  }

  if(!to->ul_tail) {
    to->ul_tail = from->ul_tail;
  }

  if(from->ul_head) {
    to->ul_head = from->ul_head;
  }

  from->ul_head = from->ul_tail = NULL;
}


userEntry *
_entryRemove(userEntry *ue, userEntryList *from)
{
  if(!ue) {
    return NULL;
  }

  if(!ue->ue_prev) {
    if(from->ul_head == ue) {
      from->ul_head = ue->ue_next;
    }
  } else {
    if(ue->ue_prev->ue_next != ue) {
      return NULL;
    }

    ue->ue_prev->ue_next = ue->ue_next;
  }

  if(!ue->ue_next) {
    if(from->ul_tail == ue) {
      from->ul_tail = ue->ue_prev;
    }
  } else {
    if(ue->ue_next->ue_prev != ue) {
      return NULL;
    }

    ue->ue_next->ue_prev = ue->ue_prev;
  }

  ue->ue_next = ue->ue_prev = NULL;
  return ue;
}


void
_entryRemoveMatchVersion(userEntryList *l1, userEntryList *l2)
{
  userEntry *ue, *ue2, *uetmp;

  for(ue = l1->ul_head; ue; ue = ue->ue_next) {
    for(ue2 = l2->ul_head; ue2; ) {
     uetmp = ue2->ue_next;

      if(!strcasecmp(ue->ue_version, ue2->ue_version)) {
        FreeMem(_entryRemove(ue2, l2));
      }

      ue2 = uetmp;
    }
  }
}


void
_freeUserEntryList(userEntryList *list)
{
  userEntry *ue;

  for(ue = list->ul_head; ue; ue = list->ul_head) {
    list->ul_head = ue->ue_next;
    FreeMem(ue);
  }

  list->ul_head = list->ul_tail = NULL;
}


int
_entryLoad(const char *filename, userEntryList *to)
{
  char *p;
  FILE *file;
  time_t datet;
  unsigned int n, def;
  char buf[PNAME_LEN * 3 + PVER_LEN+4], name[PNAME_LEN+1], version[PVER_LEN+1];

  if(!(file = fopen(filename, "r"))) {
    return HICKERR_INSTFILE;
  }

  while(fgets(buf, (PNAME_LEN * 3 + PVER_LEN + 3), file)) {
    for(p = buf, n = 0; *p && (n < PNAME_LEN) && (*p != ':'); ++p){
      name[n++] = *p; /* getting the time(2) */
    }

    name[n] = (char)0;
    datet = atol(name);

    if(!*p) {
      continue;
    }

    for(p++, n = 0; *p && (n < PNAME_LEN) && (*p != ':'); ++p) {
      name[n++] = *p; /* getting the flags */
    }

    name[n] = (char)0;
    def = atoi(name);

    if(!*p) {
      continue;
    }

    p = _get_str_word(++p, ':', ':', name, PNAME_LEN);

    if(!*p) {
      continue;
    }

    _get_str_word(++p, ':', '\n', version, PVER_LEN);

    if(!_entryFindTwo(name, version, to)) {
      if(!_entryNew(name, version, datet, def, to)) {
        _freeUserEntryList(to);
        return HICKERR_OUTOFMEMORY;
      }
    }
  }

  fclose(file);
  return 0;
}


char *
_get_str_word(char *str, int delim, int endc, char *buf, size_t maxlen)
{
  size_t n;

  for(n = 0; *str && (*str != endc) && (n < maxlen); ++n, ++str) {
    if(*str == '\\') {
      if(!*++str) {
        break;
      }

      if((*str == delim) || (*str == '\\')) {
        buf[n] = *str;
      } else {
        if(*str == 'n') {
          buf[n] = '\n';
        } else {
          if((n + 1) < maxlen) {
            buf[n] = '\\';
            buf[++n] = *str;
          }
        }
      }
    } else {
      buf[n] = *str;
    }
  }

  buf[n] = (char)0;
  return str;
}


static void
removeEntry(const char *name, const char *version, int chkVer,
            userEntryList *from)
{
  userEntry *ue, *uetmp;

  for(ue = from->ul_head; ue; ) {
    if(!strcasecmp(name, ue->ue_name) &&
       (!chkVer || (chkVer && !strcasecmp(version, ue->ue_version)))) {
      uetmp = ue->ue_next;
      FreeMem(_entryRemove(ue, from));      
      ue = uetmp;
    } else {
      ue = ue->ue_next;
    }
  }
}


int
addPacket(HickPackets *pack, const char *name, const char *version, int chkVer,
          int def)
{
  removeEntry(name, version, chkVer, &pack->hp_remove);

  if((chkVer && _entryFindTwo(name, version, &pack->hp_install)) ||
     (!chkVer && _entryFind(name, &pack->hp_install))) {
    return 0;
  }

  _entryNew(name, version, 0, (def ? UE_DEFAULT : UE_NONE), &pack->hp_install);
  return 1;
}


int
removePacket(HickPackets *pack, const char *name, const char *version,
             int chkVer, int whole)
{
  removeEntry(name, version, chkVer, &pack->hp_install);

  if((chkVer && _entryFindTwo(name, version, &pack->hp_remove)) ||
     (!chkVer && _entryFind(name, &pack->hp_remove))) {
    return 0;
  }

  _entryNew(name, version, 0, (whole ? UE_RWHOLE : 0), &pack->hp_remove);
  return 1;
}
