/* HICK Lib 1.3.2 -- User Profiler Library.
** Copyright (C) 1997 - 2000 Daniel Bergstrom <daniel@octocode.com>
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the Free
** Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <pwd.h>
#include <grp.h>
#include <ctype.h>
#include <stdlib.h>

#if STDC_HEADERS
# include <string.h>
#else
# ifndef HAVE_MEMCPY
#  define memcpy(d, s, n) bcopy ((s), (d), (n))
# endif
#endif

#include "hicksh.h"
#include "common.h"
#include "getfile.h"
#include "hickenv.h"
#include "version.h"
#include "getpacket.h"

FileToken _tokens[] = {
  { "about",    TOKEN_ABOUT      },
  { "hidden",   TOKEN_HIDDEN     },
  { "required", TOKEN_REQUIRED   },
  { "multiple", TOKEN_MULTIPLE   },
  { "created",  TOKEN_CREATED    },
  { "modified", TOKEN_MODIFIED   },
  { "expire",   TOKEN_EXPIRE     },
  { "subgroup", TOKEN_SUBGROUP   },
  { "version",  TOKEN_VERSION    },
  { "allow",    TOKEN_ALLOW      },
  { "deny",     TOKEN_DENY       },
  { "force",    TOKEN_FORCE      },
  { "addabout", TOKEN_ADDABOUT   },
  { "info",     HED_ENV_INFO     },
  { "call",     HED_ENV_CALL     },
  { "conflict", HED_ENV_CONFLICT },
  { "create",   HED_ENV_CREATE   },
  { "env",      HED_ENV_ENV      },
  { "link",     HED_ENV_LINK     },
  { "menuitem", HED_ENV_MITEM    },
  { "require",  HED_ENV_REQUIRE  },
  { "script",   HED_ENV_SCRIPT   },
  { "xres",     HED_ENV_XRES     },
  { "html",     HED_ENV_HTML     },
  { NULL, 0 }
};


int
_token_case_cmp(const char *s1, const char *s2)
{
  int c1, c2;

  do {
    c1 = tolower((int)*s1);
    c2 = (int)*s2;

    if(!c1 || (c1 != c2)) {
      break;
    }

    ++s1;
    ++s2;
  } while(1);

  return (c1 - c2);
}


int
_token_lookup(const char *s, FileToken *tokens)
{
  size_t n;

  for(n = 0; tokens[n].name; ++n) {
    if(!_token_case_cmp(s, tokens[n].name)) {
      return tokens[n].token;
    }
  }

  return -1;
}


int
_getPacketHead(FILE *file, HickEntry *result)
{
  int c;
  char *p;

  while((c = _getSpaces(file)) == '#') {
    _getComment(file);
  }

  if(c == EOF) {
    return 0;
  }

  ungetc(c, file);

  do { /* read the group(s). */
    if(!(p = _getString(file))) {
      break;
    }

    hsh_array_add(p, 0, &result->he_groups);
  } while(((c = _getSpaces(file)) != EOF) && (c == '|'));

  ungetc(c, file);

  if(!p || (c == EOF) || !(p = _getString(file)) || (_getSpaces(file) != '{')){
    hsh_array_remove(&result->he_groups);
    return 0;
  }

  strncpy(result->he_name, p, PNAME_LEN);
  result->he_name[PNAME_LEN+1] = 0;

  FreeMem(p);
  return 1;
}


void
_getPacket(FILE *file, const char *type, HickPacket *hpac, int sabout)
{
  int i, c;
  size_t defpri;
  HickSubGroup *sub;
  HickEnvList *elist;
  char buf[TMPBUFSIZ + 1];

  while(((c = _getSpaces(file)) != EOF) && (c != '}')) {
    if(c == '#') {
      _getComment(file);
      continue;
    }

    ungetc(c, file);

    if(!_getName(file, buf, TMPBUFSIZ)) {
      break;
    }

    switch((i = _token_lookup(buf, _tokens))) {
      case TOKEN_ABOUT: { /* about */
        if(!sabout) {
          _readString(file, 1);
          continue;
        }

        if(hpac->pac_about) { /* clear off old abouts.. */
          FreeMem(hpac->pac_about);
        }

        hpac->pac_about = _getString(file);
        continue;
      }
      case TOKEN_HIDDEN: { /* hidden */
        hpac->pac_hidden = 1;
        continue;
      }
      case TOKEN_REQUIRED: { /* required */
        hpac->pac_required = 1;
        continue;
      }
      case TOKEN_MULTIPLE: { /* multiple */
        hpac->pac_multiple = 1;
        continue;
      }
      case TOKEN_CREATED: { /* created */
        hpac->pac_created = _getTime(file);
        continue;
      }
      case TOKEN_MODIFIED: { /* modified */
        hpac->pac_modified = _getTime(file);
        continue;
      }
      case TOKEN_EXPIRE: { /* expire */
        hpac->pac_expire = _getTime(file);
        continue;
      }
      case TOKEN_SUBGROUP: { /* subGroup */
        if((sub = _getSubgroup(file, 1))) {
          sub->hsg_next = (hpac->pac_subGroups ? hpac->pac_subGroups : 0);
          hpac->pac_subGroups = sub;
        }
        continue;
      }
      case TOKEN_VERSION: { /* version */
        _getVersion(&hpac->pac_versions, file, 1, type, sabout);
        continue;
      }
      case TOKEN_ALLOW: { /* allow */
        _getDirection(file, &hpac->pac_allow, 0);
        continue;
      }
      case TOKEN_DENY: { /* deny */
        _getDirection(file, &hpac->pac_deny, 0);
        continue;
      }
      case TOKEN_FORCE: { /* force */
        _getDirection(file, &hpac->pac_force, 1);
        continue;
      }
    }

    if((i < HED_ENV_INFO) || (i > HED_ENV_HTML)) {
      continue;
    }

    defpri = 0;
    elist = NULL;

    switch(i) {      
      case HED_ENV_INFO: { /* info */
        elist = &hpac->pac_info;
        break;
      }
      case HED_ENV_CALL: { /* Call */
        elist = &hpac->pac_call;
        break;
      }
      case HED_ENV_CONFLICT: { /* Conflict */
        elist = &hpac->pac_conflict;
        break;
      }
      case HED_ENV_CREATE: { /* Create */
        elist = &hpac->pac_create;
        break;
      }
      case HED_ENV_ENV: { /* Env */
        elist = &hpac->pac_env;
        defpri = HICKPRI_ENVIRON;
        break;
      }
      case HED_ENV_LINK: { /* Link */
        elist = &hpac->pac_link;
        break;
      }
      case HED_ENV_MITEM: { /* MenuItem */
        elist = &hpac->pac_mitem;
        defpri = HICKPRI_MENUITEM;
        break;
      }
      case HED_ENV_REQUIRE: { /* Require */
        elist = &hpac->pac_require;
        break;
      }
      case HED_ENV_SCRIPT: { /* Script */
        elist = &hpac->pac_script;
        defpri = HICKPRI_SCRIPTS;
        break;
      }
      case HED_ENV_XRES: { /* Xres */
        elist = &hpac->pac_xres;
        break;
      }
      case HED_ENV_HTML: { /* HTML */
       elist = &hpac->pac_html;
      }
    }

    if(elist) {
      _getEnv(file, i, elist, type, defpri);
    }
  }
}


HickDirection *
_getDirection(FILE *file, HickDirection **head, int force)
{
  char *s;
  HickDirection hd, *ret;

  if(!(s = _getString(file))) {
    return 0;
  }

  hd.hd_next = NULL;
  hd.hd_group = 0;

  hsh_array_init(&hd.hd_names);
  hsh_array_init(&hd.hd_versions);

  if(_token_case_cmp(s, "user")) {
    if(_token_case_cmp(s, "group")) {
      if(!force) { /* If it's not force we got an error. */
        return 0;
      }

      hsh_array_add(s, 0, &hd.hd_versions);
    }

    hd.hd_group = 1;
  }

  if(!hd.hd_versions.siz) {
    FreeMem(s); /* we found a user or group.. */
    _getList(file, &hd.hd_names);

    if(force) {
      if(!(s = _getString(file))) {
        hsh_array_remove(&hd.hd_names);
        return 0;
      }

      hsh_array_add(s, 0, &hd.hd_versions);
    } else {
      _getList(file, &hd.hd_versions);
    }
  }

  /* User or Group names have to be specified else won't we use the
   * information. This dosen't apply to force.
   */

  if(!force && !hd.hd_names.siz) {
    hsh_array_remove(&hd.hd_names);
    hsh_array_remove(&hd.hd_versions);
    return 0;
  }

  /* Allocate the HickDirection stucture and copy over the information */

  ret = MallocMem(HickDirection);

  memcpy(ret, &hd, sizeof(HickDirection));

  ret->hd_next = *head;
  *head = ret;

  return ret;
}


HickSubGroup *
_getSubgroup(FILE *file, int ongroup)
{
  HickSubGroup *sub;
  char c, *ong, *name;

  if(!(sub = MallocMem(HickSubGroup))) {
    return 0;
  }

  sub->hsg_group[0] = sub->hsg_name[0] = 0;
  sub->hsg_down = sub->hsg_next = 0;

  hsh_array_init(&sub->hsg_versions);

  if(ongroup) {
    if(!(ong = _getString(file))) { /* on group */
      return 0;
    }

    strncpy(sub->hsg_group, ong, PGROUP_LEN);
    sub->hsg_group[PGROUP_LEN] = (char)0;
    FreeMem(ong);
  }

  if(!(name = _getString(file))) { /* subgroup name */
    return 0;
  }

  strncpy(sub->hsg_name, name, PGROUP_LEN);
  sub->hsg_name[PGROUP_LEN] = 0;
  FreeMem(name);

  c = _getSpaces(file);

  if(c == '{') {
    ungetc(c, file);
    _getList(file, &sub->hsg_versions);
    c = _getSpaces(file);
  }

  if(c == '&') {
    sub->hsg_down = _getSubgroup(file, 0);
  } else {
    ungetc(c, file);
  }

  return sub;
}

/* return values:
**
**   0 - failed
**   1 - deny all
**   2 - got access
*/

int
_checkDirection(HickDirection *head, int flag, DoubleArray *arr)
{
  HickDirection *hd;
  struct group *grp;
  struct passwd *pwd;
  int i, j, found, alldeny = 0;

  /* get the user account */
  if(!(pwd = getpwuid(geteuid())) || !pwd->pw_name) {
    return 0;
  }

  initDoubleArray(arr);

  for(hd = head; hd; hd = hd->hd_next) {
    if((flag == PERM_FORCE) && !hd->hd_names.siz) {
      hsh_array_add(hd->hd_versions.arr[0], 1, &arr->arr1);
      continue;
    }

    found = 0;

    if(!hd->hd_group) {
      /* Look and see if we can find the user in the list of names. */

      for(i = 0; i < hd->hd_names.siz; i++) {
        if(!strcasecmp(hd->hd_names.arr[i], pwd->pw_name)) {
          found = 1; /* user was found! */
          break;
        }
      }
    } else {
      /* Look in the list of group names and see if the user in it. */

      for(i = 0; (i < hd->hd_names.siz) && !found; i++) {
        if(!(grp = getgrnam(hd->hd_names.arr[i])) || !grp->gr_mem) {
          continue; /* didn't find that group. */
        }

        if(grp->gr_gid == pwd->pw_gid) { /* first we check the gid. */
          found = 1;
          break;
        }

        for(j = 0; grp->gr_mem[j]; j++) {
          if(!strcasecmp(grp->gr_mem[j], pwd->pw_name)) {
            found = 1;
            break;
          }
        }
      }
    }

    if(found) {
      /* Check if we got access to all version. If we have return that. */

      if(((flag != PERM_FORCE) && !hd->hd_versions.siz) ||
         ((flag == PERM_FORCE) && !*(hd->hd_versions.arr[0]))) {
        hsh_array_remove(&arr->arr1);
        hsh_array_remove(&arr->arr2);
        return 2;
      }

      if(flag == PERM_FORCE) {
        hsh_array_add(hd->hd_versions.arr[0], 1, &arr->arr1);
      } else {
        for(i = 0; i < hd->hd_versions.siz; i++) {
          hsh_array_add(hd->hd_versions.arr[i], 1, &arr->arr1);
        }
      }
    } else {
      if((flag != PERM_ALLOW) || alldeny) {
        continue;
      }

      if(!hd->hd_versions.siz) {
        hsh_array_remove(&arr->arr2);
        alldeny = 1;
        continue;
      }

      for(i = 0; i < hd->hd_versions.siz; i++) {
        hsh_array_add(hd->hd_versions.arr[i], 1, &arr->arr2);
      }
    }
  }

  return (!arr->arr2.arr && !alldeny ? 2 : 1);
}


int
_getList(FILE *file, ArrayInfo *arri)
{
  int c;
  char *p;

  hsh_array_init(arri);

  if((c = _getSpaces(file)) == '{') { /* read list */
    while(((c = _getSpaces(file)) != EOF) && (c != '}')) {
      if(c == '#') {
        _getComment(file);
        continue;
      }

      ungetc(c, file);

      if(!(p = _getString(file))) {
        return 0; /* something is wrong, better exit here. */
      }

      hsh_array_add(p, 0, arri);
    }
  } else { /* read single */
    ungetc(c, file);

    if(!(p = _getString(file))) {
      return 0; /* Something is wrong, better exit here. */
    }

    if(*p) { /* Don't add empty words */
      hsh_array_add(p, 0, arri);
    } else {
      FreeMem(p);
    }
  }

  return 1;
}


void
initHickPacket(HickPacket *hpac)
{
  hpac->pac_name[0] = 0;
  hpac->pac_about = NULL;
  hpac->pac_hidden = hpac->pac_required = hpac->pac_multiple = 0;
  hpac->pac_created = hpac->pac_modified = hpac->pac_expire = 0;
  hpac->pac_allow = hpac->pac_deny = hpac->pac_force = NULL;
  hpac->pac_subGroups = NULL;

  hsh_array_init(&hpac->pac_groups);

  initHickVerList(&hpac->pac_versions);

  initHickEnvList(&hpac->pac_info);
  initHickEnvList(&hpac->pac_call);
  initHickEnvList(&hpac->pac_conflict);
  initHickEnvList(&hpac->pac_create);
  initHickEnvList(&hpac->pac_env);
  initHickEnvList(&hpac->pac_link);
  initHickEnvList(&hpac->pac_mitem);
  initHickEnvList(&hpac->pac_require);
  initHickEnvList(&hpac->pac_script);
  initHickEnvList(&hpac->pac_xres);
  initHickEnvList(&hpac->pac_html);
}


HickPacket *
freeHickPacket(HickPacket *hpac)
{
  freeHickEnvList(&hpac->pac_info,     HED_ENV_INFO);
  freeHickEnvList(&hpac->pac_call,     HED_ENV_CALL);
  freeHickEnvList(&hpac->pac_conflict, HED_ENV_CONFLICT);
  freeHickEnvList(&hpac->pac_create,   HED_ENV_CREATE);
  freeHickEnvList(&hpac->pac_env,      HED_ENV_ENV);
  freeHickEnvList(&hpac->pac_link,     HED_ENV_LINK);
  freeHickEnvList(&hpac->pac_mitem,    HED_ENV_MITEM);
  freeHickEnvList(&hpac->pac_require,  HED_ENV_REQUIRE);
  freeHickEnvList(&hpac->pac_script,   HED_ENV_SCRIPT);
  freeHickEnvList(&hpac->pac_xres,     HED_ENV_XRES);
  freeHickEnvList(&hpac->pac_html,     HED_ENV_HTML);

  freeHickVerList(&hpac->pac_versions, 1);

  if(hpac->pac_about) {
    FreeMem(hpac->pac_about);
  }

  hsh_array_remove(&hpac->pac_groups);

  freeHickDirection(hpac->pac_deny);
  freeHickDirection(hpac->pac_allow);
  freeHickDirection(hpac->pac_force);

  freeHickSubGroup(hpac->pac_subGroups);

  initHickPacket(hpac);
  return hpac;
}


void
freeHickDirection(HickDirection *head)
{
  HickDirection *hd;

  for(hd = head; hd; hd = head) {
    head = hd->hd_next;

    hsh_array_remove(&hd->hd_names);
    hsh_array_remove(&hd->hd_versions);
    FreeMem(hd);
  }
}


void
freeHickSubGroup(HickSubGroup *head)
{
  HickSubGroup *sub;

  for(sub = head; sub; sub = head) {
    head = sub->hsg_next;

    hsh_array_remove(&sub->hsg_versions);

    if(sub->hsg_down) {
      freeHickSubGroup(sub->hsg_down);
    }

    FreeMem(sub);
  }
}
