/* HICK Lib 1.3.2 -- User Profiler Library.
** Copyright (C) 1997 - 2000 Daniel Bergstrom <daniel@octocode.com>
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the Free
** Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "common.h"
#include "getfile.h"
#include "hickenv.h"
#include "version.h"
#include "getpacket.h"

#define DIRECT_FORCE 0
#define DIRECT_ALLOW 1
#define DIRECT_DENY  2

static void mergeHickEnv(HickEnvList *env, int envtype, ArrayInfo *types);

static void writeHickEnv(HickEnvList *list, int type, const char *s,
                         const char *espaces, const char *hicktype,
                         Buffert *buf);
static void writeSubGroup(HickSubGroup *head, Buffert *buf);
static void writeArray(ArrayInfo *arri, Buffert *buf);
static void writeDate(const char *s, time_t t, int tlevel, Buffert *buf);
static void writeHickDirection(HickDirection *head, int type, Buffert *buf);


int
loadPacket(const char *name, const char *packpath, const char *hicktype,
           ArrayInfo *types, HickPacket *result, int sabout)
{
  FILE *file;
  HickEntry entry;
  struct stat sbuf;
  DIR *typedir, *packdir;
  unsigned int len, found = 0;
  char fname[FNAME_LEN+1], *type;
  struct dirent *tentry, *pentry;

  if(!(packdir = hsh_opendir(packpath, NULL))) {
    return HICKERR_PACKDIR;
  }

  initHickEntry(&entry);

  while((tentry = readdir(packdir))) {
    if(IS_DOT_OR_DOTDOT(tentry)) {
      continue;
    }

    if(hicktype && strcasecmp(tentry->d_name, COMMON_DIR) &&
       strcasecmp(tentry->d_name, hicktype)) {
      continue; /* didn't match the type or was common dir */
    }

    hsh_create_str(fname, FNAME_LEN, packpath, "/", tentry->d_name, NULL);

    if(!(typedir = hsh_opendir(fname, NULL))) {
      continue;
    }

    while((pentry = readdir(typedir))) {
      hsh_create_str(fname, FNAME_LEN, packpath, "/", tentry->d_name, "/",
                     pentry->d_name, NULL);

      if((hsh_stat(fname, &sbuf, NULL) ==IOW_FAIL) || !S_ISREG(sbuf.st_mode)) {
        continue;
      }

      len = NAMLEN(pentry) - 1;

      if(IS_EMACS_BACKUP(pentry, len) || !(file = fopen(fname, "r"))) {
        continue;
      }

      while(_getPacketHead(file, &entry)) {
        if(strcasecmp(entry.he_name, name)) {
          _readStmt(file, 0); /* trash the contents of the packet */
          freeHickEntry(&entry);
          continue;
        }

        if(++found == 1) { /* first packet? copy over header stuff */
          strcpy(result->pac_name, entry.he_name);

          result->pac_groups.arr = entry.he_groups.arr;
          result->pac_groups.siz = entry.he_groups.siz;
          result->pac_groups.alloc = entry.he_groups.alloc;

          hsh_array_init(&entry.he_groups);
        }

        freeHickEntry(&entry);

        if(!hicktype && strcasecmp(tentry->d_name, COMMON_DIR)) {
          type = tentry->d_name;
        } else {
          type = NULL;
        }

        _getPacket(file, type, result, sabout);

        if(hicktype) { /* that's it, we got one. */
          fclose(file);
          closedir(typedir);
          closedir(packdir);
          return 0;
        }
      }

      fclose(file);
    }

    closedir(typedir);
  }

  closedir(packdir);

  if(result->pac_name[0] == (char)0) {
    return HICKERR_NOTFOUND;
  }

  mergeHickPacket(result, types);
  return 0;
}


void
mergeHickPacket(HickPacket *hpac, ArrayInfo *types)
{
  unsigned int n;
  HickVer *ver, *vold, *vtmp;

  /* Merging versions, we assume that the contents on any of the verions
  ** is the same, but the variables. So we copy over the variables, frees
  ** the version and gets the next one. We also checks the types, and adds
  ** one if needed.
  */

  if(hpac->pac_versions.vl_head && hpac->pac_versions.vl_head->ver_next) {
    for(vold = hpac->pac_versions.vl_head; vold && vold->ver_next;
        vold = vold->ver_next) {

      for(ver = vold->ver_next; ver; ver = vtmp) {
        vtmp = ver->ver_next;

        if(strcasecmp(ver->ver_number, vold->ver_number)) {
          continue; /* didn't match */
        }

        _envMoveList(&HVER_INFO(ver),     &HVER_INFO(vold));
        _envMoveList(&HVER_CALL(ver),     &HVER_CALL(vold));
        _envMoveList(&HVER_CONFLICT(ver), &HVER_CONFLICT(vold));
        _envMoveList(&HVER_CREATE(ver),   &HVER_CREATE(vold));
        _envMoveList(&HVER_ENV(ver),      &HVER_ENV(vold));
        _envMoveList(&HVER_LINK(ver),     &HVER_LINK(vold));
        _envMoveList(&HVER_MITEM(ver),    &HVER_MITEM(vold));
        _envMoveList(&HVER_REQUIRE(ver),  &HVER_REQUIRE(vold));
        _envMoveList(&HVER_SCRIPT(ver),   &HVER_SCRIPT(vold));
        _envMoveList(&HVER_XRES(ver),     &HVER_XRES(vold));
        _envMoveList(&HVER_HTML(ver),     &HVER_HTML(vold));

        if(HVER_TYPES(vold).siz && HVER_TYPES(ver).siz) {
          for(n = 0; n < HVER_TYPES(ver).siz; n++) {
            hsh_array_add(HVER_TYPES(ver).arr[n], 1, &HVER_TYPES(vold));
          }
        } else {
          if(HVER_TYPES(vold).siz) {
            hsh_array_remove(&HVER_TYPES(vold));
          }
        }

        FreeMem(freeHickVer(_versionRemove(ver, &hpac->pac_versions), 1));
      }

      if(HVER_TYPES(vold).siz) {
        for(n = 0; n < types->siz; n++) {
          if(!hsh_array_find(types->arr[n], &HVER_TYPES(vold))) {
            break;
          }
        }

        if(n >= types->siz) {
          hsh_array_remove(&HVER_TYPES(vold));
        }
      }
    }
  }

  for(ver = hpac->pac_versions.vl_head; ver; ver = ver->ver_next) {
    mergeHickEnv(&HVER_INFO(ver), HED_ENV_INFO, types);
    mergeHickEnv(&HVER_CALL(ver), HED_ENV_CALL, types);
    mergeHickEnv(&HVER_CONFLICT(ver), HED_ENV_CONFLICT, types);
    mergeHickEnv(&HVER_CREATE(ver), HED_ENV_CREATE, types);
    mergeHickEnv(&HVER_ENV(ver), HED_ENV_ENV, types);
    mergeHickEnv(&HVER_LINK(ver), HED_ENV_LINK, types);
    mergeHickEnv(&HVER_MITEM(ver), HED_ENV_MITEM, types);
    mergeHickEnv(&HVER_REQUIRE(ver), HED_ENV_REQUIRE, types);
    mergeHickEnv(&HVER_SCRIPT(ver), HED_ENV_SCRIPT, types);
    mergeHickEnv(&HVER_XRES(ver), HED_ENV_XRES, types);
    mergeHickEnv(&HVER_HTML(ver), HED_ENV_HTML, types);
  }

  mergeHickEnv(&hpac->pac_info, HED_ENV_INFO, types);
  mergeHickEnv(&hpac->pac_call, HED_ENV_CALL, types);
  mergeHickEnv(&hpac->pac_conflict, HED_ENV_CONFLICT, types);
  mergeHickEnv(&hpac->pac_create, HED_ENV_CREATE, types);
  mergeHickEnv(&hpac->pac_env, HED_ENV_ENV, types);
  mergeHickEnv(&hpac->pac_link, HED_ENV_LINK, types);
  mergeHickEnv(&hpac->pac_mitem, HED_ENV_MITEM, types);
  mergeHickEnv(&hpac->pac_require, HED_ENV_REQUIRE, types);
  mergeHickEnv(&hpac->pac_script, HED_ENV_SCRIPT, types);
  mergeHickEnv(&hpac->pac_xres, HED_ENV_XRES, types);
  mergeHickEnv(&hpac->pac_html, HED_ENV_HTML, types);
}


static void
mergeHickEnv(HickEnvList *list, int envtype, ArrayInfo *types)
{
  unsigned int n;
  HickEnvItem *eit, *itmp;
  HickEnv *env, *eold, *etmp;
  HickRequirePacket *rpac, *rtmp;

  if(list->el_head && list->el_head->env_next) {
    for(eold = list->el_head; eold && eold->env_next; eold = eold->env_next) {
      for(env = eold->env_next; env; env = etmp) {
        etmp = env->env_next;

        switch(envtype) {
          case HED_ENV_INFO: {
            if(strcasecmp(eold->env_name, env->env_name) ||
               strcasecmp(eold->env_value, env->env_value) ||
               strcasecmp(eold->env_itemTitle, env->env_itemTitle) ||
               strcasecmp(eold->env_about, env->env_about) ||
               strcasecmp(eold->u.env_topic, env->u.env_topic)) {
              continue;
            }
            break;
          }
          case HED_ENV_HTML: {
            if(strcasecmp(eold->env_name, env->env_name) ||
               strcasecmp(eold->env_value, env->env_value)) {
              continue;
            }
            break;
          }
          case HED_ENV_CALL: {
            if((eold->env_flags != env->env_flags) ||
               strcasecmp(eold->env_name, env->env_name) ||
               strcasecmp(eold->env_value, env->env_value)) {
              continue;
            }
            break;
          }
          case HED_ENV_CONFLICT: {
            if(strcasecmp(eold->env_name, env->env_name) ||
               (eold->env_flags != env->env_flags) ||
               !hsh_array_match(env->u.env_versions, eold->u.env_versions)) {
              continue;
            }
            break;
          }
          case HED_ENV_CREATE: {
            if((eold->env_flags != env->env_flags) ||
               (eold->env_priority != env->env_priority) ||
               strcasecmp(eold->env_name, env->env_name) ||
               strcasecmp(eold->env_value, env->env_value)) {
              continue;
            }
            break;
          }
          case HED_ENV_ENV: {
            if((eold->env_flags != env->env_flags) ||
               (eold->env_priority != env->env_priority)) {
              continue;
            }

            if((eold->env_flags != HFLAG_ENV_UNSET) &&
               (strcasecmp(eold->env_name, env->env_name) ||
                strcasecmp(eold->env_value, env->env_value))) {
              continue;
            } else {
              if((eold->env_flags ==  HFLAG_ENV_UNSET) &&
                 strcasecmp(eold->env_name, env->env_name)) {
                continue;
              }
            }
            break;
          }
          case HED_ENV_LINK: {
            if((eold->env_flags != env->env_flags) ||
               strcasecmp(eold->env_name, env->env_name) ||
               strcasecmp(eold->env_value, env->env_value)) {
              continue;
            }
            break;
          }
          case HED_ENV_MITEM: {
            if((eold->env_priority != env->env_priority) ||
               strcasecmp(eold->env_name, env->env_name) ||
               strcasecmp(eold->env_value, env->env_value) ||
               strcasecmp(eold->env_itemTitle, env->env_itemTitle)) {
              continue;
            }

            if(!eold->u.env_items && !env->u.env_items) {
              break;
            }

            for(eit = eold->u.env_items; eit; eit = eit->eit_next) {
              for(itmp = env->u.env_items; itmp; itmp = itmp->eit_next) {
                if(!strcasecmp(eit->eit_name, itmp->eit_name) &&
                   !strcasecmp(eit->eit_value, itmp->eit_value)) {
                  break; /* found match */
                }
              }

              if(!itmp) {
                break; /* didn't match */
              }
            }

            if(eit) {
              continue; /* didn't match */
            }
            break;
          }
          case HED_ENV_REQUIRE: {
            for(rpac = eold->u.env_rpacs; rpac; rpac = rpac->rpac_or) {
              for(rtmp = env->u.env_rpacs; rtmp; rtmp = rtmp->rpac_or) {
                if(!strcasecmp(rpac->rpac_name, rtmp->rpac_name) &&
                   hsh_array_match(&rpac->rpac_versions,
                                   &rtmp->rpac_versions)){
                  break; /* found match */
                }
              }

              if(!rtmp) {
                break; /* didn't match */
              }
            }

            if(rpac) { /* didn't match */
              continue;
            }
            break;
          }
          case HED_ENV_SCRIPT: {
            if((eold->env_priority != env->env_priority) ||
               strcasecmp(eold->env_name, env->env_name) ||
               strcasecmp(eold->env_value, env->env_value)) {
              continue;
            }
            break;
          }
          case HED_ENV_XRES: {
            if(eold->env_flags != env->env_flags) {
              continue;
            }

            if((eold->env_flags == HFLAG_XRES_INLINE) &&
               (strcasecmp(eold->env_name, env->env_name) ||
                strcasecmp(eold->env_value, env->env_value))) {
              continue;
            } else {
              if((eold->env_flags == HFLAG_XRES_FILENAME) &&
                 strcasecmp(eold->env_name, env->env_name)) {
                continue;
              }
            }
            break;
          }
        }

        if(eold->env_types.siz && env->env_types.siz) {
          for(n = 0; n < env->env_types.siz; n++) {
            hsh_array_add(env->env_types.arr[n], 0, &eold->env_types);
          }

          FreeMem(env->env_types.arr);
          hsh_array_init(&env->env_types);
        } else {
          if(eold->env_types.siz) {
            hsh_array_remove(&eold->env_types);
          }
        }

        FreeMem(freeHickEnv(_envRemove(env, list), 1));
      }

      if(eold->env_types.siz) {
        for(n = 0; n < types->siz; n++) {
          if(!hsh_array_find(types->arr[n], &eold->env_types)) {
            break;
          }
        }

        if(n >= types->siz) {
          hsh_array_remove(&eold->env_types);
        }
      }
    }
  }
}


void
savePacket(FILE *file, HickPacket *hpac, const char *hicktype, int tlevel)
{
  Buffert buf;
  HickVer *ver;
  unsigned int n;

  buf.pos = 0;
  buf.file = file;

  bput('\n', &buf);
  
  for(n = 0; n < hpac->pac_groups.siz; n++) {
    bprint(&buf, "%S%s", hpac->pac_groups.arr[n],
           ((n + 1) < hpac->pac_groups.siz ? " | " : " "));
  }

  bprint(&buf, "%S {\n", hpac->pac_name);

  if(hpac->pac_about) {
    bprint(&buf, "  about %S\n", hpac->pac_about);
  }

  if(hpac->pac_hidden) {
    _bputs("  hidden\n", 9, &buf);
  }

  if(hpac->pac_required) {
    _bputs("  required\n", 11, &buf);
  }

  if(hpac->pac_multiple) {
    _bputs("  multiple\n", 11, &buf);
  }

  writeDate("  created", hpac->pac_created, tlevel, &buf);
  writeDate("  modified", hpac->pac_modified, tlevel, &buf);
  writeDate("  expire", hpac->pac_expire, tlevel, &buf);

  writeSubGroup(hpac->pac_subGroups, &buf);

  writeHickDirection(hpac->pac_force, DIRECT_FORCE, &buf);
  writeHickDirection(hpac->pac_allow, DIRECT_ALLOW, &buf);
  writeHickDirection(hpac->pac_deny,  DIRECT_DENY, &buf);

  writeHickEnv(&hpac->pac_info, HED_ENV_INFO, "  info", "  ", hicktype, &buf);
  writeHickEnv(&hpac->pac_call, HED_ENV_CALL, "  call", "  ", hicktype, &buf);
  writeHickEnv(&hpac->pac_conflict, HED_ENV_CONFLICT, "  conflict", "  ",
               hicktype, &buf);
  writeHickEnv(&hpac->pac_create, HED_ENV_CREATE, "  create", "  ", hicktype,
               &buf);
  writeHickEnv(&hpac->pac_env, HED_ENV_ENV, "  env", "  ", hicktype, &buf);
  writeHickEnv(&hpac->pac_link, HED_ENV_LINK, "  link", "  ", hicktype, &buf);
  writeHickEnv(&hpac->pac_mitem, HED_ENV_MITEM, "  menuItem", "  ", hicktype,
               &buf);
  writeHickEnv(&hpac->pac_require, HED_ENV_REQUIRE, "  require", "  ",
               hicktype, &buf);
  writeHickEnv(&hpac->pac_script, HED_ENV_SCRIPT, "  script", "  ", hicktype,
               &buf);
  writeHickEnv(&hpac->pac_xres, HED_ENV_XRES, "  xres", "  ", hicktype, &buf);
  writeHickEnv(&hpac->pac_html, HED_ENV_HTML, "  html", "  ", hicktype, &buf);

  for(ver = hpac->pac_versions.vl_head; ver; ver = ver->ver_next) {
    if(hicktype && HVER_TYPES(ver).arr &&
       !hsh_array_find(hicktype, &HVER_TYPES(ver))) {
      continue;
    }

    bprint(&buf, "\n  version %s%S {\n", (ver->ver_default ? "def " : ""),
           ver->ver_number);

    if(ver->ver_about) {
      bprint(&buf, "    addAbout %s%S\n",
             (ver->ver_override ? "override " : ""), ver->ver_about);
    }

    writeDate("    created", ver->ver_created, tlevel, &buf);
    writeDate("    modified", ver->ver_modified, tlevel, &buf);
    writeDate("    expire", ver->ver_expire, tlevel, &buf);

    writeHickEnv(&HVER_INFO(ver), HED_ENV_INFO, "    info", "    ", hicktype,
                 &buf);
    writeHickEnv(&HVER_CALL(ver), HED_ENV_CALL, "    call", "    ", hicktype,
                 &buf);
    writeHickEnv(&HVER_CONFLICT(ver), HED_ENV_CONFLICT, "    conflict",
                 "    ", hicktype, &buf);
    writeHickEnv(&HVER_CREATE(ver), HED_ENV_CREATE, "    create", "    ",
                 hicktype, &buf);
    writeHickEnv(&HVER_ENV(ver), HED_ENV_ENV, "    env", "    ", hicktype,
                 &buf);
    writeHickEnv(&HVER_LINK(ver), HED_ENV_LINK, "    link", "    ", hicktype,
                 &buf);
    writeHickEnv(&HVER_MITEM(ver), HED_ENV_MITEM, "    menuItem", "    ",
                 hicktype, &buf);
    writeHickEnv(&HVER_REQUIRE(ver), HED_ENV_REQUIRE, "    require", "    ",
                 hicktype, &buf);
    writeHickEnv(&HVER_SCRIPT(ver), HED_ENV_SCRIPT, "    script", "    ",
                 hicktype, &buf);
    writeHickEnv(&HVER_XRES(ver), HED_ENV_XRES, "    xres", "    ", hicktype,
                 &buf);
    writeHickEnv(&HVER_HTML(ver), HED_ENV_HTML, "    html", "    ", hicktype,
                 &buf);

    _bputs("  }\n", 4, &buf);
  }

  _bputs("}\n", 2, &buf);
  bflush(&buf); /* flush the buffert */
}


static void
writeHickEnv(HickEnvList *list, int type, const char *s, const char *espaces,
             const char *hicktype, Buffert *buf)
{
  HickEnv *env;
  const char *p;
  HickEnvItem *eit;
  HickRequirePacket *rpac;
  unsigned int f, havelist;

  if(!list->el_head) {
    return;
  }

  for(havelist = 0, env = list->el_head; env; env = env->env_next) {
    if(hicktype && env->env_types.arr &&
       !hsh_array_find(hicktype, &env->env_types)) {
      continue;
    }

    havelist++;
  }

  switch(havelist) {
    case 0: {
      return;
    }
    case 1: {
      bprint(buf, "%s", s);
      break;
    }
    default: {
      bprint(buf, "%s {", s);
    }
  }

  /* we go backwards with this to get it out as we read it. */
  for(env = list->el_tail; env; env = env->env_prev) {
    if(hicktype && env->env_types.arr &&
       !hsh_array_find(hicktype, &env->env_types)) {
      continue;
    }

    if(havelist != 1) {
      bprint(buf, "\n%s ", espaces);
    }

    switch(type) {
      case HED_ENV_INFO: {
        bprint(buf, " %S %S %S %S %S", env->env_name, env->env_value,
               env->env_itemTitle, env->env_about, env->u.env_topic);
        break;
      }
      case HED_ENV_HTML: {
        bprint(buf, " %S %S", env->env_name, env->env_value);
        break;
      }
      case HED_ENV_CALL: {
        f = env->env_flags;

        if(f == (HFLAG_CALL_INSTALL | HFLAG_CALL_UPDATE | HFLAG_CALL_REMOVE)) {
          if(!_token_case_cmp(env->env_name, "install") ||
             !_token_case_cmp(env->env_name, "update") ||
             !_token_case_cmp(env->env_name, "remove")) {
            _bputs(" install & update & remove", 27, buf);
          }
        } else {
          if(f & HFLAG_CALL_INSTALL) {
            f &= ~HFLAG_CALL_INSTALL;
            bprint(buf, " install%s", (f ? " &" : ""));
          }

          if(f & HFLAG_CALL_UPDATE) {
            f &= ~HFLAG_CALL_UPDATE;
            bprint(buf, " update%s", (f ? " &" : ""));
          }

          if(f & HFLAG_CALL_REMOVE) {
            f &= ~HFLAG_CALL_REMOVE;
            bprint(buf, " remove%s", (f ? " &" : ""));
          }
        }

        bprint(buf, " %S %S", env->env_name, env->env_value);
        break;
      }
      case HED_ENV_CONFLICT: {
        if(env->env_flags == 1) {
          bprint(buf, " %S %S", env->env_name, env->u.env_versions[0]);
        } else {
          bprint(buf, " %S", env->env_name);
          writeArray(env->u.env_versions, buf);
        }
        break;
      }
      case HED_ENV_CREATE: {
        if(env->env_flags == HFLAG_CREATE_APPEND) {
          bprint(buf, " append");
        }

        if(env->env_priority != HED_DEF_PERM) {
          bprint(buf, " %o", env->env_priority);
        }

        bprint(buf, " %S %S", env->env_name, env->env_value);
        break;
      }
      case HED_ENV_ENV: {
        if(env->env_flags == HFLAG_ENV_UNSET) {
          if(env->env_priority == HICKPRI_ENVIRON) {
            bprint(buf, " unset %S", env->env_name);
          } else {
            bprint(buf, " unset %d %S", env->env_priority, env->env_name);
          }
          break;
        }

        if(env->env_flags == HFLAG_ENV_PREPEND) {
          p = (!_token_case_cmp(env->env_name, "prepend") ? "prepend " : "");
        } else {
          p = (env->env_flags == HFLAG_ENV_APPEND ? "append " : "set ");
        }

        if(env->env_priority == HICKPRI_ENVIRON) {
          bprint(buf, " %s%S %S", p, env->env_name, env->env_value);
        } else {
          bprint(buf, " %s%d %S %S", p, env->env_priority, env->env_name,
                 env->env_value);
        }
        break;
      }
      case HED_ENV_LINK: {
        p = (env->env_flags == HFLAG_LINK_RENAME ? " rename" :
             (!_token_case_cmp(env->env_name, "normal") ? " normal" : ""));

        bprint(buf, "%s %S %S", p, env->env_name, env->env_value);
        break;
      }
      case HED_ENV_MITEM: {
        if(env->env_priority == HICKPRI_MENUITEM) {
          bprint(buf, " %S %S %S", env->env_name, env->env_itemTitle,
                 env->env_value);
        } else {
          bprint(buf, " %d %S %S %S", env->env_priority, env->env_name,
                 env->env_itemTitle, env->env_value);
        }

        if(env->u.env_items) {
          _bputs(" {\n", 3, buf);

          for(eit = env->u.env_items; eit; eit = eit->eit_next) {
            bprint(buf, "%s%s  %S %S\n", espaces, (havelist != 1 ? "  " : ""),
                   eit->eit_name, eit->eit_value);
          }

          bprint(buf, "%s%s}", espaces, (havelist != 1 ? "  " : ""));
        }
        break;
      }
      case HED_ENV_REQUIRE: {
        for(rpac = env->u.env_rpacs; rpac; rpac = rpac->rpac_or) {
          bprint(buf, " %S", rpac->rpac_name);

          switch((int)rpac->rpac_versions.siz) {
            case 0: {
              _bputs(" \"\"", 3, buf);
              break;
            }
            case 1: {
              bprint(buf, " %S", rpac->rpac_versions.arr[0]);
              break;
            }
            default: {
              writeArray(&rpac->rpac_versions, buf);
            }
          }

          if(rpac->rpac_or) {
            _bputs(" |", 2, buf);
          }
        }
        break;
      }
      case HED_ENV_SCRIPT: {
        if(env->env_priority != HICKPRI_SCRIPTS) {
          bprint(buf, " %o", env->env_priority);
        }

        bprint(buf, " %S %S", env->env_name, env->env_value);
        break;
      }
      case HED_ENV_XRES: {
        if(env->env_flags == HFLAG_XRES_FILENAME) {
          bprint(buf, " filename %S", env->env_name);
          break;
        }

        bprint(buf, " %s%S %S",
               (!_token_case_cmp(env->env_name, "inline") ? "inline " : ""),
               env->env_name, env->env_value);
        break;
      }
    }
  }

  if(havelist != 1) {
    bprint(buf, "\n%s}\n", espaces);
  } else {
    bput('\n', buf);
  }
}


static void
writeHickDirection(HickDirection *head, int type, Buffert *buf)
{
  HickDirection *hd;

  if(!head) {
    return;
  }

  for(hd = head; hd; hd = hd->hd_next) {
    bprint(buf, "\n  %s ", (type == DIRECT_FORCE ? "force" :
                            (type == DIRECT_ALLOW ? "allow" : "deny")));

    if((type == DIRECT_FORCE) && !hd->hd_names.siz) {
      bprint(buf, "%S", hd->hd_versions.arr[0]);
      continue;
    }

    if(hd->hd_group) {
      _bputs("group", 5, buf);
    } else {
      _bputs("user", 4, buf);
    }

    if(hd->hd_names.siz == 1) {
      bprint(buf, " %S", hd->hd_names.arr[0]);
    } else {
      writeArray(&hd->hd_names, buf);
    }

    if(type == DIRECT_FORCE) {
      bprint(buf, " %S", hd->hd_versions.arr[0]);
      continue;
    }

    if(hd->hd_versions.siz == 1) {
      bprint(buf, " %S", hd->hd_versions.arr[0]);
    } else {
      writeArray(&hd->hd_versions, buf);
    }
  }

  bput('\n', buf);
}


static void
writeSubGroup(HickSubGroup *head, Buffert *buf)
{
  HickSubGroup *sub, *sdown;

  for(sub = head; sub; sub = sub->hsg_next) {
    bprint(buf, "\n  subGroup %S %S", sub->hsg_group, sub->hsg_name);

    if(sub->hsg_versions.siz) {
      writeArray(&sub->hsg_versions, buf);
    }

    if(sub->hsg_down) {
      for(sdown = sub->hsg_down; sdown; sdown = sdown->hsg_down) {
        bprint(buf, " & %S", sdown->hsg_name);

        if(sdown->hsg_versions.siz) {
          writeArray(&sdown->hsg_versions, buf);
        }
      }
    }
  }

  bput('\n', buf);
}


static void
writeArray(ArrayInfo *arri, Buffert *buf)
{
  unsigned int n;

  _bputs(" { ", 3, buf);

  for(n = 0; n < arri->siz; n++) {
    bprint(buf, "%S", arri->arr[n]);

    if((n + 1) < arri->siz) {
      bput(' ', buf);
    }
  }

  _bputs(" }", 2, buf);
}


static void
writeDate(const char *s, time_t t, int tlevel, Buffert *buf)
{
  char tmps[BUFLEN+1];

  if(!t) {
    return;
  }

  switch(tlevel) {
    case HTIME_TIMET: {
      bprint(buf, "%s\t%d\n", s, t);
      return;
    }
    case HTIME_DATE: {
      strftime(tmps, BUFLEN, "%d/%m/%Y", localtime(&t));
      break;
    }
    case HTIME_HOURS: {
      strftime(tmps, BUFLEN, "%d/%m/%Y,%R", localtime(&t));
      break;
    }
    case HTIME_SECONDS: {
      strftime(tmps, BUFLEN, "%d/%m/%Y,%R.%S", localtime(&t));
      break;
    }
  }

  bprint(buf, "%s\t%s\n", s, tmps);
}
