/* HICK Lib 1.3.2 -- User Profiler Library.
** Copyright (C) 1997 - 2000 Daniel Bergstrom <daniel@octocode.com>
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the Free
** Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <ctype.h>
#include <stdlib.h>

#include "hickenv.h"
#include "common.h"
#include "getfile.h"
#include "getpacket.h"

static FileToken envFlags[] = {
  { "prepend", HFLAG_ENV_PREPEND },
  { "append",  HFLAG_ENV_APPEND  },
  { "set",     HFLAG_ENV_SET     },
  { "unset",   HFLAG_ENV_UNSET   },
  { NULL, 0 }
};

static FileToken createFlags[] = {
  { "new", HFLAG_CREATE_NEW       },
  { "append", HFLAG_CREATE_APPEND },
  { NULL, 0 }
};

static FileToken linkFlags[] = {
  { "normal", HFLAG_LINK_NORMAL },
  { "rename", HFLAG_LINK_RENAME },
  { NULL, 0 }
};

static FileToken xresFlags[] = {
  { "inline",   HFLAG_XRES_INLINE   },
  { "filename", HFLAG_XRES_FILENAME },
  { NULL, 0 }
};

static FileToken callFlags[] = {
  { "install", HFLAG_CALL_INSTALL },
  { "update",  HFLAG_CALL_UPDATE  },
  { "remove",  HFLAG_CALL_REMOVE  },
  { NULL, 0 }
};

static void getHickEnv(FILE *file, int var, HickEnvList *list,
                       const char *type, unsigned int defpri);
static void trashHickEnv(FILE *file, int var);
static HickRequirePacket *getHickEnvRequire(FILE *file);
static void trashHickEnvRequire(FILE *file);
static void freeHickEnvItem(HickEnvItem *head);
static int findFlag(const char *name, FileToken *flags, unsigned int *value);


void
_getEnv(FILE *file, int var, HickEnvList *list, const char *type,
        unsigned int defpri)
{
  int c;

  if((c = _getSpaces(file)) == '{') { /* see if we are going to read a list */
    while(((c = _getSpaces(file)) != EOF) && (c != '}')) {
      if(c == '#') {
        _getComment(file);
        continue;
      }

      ungetc(c, file);

      if(list) {
        getHickEnv(file, var, list, type, defpri);
      } else {
        trashHickEnv(file, var);
      }
    }
  } else {
    ungetc(c, file);

    if(list) {
      getHickEnv(file, var, list, type, defpri);
    } else {
      trashHickEnv(file, var);
    }
  }
}


static void
trashHickEnv(FILE *file, int var)
{
  char *s = 0, *p, c;
  unsigned int f = 0, flag, gnew = 0;

  if(var == HED_ENV_REQUIRE) {
    trashHickEnvRequire(file);
    return;
  }
  
  if(!(s = _getString(file))) {
    return;
  }

  switch(var) {
    case HED_ENV_ENV: {
      gnew = findFlag(s, envFlags, &flag);
      f = (flag == HFLAG_ENV_UNSET);
      break;
    }
    case HED_ENV_LINK: {
      gnew = findFlag(s, linkFlags, &flag);
      break;
    }
    case HED_ENV_XRES: {
      gnew = findFlag(s, xresFlags, &flag);
      f = (flag == HFLAG_XRES_FILENAME);
      break;
    }
    case HED_ENV_CREATE: {
      gnew = findFlag(s, createFlags, &flag);
      break;
    }
    case HED_ENV_CALL: {
      if((gnew = findFlag(s, callFlags, &flag))) {
        for(c = _getSpaces(file); (c == '&'); c = _getSpaces(file)) {
          FreeMem(s);
          if(!(s = _getString(file))) {
            return;
          }

          if(!(gnew = findFlag(s, callFlags, &flag))) {
            break;
          }
        }

        ungetc(c, file);
        gnew = 1;
      }
    }
  }

  if(gnew) { /* if we got a flag, get a new string */
    FreeMem(s);

    if(!(s = _getString(file))) {
      return;
    }
  }

  if((var == HED_ENV_SCRIPT) || (var == HED_ENV_ENV) ||
     (var == HED_ENV_MITEM) || (var == HED_ENV_CREATE)) {

    for(p = s, gnew = 1; *p; p++) {
      if(!isdigit((int)*p)) {
        gnew = 0;
      }
    }

    if(gnew) {
      _readString(file, 1);
    }
  }

  FreeMem(s);

  if(var == HED_ENV_CONFLICT) {
    if((c = _getSpaces(file)) == '{') {
      _readStmt(file, 0);
    } else {
      ungetc(c, file);
      _readString(file, 1);
    }
  }

  if(!f && (var != HED_ENV_CONFLICT)) {
    _readString(file, 1);

    if((var == HED_ENV_MITEM) || (var == HED_ENV_INFO)) {
      _readString(file, 1);

      if(var == HED_ENV_MITEM) {
        if((c = _getSpaces(file)) == '{') {
          _readStmt(file, 0);
        } else {
          ungetc(c, file);
        }
      }
    }

    if(var == HED_ENV_INFO) {
      _readString(file, 1);
      _readString(file, 1);
    }
  }
}


static void
getHickEnv(FILE *file, int var, HickEnvList *list, const char *hicktype,
           unsigned int defpri)
{
  HickEnv *henv;
  int f, gnew, c;
  ArrayInfo arri;
  HickEnvItem *item, *head;
  unsigned int pri, flags;
  char *s1, *s2, *s3, *s4, *s5, *p;
  HickRequirePacket *rpac, *rptmp, *rpold, *rphead;

  /* If we got an require variable, then we read in a different way. Since the
   * require has or statements.
   */

  if(var == HED_ENV_REQUIRE) {
    if(!(rphead = getHickEnvRequire(file))) {
      return;
    }

    /* Time to optimize the OR statements for the require information. If
     * there is a require in the OR statment with the same packet name, we
     * copy over the information into one, and remove the other.
     */

    for(rpac = rphead; rpac; rpac = rpac->rpac_or) {
      for(rpold = rpac, rptmp = rpac->rpac_or; rptmp; ) {
        if(!strcasecmp(rpac->rpac_name, rptmp->rpac_name)) {
          /* we found one, if either of them implies that we want all the
           * versions, we'll go with that.
           */

          if(!rpac->rpac_versions.siz || !rptmp->rpac_versions.siz) {
            hsh_array_remove(&rpac->rpac_versions);
          } else {
            /* Oh, well.. copy over the information then.. */

            for(c = 0; c < rptmp->rpac_versions.siz; c++) {
              hsh_array_add(rptmp->rpac_versions.arr[c], 0,
                            &rpac->rpac_versions);
            }
          }

          FreeMem(rptmp->rpac_versions.arr);

          rpold->rpac_or = rptmp->rpac_or;
          FreeMem(rptmp);
          rptmp = rpold->rpac_or;
          continue;
        }

        rpold = rptmp;
        rptmp = rptmp->rpac_next;
      }
    }
  } else {
    gnew = f = flags = 0;
    s1 = s2 = s3 = s4 = s5 = NULL;
    pri = (var == HED_ENV_CREATE ? HED_DEF_PERM : defpri);

    hsh_array_init(&arri);

    /* Read the first argument, it's the same for all of the variables */
  
    if(!(s1 = _getString(file))) {
      return;
    }

    switch(var) {
      case HED_ENV_ENV: { /* check for env flags */
        gnew = findFlag(s1, envFlags, &flags);

        /* If we have the unset flag we have to remember that we only want
         * to read in the name of the variable, and no value.
         */

        f = (flags == HFLAG_ENV_UNSET);
        break;
      }
      case HED_ENV_LINK: { /* check for link flags */
        gnew = findFlag(s1, linkFlags, &flags);
        break;
      }
      case HED_ENV_CREATE: { /* check for create flags */
        gnew = findFlag(s1, createFlags, &flags);
        break;
      }
      case HED_ENV_XRES: { /* check for xres flags */
        gnew = findFlag(s1, xresFlags, &flags);
        
        /* We have to store this information to be used later, because the
         * XRes has a little exception with the filename flag.
         */

        f = (flags == HFLAG_XRES_FILENAME);
        break;
      }
      case HED_ENV_CALL: {
        if(!(gnew = findFlag(s1, callFlags, &flags))) {
          flags = (HFLAG_CALL_INSTALL | HFLAG_CALL_UPDATE | HFLAG_CALL_REMOVE);
        } else {
          f = 0;
          f |= flags;

          for(c = _getSpaces(file); (c == '&'); c = _getSpaces(file)) {
            FreeMem(s1);
            if(!(s1 = _getString(file))) {
              return;
            }

            if(!(gnew = findFlag(s1, callFlags, &flags))) {
              break;
            }

            f |= flags;
          }

          ungetc(c, file);

          flags = f;
          gnew = 1;
          f = 0;
        }
      }
    }

    /* If we had a flag, we need to get the next argument */

    if(gnew) {
      FreeMem(s1);

      if(!(s1 = _getString(file))) {
        return;
      }
    }

    /* Now we have to check if the first (and in some cases the second) value
     * is a number. Which in most cases means that we got a priority level.
     */
    
    if((var == HED_ENV_SCRIPT) || (var == HED_ENV_ENV) ||
       (var == HED_ENV_MITEM) || (var == HED_ENV_CREATE)) {

      /* Go though the string and check for not digits. */

      for(p = s1; isdigit((int)*p); p++);

      if(*p == 0) {
        /* The creates number value is an octal number, so we have to
         * convert it in an other fashion then the normal priority digits.
         */

        if(var == HED_ENV_CREATE) {
          for(pri = 0, p = s1; ((*p >= '0') && (*p <= '7')); p++) {
            pri = pri * 8 + *p - '0';
          }
        } else {
          pri = atoi(s1); /* store the number in the priority var. */
        }
        
        FreeMem(s1);

        if(!(s1 = _getString(file))) {
          return;
        }
      }
    }

    if(var == HED_ENV_CONFLICT) {
      _getList(file, &arri);
      s4 = NULL;
    }

    /* XRes only want a filename or the Env have unset, we don't want to read
     * more.
     */
    
    if(!f && (var != HED_ENV_CONFLICT)) {
      if(!s1 || !(s2 = _getString(file))) {
        if(s1) {
          FreeMem(s1);
        }
        return;
      }

      /* Most of them are now finished, we got menuItem and Info left
       */

      if((var == HED_ENV_MITEM) || (var == HED_ENV_INFO)) {
        if(!(s3 = _getString(file))) {
          FreeMem(s1);
          FreeMem(s2);
          return;
        }

        if(var == HED_ENV_MITEM) {
          /* menuItem can have a list of menu arguments, so we have to check
           * for that here, and read it.
           */

          head = NULL;

          if((c = _getSpaces(file)) != '{') {
            ungetc(c, file); /* No menuItem variables. */
          } else {
            while(((c = _getSpaces(file)) != (char)EOF) && (c != '}')) {
              if(c == '#') {
                _getComment(file);
                continue;
              }

              ungetc(c, file);

              if(!(p = _getString(file)) || !(s5 = _getString(file))) {
                FreeMem(s1);
                FreeMem(s2);
                FreeMem(s3);
                FreeMem(s4);

                if(p) {
                  FreeMem(p);
                }
                return; /* something is wrong, better exit here. */
              }

              item = MallocMem(HickEnvItem);
              item->eit_name  = p;
              item->eit_value = s5;
              item->eit_next = (head ? head : NULL); 
              head = item;
            }
          }
        }

        /* Info got the two last arguments left, we read them here. */

        if((var == HED_ENV_INFO) &&
           (!(s4 = _getString(file)) || !(s5 = _getString(file)))) {
          FreeMem(s1);
          FreeMem(s2);
          FreeMem(s3);

          if(s4) {
            FreeMem(s4);
          }
          return;
        }
      }
    }
  }

  if(var == HED_ENV_CONFLICT) {
    for(henv = list->el_head; henv; henv = henv->env_next) {
      if(!strcasecmp(s1, henv->env_name)) {
        if(!henv->u.env_versions->siz || !arri.siz) {
          hsh_array_remove(&arri);
          hsh_array_remove(henv->u.env_versions);
        } else {
          for(pri = 0; pri < arri.siz; pri++) {
            hsh_array_add(arri.arr[pri], 0, henv->u.env_versions);
          }

          FreeMem(arri.arr);
        }
        return;
      }
    }
  }

  /* Now allocate the HickEnv structure, and point all the information */

  henv = MallocMem(HickEnv);
  henv->env_next = henv->env_prev = NULL;
  henv->env_flags = henv->env_priority = 0;
  henv->env_name = henv->env_value = henv->env_itemTitle = henv->env_about = 0;

  hsh_array_init(&henv->env_types);

  /* We don't have to much crap if we have an require variable */

  if(var != HED_ENV_REQUIRE) {
    henv->env_flags    = flags;
    henv->env_priority = pri;
    henv->env_name     = s1;
    henv->env_about    = s4;
  }

  switch(var) {
    case HED_ENV_MITEM: { /* MenuItem */
      henv->env_value     = s3;
      henv->env_itemTitle = s2;
      henv->u.env_items   = head;
      break;
    }
    case HED_ENV_REQUIRE: { /* Require */
      henv->u.env_rpacs = rphead;
      break;
    }
    case HED_ENV_CONFLICT: {
      henv->u.env_versions = MallocMem(ArrayInfo);
      henv->u.env_versions->arr = arri.arr;
      henv->u.env_versions->siz = arri.siz;
      henv->u.env_versions->alloc = arri.alloc;
      break;
    }
    case HED_ENV_INFO: { /* Info, we do want to fall thru ;) */
      henv->u.env_topic = s5;
    }
    default: {
      henv->env_value     = s2;
      henv->env_itemTitle = s3;
    }
  }

  if(hicktype) {
    hsh_array_add((char *)hicktype, 1, &henv->env_types);
  }

  henv->env_next = NULL;
  henv->env_prev = list->el_tail;
  list->el_tail = (list->el_tail ? (list->el_tail->env_next = henv) :
                   (list->el_head = henv));
}


static HickRequirePacket *
getHickEnvRequire(FILE *file)
{
  int c;
  char *s1;
  ArrayInfo arri;
  HickRequirePacket *rpac = 0;

  if(!(s1 = _getString(file))) {
    return 0;
  }

  _getList(file, &arri);

  if(!(rpac = MallocMem(HickRequirePacket))) {
    FreeMem(s1);
    hsh_array_remove(&arri);
    return 0;
  }

  strncpy(rpac->rpac_name, s1, PNAME_LEN);
  rpac->rpac_name[PNAME_LEN] = (char)0;
  rpac->rpac_versions.arr   = arri.arr;
  rpac->rpac_versions.siz   = arri.siz;
  rpac->rpac_versions.alloc = arri.alloc;
  rpac->rpac_next = NULL;

  if((c = _getSpaces(file)) == '|') {
    rpac->rpac_or = getHickEnvRequire(file);
  } else {
    rpac->rpac_or = NULL;
    ungetc(c, file);
  }

  FreeMem(s1);
  return rpac;
}


static void
trashHickEnvRequire(FILE *file)
{
  int c;

  _readString(file, 1);

  if((c = _getSpaces(file)) == '{') {
    _readStmt(file, 0);
  } else {
    ungetc(c, file);
    _readString(file, 1);
  }

  if((c = _getSpaces(file)) == '|') {
    trashHickEnvRequire(file);
  } else {
    ungetc(c, file);
  }
}


static int
findFlag(const char *name, FileToken *flags, unsigned int *value)
{
  int i;

  for(i = 0; flags[i].name; i++) {
    if(!_token_case_cmp(name, flags[i].name)) {
      *value = flags[i].token;
      return 1;
    }
  }

  *value = flags[i].token;
  return 0;
}


HickEnvList *
freeHickEnvList(HickEnvList *list, int type)
{
  HickEnv *env;

  for(env = list->el_head; env; env = list->el_head) {
    list->el_head = env->env_next;

    FreeMem(freeHickEnv(env, type));
  }

  return list;
}


HickEnv *
freeHickEnv(HickEnv *env, int type)
{
  if(env->env_name) {
    FreeMem(env->env_name);
  }

  if(env->env_value) {
    FreeMem(env->env_value);
  }

  if(env->env_itemTitle) {
    FreeMem(env->env_itemTitle);
  }

  if(env->env_about) {
    FreeMem(env->env_about);
  }

  switch(type) {
    case HED_ENV_INFO: { /* Info */
      FreeMem(env->u.env_topic);
      break;
    }
    case HED_ENV_MITEM: { /* MenuItem */
      freeHickEnvItem(env->u.env_items);
      break;
    }
    case HED_ENV_REQUIRE: { /* Require */
      _freeHickRequirePacket(env->u.env_rpacs);
      break;
    }
    case HED_ENV_CONFLICT: {
      hsh_array_remove(env->u.env_versions);
      FreeMem(env->u.env_versions);
    }
  }

  hsh_array_remove(&env->env_types);
  return env;
}


static void
freeHickEnvItem(HickEnvItem *head)
{
  HickEnvItem *item;

  for(item = head; item; item = head) {
    head = item->eit_next;

    FreeMem(item->eit_name);
    FreeMem(item->eit_value);
    FreeMem(item);
  }
}


void
_freeHickRequirePacket(HickRequirePacket *head)
{
  HickRequirePacket *rpac;

  for(rpac = head; rpac; rpac = head) {
    head = rpac->rpac_or;

    hsh_array_remove(&rpac->rpac_versions);
    FreeMem(rpac);
  }
}


void
_envMoveList(HickEnvList *from, HickEnvList *to)
{
  if(to->el_head) {
    to->el_head->env_prev = from->el_tail;
  }

  if(from->el_tail) {
    from->el_tail->env_next = to->el_head;
  }

  if(!to->el_tail) {
    to->el_tail = from->el_tail;
  }

  if(from->el_head) {
    to->el_head = from->el_head;
  }

  from->el_head = from->el_tail = NULL;
}


HickEnv *
_envRemove(HickEnv *env, HickEnvList *from)
{
  if(env) {
    if(!env->env_prev) {
      if(from->el_head == env) {
        from->el_head = env->env_next;
      }
    } else {
      if(env->env_prev->env_next != env) {
        return NULL;
      }

      env->env_prev->env_next = env->env_next;
    }

    if(!env->env_next) {
      if(from->el_tail == env) {
        from->el_tail = env->env_prev;
      }
    } else {
      if(env->env_next->env_prev != env) {
        return NULL;
      }

      env->env_next->env_prev = env->env_prev;
    }

    env->env_next = env->env_prev = NULL;
  }

  return env;
}
