/* HICK Lib 1.3.2 -- User Profiler Library.
** Copyright (C) 1997 - 2000 Daniel Bergstrom <daniel@octocode.com>
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the Free
** Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>

#include "entry.h"
#include "common.h"
#include "getfile.h"
#include "hickenv.h"
#include "version.h"
#include "getpacket.h"

#define CLOSE_FILE(file, flag) flag = 0; fclose(file);


int
openPackets(const char *packpath, const char *homepath,
            const char *hicktype, const char *group, HickPackets *result)
{
  int err;
  char fname[FNAME_LEN+1];

  result->hp_typedir = result->hp_packdir = NULL;
  result->hp_curtype = result->hp_curpack = NULL;
  result->hp_file = NULL;
  result->hp_isopen = 0;
  result->hp_packpath = result->hp_homepath = result->hp_type = NULL;
  result->hp_group[0] = (char)0;
  result->hp_remove.ul_head = result->hp_remove.ul_tail = NULL;
  result->hp_install.ul_head = result->hp_install.ul_tail = NULL;
  result->hp_installed.ul_head = result->hp_installed.ul_tail = NULL;

  if(!(result->hp_typedir = hsh_opendir(packpath, NULL))) {
    return HICKERR_PACKDIR;
  }

  if(!(result->hp_packpath = MallocArr(char, strlen(packpath) + 1))) {
    return HICKERR_OUTOFMEMORY;
  }

  strcpy(result->hp_packpath, packpath);

  if(homepath) {
    if(!(result->hp_homepath = MallocArr(char, strlen(homepath) + 1))) {
      FreeMem(result->hp_packpath);
      return HICKERR_OUTOFMEMORY;
    }

    strcpy(result->hp_homepath, homepath);
  }

  if(hicktype) {
    if(!(result->hp_type = MallocArr(char, strlen(hicktype) + 1))) {
      FreeMem(result->hp_packpath);
      FreeMem(result->hp_homepath);
      return HICKERR_OUTOFMEMORY;
    }
 
    strcpy(result->hp_type, hicktype);
  }

  if(group) {
    strncpy(result->hp_group, group, PGROUP_LEN);
    result->hp_group[PGROUP_LEN] = (char)0;
  }

  if(!homepath || !hicktype) {
    return 0;
  }

  hsh_create_str(fname, FNAME_LEN, homepath, "/.", hicktype, "/",
                 INSTALLED_FILE, NULL);

  err = _entryLoad(fname, &result->hp_installed);

  return (err == HICKERR_INSTFILE ? 0 : err);
}


void
closePackets(HickPackets *pack)
{
  if(pack->hp_typedir) {
    closedir(pack->hp_typedir);
  }

  if(pack->hp_packdir) {
    closedir(pack->hp_packdir);
  }

  if(pack->hp_isopen) {
    fclose(pack->hp_file);
  }

  if(pack->hp_packpath) {
    FreeMem(pack->hp_packpath);
  }

  if(pack->hp_homepath) {
    FreeMem(pack->hp_homepath);
  }

  if(pack->hp_type) {
    FreeMem(pack->hp_type);
  }

  _freeUserEntryList(&pack->hp_remove);
  _freeUserEntryList(&pack->hp_install);
  _freeUserEntryList(&pack->hp_installed);
}


void
initHickEntry(HickEntry *entry)
{
  entry->he_name[0] = (char)0;
  entry->he_about = NULL;
  entry->he_created = entry->he_modified = entry->he_expire = 0;
  entry->he_subGroups = NULL;
  entry->he_hidden = entry->he_required = entry->he_multiple = 0;
  hsh_array_init(&entry->he_using);
  hsh_array_init(&entry->he_force);
  hsh_array_init(&entry->he_groups);
  initHickVerList(&entry->he_versions);
}


HickEntry *
freeHickEntry(HickEntry *entry)
{
  freeHickVerList(&entry->he_versions, 0);
  freeHickSubGroup(entry->he_subGroups);

  hsh_array_remove(&entry->he_using);
  hsh_array_remove(&entry->he_force);
  hsh_array_remove(&entry->he_groups);

  if(entry->he_about) {
    FreeMem(entry->he_about);
  }

  initHickEntry(entry); /* clear the HickEntry */

  return entry;
}


int
nextPacketHeader(HickPackets *pack, HickEntry *result)
{
  while(1) {
    freeHickEntry(result);

    if(!pack->hp_isopen && !_getNextEntry(pack)) {
      return 0;
    }

    if(!_getPacketHead(pack->hp_file, result)) {
      CLOSE_FILE(pack->hp_file, pack->hp_isopen);
      continue; /* we failed, try to get a new one. */
    }

    _readStmt(pack->hp_file, 0); /* trash the packet contents. */

    if(pack->hp_group[0] &&
       !hsh_array_find(pack->hp_group, &result->he_groups)) {
      continue;
    }

    return 1;
  }
}


int
nextPacketFull(HickPackets *pack, HickPacket *result)
{
  HickEntry entry;

  initHickEntry(&entry);

  while(1) {
    freeHickEntry(&entry);
    freeHickPacket(result);

    if(!pack->hp_isopen && !_getNextEntry(pack)) {
      return 0;
    }

    if(!_getPacketHead(pack->hp_file, &entry)) {
      CLOSE_FILE(pack->hp_file, pack->hp_isopen);
      continue;
    }

    _getPacket(pack->hp_file, NULL, result, 1);

    if(pack->hp_group[0] &&
       !hsh_array_find(pack->hp_group, &result->pac_groups)) {
      continue;
    }

    strcpy(result->pac_name, entry.he_name);
    result->pac_groups.arr = entry.he_groups.arr;
    result->pac_groups.siz = entry.he_groups.siz;
    result->pac_groups.alloc = entry.he_groups.alloc;

    entry.he_groups.arr = NULL;
    entry.he_groups.siz = 0;
    entry.he_groups.alloc = 0;
    return 1;
  }
}


int
nextPacket(HickPackets *pack, HickEntry *result, int sabout)
{
  int c;
  userEntry *ue;
  HickSubGroup *sub;
  HickVerList denied;
  DoubleArray forceArr;
  char buf[TMPBUFSIZ+1];
  HickDirection *allow, *deny, *force;

  allow = deny = force = NULL;
  denied.vl_head = denied.vl_tail = NULL;

  while(1) {
    freeHickEntry(result); /* we start of cleaning the entry */
    freeHickDirection(deny);
    freeHickDirection(allow);
    freeHickDirection(force);
    freeHickVerList(&denied, 0);

    allow = deny = force = NULL;

    if(!pack->hp_isopen && !_getNextEntry(pack)) {
      return 0;
    }

    if(!_getPacketHead(pack->hp_file, result)) {
      CLOSE_FILE(pack->hp_file, pack->hp_isopen);
      continue; /* we failed, try to get a new one. */
    }

    if(pack->hp_group[0] &&
       !hsh_array_find(pack->hp_group, &result->he_groups)) {
      _readStmt(pack->hp_file, 0); /* trash the packet contents. */
      continue;
    }

    while(((c = _getSpaces(pack->hp_file)) != (char)EOF) && (c != '}')) {
      if(c == '#') {
        _getComment(pack->hp_file);
        continue;
      }

      ungetc(c, pack->hp_file);

      if(!_getName(pack->hp_file, buf, TMPBUFSIZ)) {
        break;
      }

      switch((c = _token_lookup(buf, _tokens))) {
        case TOKEN_ABOUT: {
          if(!sabout) {
            _readString(pack->hp_file, 1);
          } else {
            if(result->he_about) {
              FreeMem(result->he_about);
            }

            result->he_about = _getString(pack->hp_file);
          }
          break;
        }
        case TOKEN_HIDDEN: {
          result->he_hidden = 1;
          break;
        }
        case TOKEN_REQUIRED: {
          result->he_required = 1;
          break;
        }
        case TOKEN_MULTIPLE: {
          result->he_multiple = 1;
          break;
        }
        case TOKEN_CREATED: {
          result->he_created = _getTime(pack->hp_file);
          break;
        }
        case TOKEN_MODIFIED: {
          result->he_modified = _getTime(pack->hp_file);
          break;
        }
        case TOKEN_EXPIRE: {
          result->he_expire = _getTime(pack->hp_file);
          break;
        }
        case TOKEN_ALLOW: {
          _getDirection(pack->hp_file, &allow, 0);
          break;
        }
        case TOKEN_DENY: {
          _getDirection(pack->hp_file, &deny, 0);
          break;
        }
        case TOKEN_FORCE: {
          _getDirection(pack->hp_file, &force, 1);
          break;
        }
        case TOKEN_SUBGROUP: {
          if((sub = _getSubgroup(pack->hp_file, 1))) {
            sub->hsg_next = (result->he_subGroups ? result->he_subGroups : 0);
            result->he_subGroups = sub;
          }
          break;
        }
        case TOKEN_VERSION: {
          _getVersion(&result->he_versions, pack->hp_file, 0, NULL, sabout);
          break;
        }
        case HED_ENV_INFO:
        case HED_ENV_CALL:
        case HED_ENV_CONFLICT:
        case HED_ENV_CREATE:
        case HED_ENV_ENV:
        case HED_ENV_LINK:
        case HED_ENV_MITEM:
        case HED_ENV_REQUIRE:
        case HED_ENV_SCRIPT:
        case HED_ENV_XRES:
        case HED_ENV_HTML: {
          _getEnv(pack->hp_file, c, 0, 0, 0); /* trash the variable. */
          break;
        }
      }
    }

    if(!_versionPermissions(&result->he_versions, allow, deny, &denied)) {
      continue; /* no permissions to packet */
    }

    _checkDirection(force, PERM_FORCE, &forceArr); /* get force information */
    result->he_force.arr = forceArr.arr1.arr;
    result->he_force.siz = forceArr.arr1.siz;
    result->he_force.alloc = forceArr.arr1.alloc;

    freeHickDirection(deny);
    freeHickDirection(allow);
    freeHickDirection(force);
    freeHickVerList(&denied, 0);

    if(!result->he_multiple) {
      if((ue = _entryFind(result->he_name, &pack->hp_installed))) {
        hsh_array_add(ue->ue_version, 1, &result->he_using);
      }
    } else {
      for(ue = pack->hp_installed.ul_head; ue; ue = ue->ue_next) {
        if(!strcasecmp(result->he_name, ue->ue_name)) {
          hsh_array_add(ue->ue_version, 1, &result->he_using);
        }
      }
    }

    return 1;
  }
}


HickResult *
freeHickResult(HickResult *result)
{
  HickForced *fo;
  HickRequire *req;
  HickFailure *fail;
  HickConflict *con;
  HickRequirePacket *rpac;

  for(req = result->hr_required; req; req = result->hr_required) {
    result->hr_required = req->req_next;

    for(rpac = req->req_list; rpac; rpac = req->req_list) {
      req->req_list = rpac->rpac_next;

      _freeHickRequirePacket(rpac->rpac_or);
      hsh_array_remove(&rpac->rpac_versions);
      FreeMem(rpac);
    }

    FreeMem(req);
  }

  for(con = result->hr_conflicts; con; con = result->hr_conflicts) {
    result->hr_conflicts = con->con_next;
    FreeMem(con);
  }

  for(fail = result->hr_failure; fail; fail = result->hr_failure) {
    result->hr_failure = fail->fail_next;
    FreeMem(fail);
  }

  for(fo = result->hr_forced; fo; fo = result->hr_forced) {
    result->hr_forced = fo->fo_next;
    FreeMem(fo);
  }

  return result;
}
