/* HICK Lib 1.3.2 -- User Profiler Library.
** Copyright (C) 1997 - 2000 Daniel Bergstrom <daniel@octocode.com>
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the Free
** Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>

#include "common.h"
#include "getfile.h"
#include "version.h"
#include "hickenv.h"
#include "getpacket.h"


HickVer *
_versionAppend(HickVer *ver, HickVerList *to)
{
  if(ver) {
    ver->ver_next = NULL;
    ver->ver_prev = to->vl_tail;
    to->vl_tail = (to->vl_head ? (to->vl_tail->ver_next = ver) :
                   (to->vl_head = ver));
  }

  return ver;
}


HickVer *
_versionFind(const char *s, HickVerList *list)
{
  HickVer *ver;

  for(ver = list->vl_head; ver; ver = ver->ver_next) {
    if(!strcasecmp(s, ver->ver_number)) {
      return ver;
    }
  }

  return NULL;
}


HickVer *
_versionMove(HickVer *ver, HickVerList *from, HickVerList *to)
{
  if(_versionRemove(ver, from)) {
    return _versionAppend(ver, to);
  }

  return NULL;
}


void
_versionMoveList(HickVerList *from, HickVerList *to)
{
  if(to->vl_head) {
    to->vl_head->ver_prev = from->vl_tail;
  }

  if(from->vl_tail) {
    from->vl_tail->ver_next = to->vl_head;
  }

  if(!to->vl_tail) {
    to->vl_tail = from->vl_tail;
  }

  if(from->vl_head) {
    to->vl_head = from->vl_head;
  }

  from->vl_head = from->vl_tail = NULL;
}


HickVer *
_versionRemove(HickVer *ver, HickVerList *from)
{
  if(!ver) {
    return NULL;
  }

  if(!ver->ver_prev) {
    if(from->vl_head == ver) {
      from->vl_head = ver->ver_next;
    }
  } else {
    if(ver->ver_prev->ver_next != ver) {
      return NULL;
    }

    ver->ver_prev->ver_next = ver->ver_next;
  }

  if(!ver->ver_next) {
    if(from->vl_tail == ver) {
      from->vl_tail = ver->ver_prev;
    }
  } else {
    if(ver->ver_next->ver_prev != ver) {
      return NULL;
    }

    ver->ver_next->ver_prev = ver->ver_prev;
  }

  ver->ver_next = ver->ver_prev = NULL;
  return ver;
}


void
freeHickVerList(HickVerList *list, int hver)
{
  HickVer *ver;

  for(ver = list->vl_head; ver; ver = list->vl_head) {
    list->vl_head = ver->ver_next;
    FreeMem(freeHickVer(ver, hver));
  }

  list->vl_head = list->vl_tail = NULL;
}


HickVer *
freeHickVer(HickVer *ver, int hver)
{
  if(ver->ver_about) {
    FreeMem(ver->ver_about);
  }

  if(hver) {
    freeHickEnvList(&HVER_INFO(ver),     HED_ENV_INFO);
    freeHickEnvList(&HVER_CALL(ver),     HED_ENV_CALL);
    freeHickEnvList(&HVER_CONFLICT(ver), HED_ENV_CONFLICT);
    freeHickEnvList(&HVER_CREATE(ver),   HED_ENV_CREATE);
    freeHickEnvList(&HVER_ENV(ver),      HED_ENV_ENV);
    freeHickEnvList(&HVER_LINK(ver),     HED_ENV_LINK);
    freeHickEnvList(&HVER_MITEM(ver),    HED_ENV_MITEM);
    freeHickEnvList(&HVER_REQUIRE(ver),  HED_ENV_REQUIRE);
    freeHickEnvList(&HVER_SCRIPT(ver),   HED_ENV_SCRIPT);
    freeHickEnvList(&HVER_XRES(ver),     HED_ENV_XRES);
    freeHickEnvList(&HVER_HTML(ver),     HED_ENV_HTML);

    hsh_array_remove(&HVER_TYPES(ver));
  }

  return ver;
}


HickVer *
_getVersion(HickVerList *list, FILE *file, int hver, const char *hicktype,
            int sabout)
{
  int f, i;
  HickVer *ver;
  HickVersion *hv;
  HickEnvList *elist;
  unsigned int defpri;
  char c, *p, buf[TMPBUFSIZ + 1];

  if(!(p = _getString(file))) { /* get the a flag or the version number */
    return NULL;
  }

  if((f = (_token_case_cmp(p, "def") == 0))) {
    FreeMem(p); /* we got a flag, get the version number */

    if(!(p = _getString(file))) {
      return NULL;
    }
  }

  if(_getSpaces(file) != '{') { /* missing a '{' */
    FreeMem(p);
    return NULL;
  }

  if(hver) { /* Allocate and initialize the HickVersion structure */
    hv = MallocMem(HickVersion);

    initHickEnvList(&hv->hver_info);
    initHickEnvList(&hv->hver_call);
    initHickEnvList(&hv->hver_conflict);
    initHickEnvList(&hv->hver_create);
    initHickEnvList(&hv->hver_env);
    initHickEnvList(&hv->hver_link);
    initHickEnvList(&hv->hver_mitem);
    initHickEnvList(&hv->hver_require);
    initHickEnvList(&hv->hver_script);
    initHickEnvList(&hv->hver_xres);
    initHickEnvList(&hv->hver_html);

    hsh_array_init(&hv->hver_types);

    ver = &hv->hver_version;

    if(hicktype) {
      hsh_array_add((char *)hicktype, 1, &hv->hver_types);
    }
  } else { /* Allocate and initialize the HickVer structure */
    ver = MallocMem(HickVer);
  }

  strncpy(ver->ver_number, p, PVER_LEN);
  ver->ver_number[PVER_LEN] = (char)0;
  ver->ver_about = NULL;
  ver->ver_override = 0;
  ver->ver_created = ver->ver_modified = ver->ver_expire = 0;
  ver->ver_default = f;

  FreeMem(p);

  _versionAppend(ver, list); /* Append version into the list */

  /* read all the variables of the version */

  while(((c = _getSpaces(file)) != (char)EOF) && (c != '}')) {
    if(c == '#') {
      _getComment(file);
      continue;
    }

    ungetc(c, file);

    if(!_getName(file, buf, TMPBUFSIZ)) {
      return 0;
    }

    switch((i = _token_lookup(buf, _tokens))) {
      case TOKEN_CREATED: {
        ver->ver_created = _getTime(file);
        continue;
      }
      case TOKEN_MODIFIED: {
        ver->ver_modified = _getTime(file);
        continue;
      }
      case TOKEN_EXPIRE: {
        ver->ver_expire = _getTime(file);
        continue;
      }
      case TOKEN_ADDABOUT: {
        if(ver->ver_about) { /* clear off old about.. */
          FreeMem(ver->ver_about);
        }

        ver->ver_about = _getString(file);

        if(!_token_case_cmp(ver->ver_about, "override")) {
          ver->ver_override = 1;

          FreeMem(ver->ver_about);

          if(sabout) {
            ver->ver_about = _getString(file);
          } else {
            _readString(file, 1);
            continue;
          }
        }

        if(!sabout) {
          FreeMem(ver->ver_about);
          ver->ver_about = NULL;
        }
        continue;
      }
    }

    if((i < HED_ENV_INFO) || (i > HED_ENV_HTML)) {
      continue;
    }

    if(!hver) {
      _getEnv(file, i, 0, 0, 0);
      continue;
    }

    defpri = 0;
    elist = 0;

    switch(i) {
      case HED_ENV_INFO: { /* info */
        elist = &hv->hver_info;
        break;
      }
      case HED_ENV_CALL: { /* Call */
        elist = &hv->hver_call;
        break;
      }
      case HED_ENV_CONFLICT: { /* Conflict */
        elist = &hv->hver_conflict;
        break;
      }
      case HED_ENV_CREATE: { /* Create */
        elist = &hv->hver_create;
        break;
      }
      case HED_ENV_ENV: { /* Env */
        elist = &hv->hver_env;
        defpri = HICKPRI_ENVIRON;
        break;
      }
      case HED_ENV_LINK: { /* Link */
        elist = &hv->hver_link;
        break;
      }
      case HED_ENV_MITEM: { /* MenuItem */
        elist = &hv->hver_mitem;
        defpri = HICKPRI_MENUITEM;
        break;
      }
      case HED_ENV_REQUIRE: { /* Require */
        elist = &hv->hver_require;
        break;
      }
      case HED_ENV_SCRIPT: { /* Script */
        elist = &hv->hver_script;
        defpri = HICKPRI_SCRIPTS;
        break;
      }
      case HED_ENV_XRES: { /* Xres */
        elist = &hv->hver_xres;
        break;
      }
      case HED_ENV_HTML: { /* HTML */
        elist = &hv->hver_html;
      }
    }

    if(elist) {
      _getEnv(file, i, elist, hicktype, defpri);
    }
  }

  return ver;
}


int
_versionPermissions(HickVerList *list, HickDirection *allow,
                    HickDirection *deny, HickVerList *deniedList)
{
  int i;
  HickVer *ver, *tver;
  DoubleArray allowArr, denyArr;

  if(!allow && !deny) {
    return 1;
  }

  initDoubleArray(&denyArr);
  initDoubleArray(&allowArr);

  /* Check for user / group premissions to this file. */

  if(!(i = _checkDirection(allow, PERM_ALLOW, &allowArr)) ||
     (allow && (i == 1) && !allowArr.arr1.siz && !allowArr.arr2.siz)) {
    hsh_array_remove(&allowArr.arr1);
    _versionMoveList(list, deniedList);
    return 0; /* no access to packet, get next */
  }

  /* Remove denied versions from the allow information */

  for(i = 0; i < allowArr.arr2.siz; i++) {
    for(ver = list->vl_head; ver; ) {
      if(!hsh_array_find(allowArr.arr2.arr[i], &allowArr.arr1) &&
         !strcasecmp(ver->ver_number, allowArr.arr2.arr[i])) {
        tver = ver->ver_next;
        _versionMove(ver, list, deniedList);
        ver = tver;
      } else {
        ver = ver->ver_next;
      }
    }
  }

  if(!list->vl_head) {
    hsh_array_remove(&denyArr.arr1);
    hsh_array_remove(&allowArr.arr1);
    _versionMoveList(list, deniedList);
    return 0;
  }

  if(!(i = _checkDirection(deny, PERM_DENY, &denyArr)) || 
     ((i == 1) && deny && !denyArr.arr1.siz && !allow)) {
    hsh_array_remove(&denyArr.arr1);
    hsh_array_remove(&allowArr.arr1);
    _versionMoveList(list, deniedList);
    return 0;
  }

  if((i == 1) && deny && !denyArr.arr1.siz) {
    for(ver = list->vl_head; ver; ) { /* Remove all verions, but allowed */
      if(!hsh_array_find(ver->ver_number, &allowArr.arr1)) {
        tver = ver->ver_next;
        _versionMove(ver, list, deniedList);
        ver = tver;
      } else {
        ver = ver->ver_next;
      }
    }
  } else { /* Remove some versions, but allowed */
    for(i = 0; i < denyArr.arr1.siz; i++) {
      if(hsh_array_find(denyArr.arr1.arr[i], &allowArr.arr1)) {
        continue;
      }

      if((ver = _versionFind(denyArr.arr1.arr[i], list))) {
        _versionMove(ver, list, deniedList);
      }
    }
  }

  hsh_array_remove(&denyArr.arr1);
  hsh_array_remove(&allowArr.arr1);

  return (!list->vl_head ? 0 : 1); /* no versions left? get next packet.. */
}
