/* HickMK 1.2.2 -- HICK Make Utility.
** Copyright (C) 1998 - 2001 Daniel Bergstrom <daniel@octocode.com>
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the Free
** Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <ctype.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "utils.h"
#include "hickmk.h"

#include "memd.h"
#include "linker.h" /* Include files from HICK Lib 1.3 */
#include "common.h"

#define MIN_MENUSIZE  0
#define BONUS_SIZE    9 /* the length of '* : ().  ' */

/* struct menuItem ********************************************************* */

struct menuItem {
  struct menuItem *mi_next; /* pointer to next menuItem. */

  char *mi_name;            /* Name of the menu item. */
  char *mi_filename;        /* Filename which will be linked to. */
  char *mi_subMenu;         /* If any sub menu, it will be set here. */
  char *mi_about;           /* A short about message. */
};

/* struct menuTopic ******************************************************** */

struct menuTopic {
  struct menuTopic *mt_next; /* Pointer to next menuItems. */

  char *mt_topic;            /* Major topic of this menuItems. */
  size_t mt_menuSize;        /* Calculated length needed for the name,
                              * filename and submenu.
                              */
  struct menuItem *mt_head;  /* Pointer to first menuItem in list. */
};

static void remove_mitem(struct menuItem *head);
static void formatString(Buffert *buf, char *s, size_t insSpc, size_t maxLen);
static int writeFile(Buffert *b, ...);


void
hickmk_info_remove(const char *homepath, const char *hicktype)
{
  char fname[FNAME_LEN+1];

  hsh_create_str(fname, FNAME_LEN, homepath, "/", hicktype, "/", INFO_DIR,
                 "/", INFO_DIR_FILE, NULL);

  hsh_unlink(fname, NULL);
}


int
hickmk_info_create(const char *infohead, const char *homepath,
                   const char *hicktype, size_t maxLen)
{
  FILE *file;
  Buffert buf;
  size_t menuSize;
  struct menuItem *mitem;
  struct menuTopic noTopic, *topic;
  char line[80], fname[FNAME_LEN+1], *s1, *s2, *s3, *s4, *s5;

  buf.pos = 0;
  noTopic.mt_next     = NULL;
  noTopic.mt_topic    = NULL;
  noTopic.mt_menuSize = MIN_MENUSIZE;
  noTopic.mt_head     = NULL;

  /* make sure that we got the info dir under the .hick/<hick type>  */

  hsh_create_str(fname, FNAME_LEN, homepath, "/", hicktype, "/", INFO_DIR,
                 NULL);

  if(hsh_path_create(fname, NULL) == IOW_FAIL) {
    return -1;
  }

  hsh_create_str(fname, FNAME_LEN, homepath, "/", hicktype, "/", INFO_DIR, 
                 "/", INFO_DIR_FILE, NULL);

  if(!(buf.file = fopen(fname, "w"))) { /* Open the output file. */
    return -2;
  }

  bprint(&buf, "-*- Text -*-\nThis file is generated by HickMK %s.\n\n",
         HICKMKLIB_STRING_VERSION);

  if(!infohead || !writeFile(&buf, infohead, NULL)) {
    bputs("File: dir\tNode: Top\tThis is the top of the INFO tree\n\n* Menu:\n", &buf);
  }

  hsh_create_str(fname, FNAME_LEN, homepath, "/.", hicktype, "/", INFO_FILE,
                 NULL);

  if(!(file = fopen(fname, "r"))) {
    return -3;
  }

  while(!feof(file)) {
    if(!(s1 = _getCString(file))) { /* Topic */
      continue;
    }

    if(!(s2 = _getCString(file))) { /* Menu name */
      FreeMem(s1);
      continue;
    }

    if(!(s3 = _getCString(file))) { /* Info filename */
      FreeMem(s1);
      FreeMem(s2);
      continue;
    }

    if(!(s4 = _getCString(file))) { /* Sub menu */
      FreeMem(s1);
      FreeMem(s2);
      FreeMem(s3);
      continue;
    }

    if(!(s5 = _getCString(file))) { /* About message */
      FreeMem(s1);
      FreeMem(s2);
      FreeMem(s3);
      FreeMem(s4);
      continue;
    }

    if(*s1) { /* lookup the menu topic to use or create it */
      for(topic = noTopic.mt_next; topic; topic = topic->mt_next) {
        if(!strcasecmp(topic->mt_topic, s1)) {
          break;
        }
      }

      if(!topic) { /* Create new topic? */
        topic = MallocMem(struct menuTopic);
        topic->mt_topic    = s1;
        topic->mt_head     = NULL;
        topic->mt_next     = noTopic.mt_next;
        topic->mt_menuSize = MIN_MENUSIZE;
        noTopic.mt_next = topic;
        s1 = NULL;
      } else {
        FreeMem(s1);
      }
    } else {
      FreeMem(s1);
      topic = &noTopic;
    }

    mitem = MallocMem(struct menuItem); /* Allocate menu item */
    mitem->mi_name     = s2;
    mitem->mi_filename = s3;
    mitem->mi_subMenu  = s4;
    mitem->mi_about    = s5;
    mitem->mi_next = topic->mt_head;
    topic->mt_head = mitem;

    menuSize = strlen(s2) + strlen(s3) + strlen(s4) + BONUS_SIZE;

    topic->mt_menuSize = (topic->mt_menuSize < menuSize ? menuSize :
                          topic->mt_menuSize);
  }

  fclose(file);

  for(topic = &noTopic; topic; topic = topic->mt_next) {
    if(topic->mt_topic) {
      bprint(&buf, "\n%s\n\n", topic->mt_topic);
    }

    for(mitem = topic->mt_head; mitem; mitem = mitem->mi_next) {
      hsh_create_str(line, 80, "* ", mitem->mi_name, ": (", mitem->mi_filename,
                     ")", mitem->mi_subMenu, ".  ", NULL);

      for(menuSize = strlen(line); menuSize < topic->mt_menuSize; ++menuSize) {
        line[menuSize] = ' ';
      }

      line[topic->mt_menuSize] = (char)0;

      _bputs(line, topic->mt_menuSize, &buf);

      formatString(&buf, mitem->mi_about, topic->mt_menuSize, maxLen);

      bput('\n', &buf);
    }
  }

  bflush(&buf);
  fclose(buf.file);

  for(topic = noTopic.mt_next; topic; topic = noTopic.mt_next) {
    noTopic.mt_next = topic->mt_next;

    remove_mitem(topic->mt_head);
    FreeMem(topic->mt_topic);
    FreeMem(topic);
  }

  remove_mitem(noTopic.mt_head);
  return 0;
}


static void
remove_mitem(struct menuItem *head)
{
  struct menuItem *mitem;

  for(mitem = head; mitem; mitem = head) {
    head = mitem->mi_next;

    FreeMem(mitem->mi_name);
    FreeMem(mitem->mi_filename);
    FreeMem(mitem->mi_subMenu);
    FreeMem(mitem->mi_about);
    FreeMem(mitem);
  }
}


static void
formatString(Buffert *buf, char *s, size_t insSpc, size_t maxLen)
{
  char *wStart, *wEnd;
  size_t len, n, curLen;

  len    = (maxLen - insSpc);
  curLen = insSpc;

  for(wStart = wEnd = 0; *s; ++s) {
    if(isspace((int)*s)) {
      /* We got a word in our backpack? if so we'll add it to the current
       * line.
       */

      if(wStart) {
        ++wEnd; /* move the end one step */

        /* If this word togeather with the current length of the line is
         * to long to fix we'll have to add a new line.
         */

        if((curLen + (wEnd - wStart)) >= maxLen) {
          /* If the word is greater then the line length we'll have to split
           * it up.
           */

          if((wEnd - wStart) > (maxLen - insSpc)) {
            do {
              /* First we add what we can on the current line. */

              for(;(curLen < maxLen) && (wStart != wEnd); ++curLen, ++wStart) {
                bput(*wStart, buf);
              }

              /* If we haven't reached the end yet, we'll add a new line */

              if(wStart != wEnd) {
                bput('\n', buf);

                for(n = 0; n < insSpc; ++n) {
                  bput(' ', buf);
                }

                curLen = insSpc;
              }
            } while(wStart != wEnd);
          } else {
            bput('\n', buf);

            for(n = 0; n < insSpc; ++n) {
              bput(' ', buf);
            }

            for(curLen = insSpc; wStart != wEnd; ++wStart, ++curLen) {
              bput(*wStart, buf);
            }
          }
        } else { /* add the word */
          for( ; wStart != wEnd; ++wStart, ++curLen) {
            bput(*wStart, buf);
          }
        }

        wStart = wEnd = 0;
      }

      /* if we got a space or a tab we'll add a space into the buffert */

      if((*s == ' ') || (*s == '\t')) {
        ++curLen;
        bput(' ', buf);
      }
      continue;
    }

    if(!wStart) {
      wStart = wEnd = s;
    } else {
      ++wEnd;
    }
  }

  if(wStart) {
    ++wEnd; /* move the end one step */

    /* If this word togeather with the current length of the line is
     * to long to fix we'll have to add a new line.
     */

    if((curLen + (wEnd - wStart)) >= maxLen) {
      /* If the word is greater then the line length we'll have to split
       * it up.
       */

      if((wEnd - wStart) > (maxLen - insSpc)) {
        do {
          /* First we add what we can on the current line. */

          for( ; (curLen < maxLen) && (wStart != wEnd); curLen++, wStart++) {
            bput(*wStart, buf);
          }

          /* If we haven't reached the end yet, we'll add a new line */

          if(wStart != wEnd) {
            bput('\n', buf);
            
            for(n = 0; n < insSpc; ++n) {
              bput(' ', buf);
            }
            
            curLen = insSpc;
          }
        } while(wStart != wEnd);
      } else {
        bput('\n', buf);
        
        for(n = 0; n < insSpc; ++n) {
          bput(' ', buf);
        }
        
        for(curLen = insSpc; wStart != wEnd; ++wStart, ++curLen) {
          bput(*wStart, buf);
        }
      }
    } else { /* add the word */
      for( ; wStart != wEnd; ++wStart, ++curLen) {
        bput(*wStart, buf);
      }
    }
  }
}


static int
writeFile(Buffert *b, ...)
{
  int fd;
  va_list ap;
  struct stat sbuf;
  unsigned int n = 0;
  char buf[FNAME_LEN+1], *p;

  va_start(ap, b);

  while((p = va_arg(ap, char *)) && (n < FNAME_LEN)) {
    for( ; *p && (n < FNAME_LEN); ++n, ++p) {
      buf[n] = *p;
    }
  }

  buf[n] = (char)0;

  va_end(ap);

  if((stat(buf, &sbuf) == -1) || ((fd = open(buf, O_RDONLY)) == -1)) {
    return 0;
  }

  p = MallocArr(char, sbuf.st_size +1);
  if((n = read(fd, p, sbuf.st_size)) != -1) {
    _bputs(p, n, b);
  }

  FreeMem(p);
  close(fd);
  return 1;
}
