/* HickMK 1.2.2 -- HICK Make Utility.
** Copyright (C) 1998 - 2001 Daniel Bergstrom <daniel@octocode.com>
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the Free
** Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <ctype.h>
#include <unistd.h>

#include "common.h" /* from hick lib 1.3.1 */
#include "getfile.h"

#include "hickmk.h"

#include "utils.h"
#include "mkinterp.h"


typedef struct {
  NodeNum nnum;       /* sorted by priority level */

  char *name;         /* item name */
  char *ident;        /* ident string based on item name */
  char *command;      /* command */

  NodeList variables; /* variables */
} MenuItem;


typedef struct {
  NodeName nname;       /* sorted by name */

  char *ident;          /* ident string based on name (nname.name) */
  unsigned int has_sub; /* this menu has the sub menus? */

  NodeList items;       /* list of items. */
} MenuTopic;


typedef struct {
  NodeName nname;

  char *value;
} MenuValue;


typedef struct {
  MenuTopic *menu;  /* pointer to current menu */
  NodeList *topics; /* pointer menu topic list */
} MenuInfo;


static void loop_menu(MKLoop *ldata, MKInterp *interp, NodeList *stmts);
static void loop_item(MKLoop *ldata, MKInterp *interp, NodeList *stmts);
static void loop_menuitem(MKLoop *ldata, MKInterp *interp, NodeList *stmts);

static void local_var_add(MKInterp *interp, MenuItem *mitem);
static void local_var_remove(MKInterp *interp, MenuItem *mitem);

static void menus_load(const char *homepath, const char *hicktype, 
                       NodeList *list);
static void skip_variables(FILE *file);
static void menus_free(NodeList *list);
MenuTopic *topic_new(char *name, int has_sub, NodeList *list);
static char *create_ident(const char *s);
static int menuok(const char *name, MKInterp *interp);


char **
hickmk_menus_get(const char *menupath, size_t *siz)
{
  return class_get(menupath, MENU_FILE, MKC_MENU, siz);
}


void
hickmk_menus_remove(const char *homepath, const char *hicktype)
{
  char fname[FNAME_LEN+1];

  hsh_create_str(fname, FNAME_LEN, homepath, "/", MENU_DIR, "/", hicktype,
                 NULL);

  _unlinkRecursive(fname, NULL, NULL, NULL);
}


int
hickmk_menus_create(const char **names, unsigned int nameSize,
                    const char *menupath, const char *homepath,
                    const char *hicktype, const char *menusub)
{
  int n;
  char *p;
  Node *node;
  MKInterp interp;
  NodeList menus, funcs;
  char fname[FNAME_LEN+1], path[FNAME_LEN+1];

  hsh_nodelist_init(&menus);
  hsh_nodelist_init(&funcs);
  mkinterp_init(&interp);

  hsh_create_str(fname, FNAME_LEN, menupath, "/", MENU_FILE, NULL);

  if((n = mklang_functions_load(fname, &funcs))) {
    return n;
  }

  hsh_create_str(path, FNAME_LEN, homepath, "/", MENU_DIR, "/", hicktype, NULL);

  if(hsh_path_create(path, NULL) == IOW_FAIL) {
    return -1;
  }

  if(menusub) {
    p = MallocArr(char, strlen(menusub)+1);
    strcpy(p, menusub);

    topic_new(p, 1, &menus);
  }

  menus_load(homepath, hicktype, &menus);

  hsh_create_str(fname, FNAME_LEN, menupath, "/", hicktype, NULL);

  interp.userhome = homepath;
  interp.stdpath  = fname;

  for(node = funcs.head; node; node = node->next) {
    if(((MKFunction *)node)->fclass == MKC_MENU) {
      if(names) {
        for(n = 0; n < nameSize; n++) {
          if(!strcasecmp(names[n], ((MKFunction *)node)->name)) {
            break;
          }
        }

        if(n >= nameSize) {
          continue;
        }
      }

      mkloop_set(&interp, "menu",     loop_menu,     &menus);
      mkloop_set(&interp, "item",     loop_item,     NULL);
      mkloop_set(&interp, "menuitem", loop_menuitem, &menus);

      mkvar_set(&interp, "rootmenu", "",         NULL, NULL, 1, 0, 0);
      mkvar_set(&interp, "MENUSPATH", path,      NULL, NULL, 1, 0, 0);
      mkvar_set(&interp, "MENU",      "",        NULL, NULL, 1, 0, 0);
      mkvar_set(&interp, "MENUID",    "",        NULL, NULL, 1, 0, 0);
      mkvar_set(&interp, "ITEM",      "",        NULL, NULL, 1, 0, 0);
      mkvar_set(&interp, "ITEMID",    "",        NULL, NULL, 1, 0, 0);
      mkvar_set(&interp, "COMMAND",   "",        NULL, NULL, 1, 0, 0);
      mkvar_set(&interp, "TYPE",      "",        NULL, NULL, 1, 0, 0);
      mkvar_set(&interp, "FIRST",     "",        NULL, NULL, 1, 0, 0);
      mkvar_set(&interp, "LAST",      "",        NULL, NULL, 1, 0, 0);

      mkvar_set(&interp, HICKUPTYPE, hicktype,   NULL, NULL, 1, 0, 0);
      mkvar_set(&interp, "USER",     hmk_getlogin(), NULL, NULL, 1, 0, 0);
      mkvar_set(&interp, "PATH",     path,       NULL, NULL, 1, 0, 0);
      mkvar_set(&interp, "FILE",     "",         NULL, NULL, 1, 0, 0);

      mkinterp_exec_stmt(&((MKFunction *)node)->stmt, &interp);
      mkinterp_clear(&interp);
    }
  }

  menus_free(&menus);
  mklang_functions_free(&funcs);
  return 0;
}


static void
loop_menu(MKLoop *ldata, MKInterp *interp, NodeList *stmts)
{
  Node *node;
  MenuInfo info;

  info.topics = (NodeList *)ldata->data;
  mkloop_set(interp, "item", loop_item, &info);

  for(node = ((NodeList *)ldata->data)->head; node; node = node->next) {
    info.menu = (MenuTopic *)node;

    if(!menuok(((NodeName *)node)->name, interp)) {
      continue;
    }

    mkvar_fl_set(interp, info.topics, node);

    if(info.menu->has_sub) {
      mkvar_set(interp, "rootmenu", "1", NULL, NULL, 1, 0, 1);
    } else {
      mkvar_set(interp, "rootmenu", "", NULL, NULL, 1, 0, 1);
    }

    mkvar_set(interp, "MENU", ((NodeName *)node)->name, NULL, NULL, 1, 0, 1);
    mkvar_set(interp, "MENUID", ((MenuTopic *)node)->ident, NULL, NULL, 1,0,1);

    mkinterp_exec_stmts(stmts, interp);
  }

  mkloop_set(interp, "item", loop_item, NULL);
  mkvar_set(interp, "rootmenu", "", NULL, NULL, 1, 0, 1);
  mkvar_set(interp, "FIRST", "", NULL, NULL, 1, 0, 1);
  mkvar_set(interp, "LAST", "", NULL, NULL, 1, 0, 1);
  mkvar_set(interp, "MENU", "", NULL, NULL, 1, 0, 1);
  mkvar_set(interp, "MENUID", "", NULL, NULL, 1, 0, 1);
}


static void
loop_item(MKLoop *ldata, MKInterp *interp, NodeList *stmts)
{
  Node *node;
  MenuInfo *info;

  if(!(info = (MenuInfo *)ldata->data)) {
    return;
  }

  if(info->menu->has_sub) {
    mkvar_set(interp, "TYPE", "menu", NULL, NULL, 1, 0, 1);
    mkvar_set(interp, "COMMAND", "", NULL, NULL, 1, 0, 1);

    for(node = info->topics->head; node; node = node->next) {
      if(((MenuTopic *)node)->has_sub ||
         !menuok(((NodeName *)node)->name, interp)) {
        continue;
      }

      mkvar_fl_set(interp, info->topics, node);
      mkvar_set(interp, "ITEM", ((NodeName *)node)->name, NULL, NULL, 1,0,1);
      mkvar_set(interp, "ITEMID", ((MenuTopic *)node)->ident, NULL,NULL,1,0,1);
      mkinterp_exec_stmts(stmts, interp);
    }
  }

  mkvar_set(interp, "TYPE", "exec", NULL, NULL, 1, 0, 1);

  for(node = info->menu->items.head; node; node = node->next) {
    mkvar_fl_set(interp, &info->menu->items, node);
    mkvar_set(interp, "ITEM", ((MenuItem *)node)->name, NULL, NULL, 1, 0, 1);
    mkvar_set(interp, "ITEMID", ((MenuItem *)node)->ident, NULL, NULL, 1, 0,1);
    mkvar_set(interp, "COMMAND", ((MenuItem *)node)->command, NULL,NULL,1,0,1);
    local_var_add(interp, (MenuItem *)node);
    mkinterp_exec_stmts(stmts, interp);
    local_var_remove(interp, (MenuItem *)node);
  }

  mkvar_set(interp, "TYPE", "", NULL, NULL, 1, 0, 1);
  mkvar_set(interp, "ITEM", "", NULL, NULL, 1, 0, 1);
  mkvar_set(interp, "ITEMID", "", NULL, NULL, 1, 0, 1);
  mkvar_set(interp, "COMMAND", "", NULL, NULL, 1, 0, 1);
}


static void
loop_menuitem(MKLoop *ldata, MKInterp *interp, NodeList *stmts)
{
  Node *node, *inode;

  for(node = ((NodeList *)ldata->data)->head; node; node = node->next) {
    mkvar_set(interp, "MENU", ((NodeName *)node)->name, NULL, NULL, 1, 0, 1);
    mkvar_set(interp, "MENUID", ((MenuTopic *)node)->ident, NULL, NULL, 1,0,1);

    for(inode = ((MenuTopic *)node)->items.head; inode; inode = inode->next) {
      if((((NodeList *)ldata->data)->head == node) && 
         (((MenuTopic *)node)->items.head == inode)) {
        mkvar_set(interp, "FIRST", "1", NULL, NULL, 1, 0,1);
      } else {
        mkvar_set(interp, "FIRST", "", NULL, NULL, 1, 0,1);
      }

      if((((NodeList *)ldata->data)->tail == node) && 
         (((MenuTopic *)node)->items.tail == inode)) {
        mkvar_set(interp, "LAST", "1", NULL, NULL, 1, 0,1);
      } else {
        mkvar_set(interp, "LAST", "", NULL, NULL, 1, 0,1);
      }

      mkvar_set(interp, "ITEM", ((MenuItem *)inode)->name, NULL, NULL, 1, 0,1);
      mkvar_set(interp, "ITEMID", ((MenuItem *)inode)->ident, NULL,NULL,1,0,1);
      mkvar_set(interp, "COMMAND", ((MenuItem *)inode)->command, NULL, NULL,
                1, 0, 1);

      local_var_add(interp, (MenuItem *)inode);
      mkinterp_exec_stmts(stmts, interp);
      local_var_remove(interp, (MenuItem *)inode);
    }
  }

  mkvar_set(interp, "FIRST", "", NULL, NULL, 1, 0, 1);
  mkvar_set(interp, "LAST", "", NULL, NULL, 1, 0, 1);
  mkvar_set(interp, "MENU", "", NULL, NULL, 1, 0, 1);
  mkvar_set(interp, "MENUID", "", NULL, NULL, 1, 0, 1);
  mkvar_set(interp, "ITEM", "", NULL, NULL, 1, 0, 1);
  mkvar_set(interp, "ITEMID", "", NULL, NULL, 1, 0, 1);
  mkvar_set(interp, "COMMAND", "", NULL, NULL, 1, 0, 1); 
}


static void
local_var_add(MKInterp *interp, MenuItem *mitem)
{
  Node *node;
  MenuValue *mvalue;
  MKVariable *mkvar;

  for(node = mitem->variables.head; node; node = node->next){
    mvalue = (MenuValue *)node;

    if((mkvar = mkvar_find(interp, mvalue->nname.name)) && mkvar->read_only) {
      continue;
    }
     
    mkvar_set(interp, mvalue->nname.name, mvalue->value, NULL, NULL, 1, 5, 1);
  }
}


static void
local_var_remove(MKInterp *interp, MenuItem *mitem)
{
  mkvar_temp_unset(interp, 5);
}


static void
menus_load(const char *homepath, const char *hicktype, NodeList *list)
{
  FILE *file;
  int priority;
  MenuItem *item;
  MenuValue *value;
  MenuTopic *topic;
  char fname[FNAME_LEN+1], *s1, *s2, *s3, c, *name, *val;

  hsh_create_str(fname, FNAME_LEN, homepath, "/.", hicktype, "/", MENUS_FILE,
                 NULL);

  if(!(file = fopen(fname, "r"))) {
    return;
  }

  while((c = _getSpaces(file)) != EOF) {
    ungetc(c, file);

    if(!(s1 = _getCString(file))) { /* priority level */
      break;
    }

    priority = atol(s1);
    FreeMem(s1);

    if(!(s1 = _getCString(file))) { /* menu topic */
      break;
    }

    if(!(s2 = _getCString(file))) { /* item name */
      FreeMem(s1);
      break;
    }

    if(!(s3 = _getCString(file))) { /* command name */
      FreeMem(s1);
      FreeMem(s2);
      break;
    }

    if((c = fgetc(file)) != '{') { /* something is wrong here. */
      skip_variables(file);
      FreeMem(s1);
      FreeMem(s2);
      FreeMem(s3);
      continue;
    }

    if(!(topic = (MenuTopic *)hsh_node_name_find(s1, list))) {
      topic = topic_new(s1, 0, list);
    }

    item = MallocMem(MenuItem);
    item->nnum.num = priority;
    item->name = s2;
    item->ident = create_ident(s2);
    item->command = s3;
    hsh_nodelist_init(&item->variables);

    hsh_node_num_insert(&item->nnum, &topic->items, 1);

    while(((c = fgetc(file)) != EOF) && (c != '}')) {
      if(c == '"') {
        ungetc(c, file);
      } else {
        if(c != ':') {
          skip_variables(file);
          break;
        }
      }

      if(!(name = _getString(file))) {
        break;
      }

      if((fgetc(file) == EOF) || !(val = _getString(file))) {
        FreeMem(name);
        break;
      }

      value = MallocMem(MenuValue);
      value->nname.name = name;
      value->value = val;
      hsh_node_append(&value->nname.node, &item->variables);
    }
  }

  fclose(file);
}


static void
skip_variables(FILE *file)
{
  char c;

  while(!feof(file) && (((c = fgetc(file))) != '}') && (c != '\n')) ;
}


static void
menus_free(NodeList *list)
{
  Node *n1, *n2, *n3;

  for(n1 = list->head; n1; n1 = list->head) {
    list->head = n1->next;

    for(n2 = ((MenuTopic *)n1)->items.head; n2; 
        n2 = ((MenuTopic *)n1)->items.head) {

      ((MenuTopic *)n1)->items.head = n2->next;

      for(n3 = ((MenuItem *)n2)->variables.head; n3; 
          n3 = ((MenuItem *)n2)->variables.head) {

        ((MenuItem *)n2)->variables.head = n3->next;

        FreeMem(((MenuValue *)n3)->nname.name);
        FreeMem(((MenuValue *)n3)->value);
        FreeMem((MenuValue *)n3);
      }

      FreeMem(((MenuItem *)n2)->name);
      FreeMem(((MenuItem *)n2)->ident);
      FreeMem(((MenuItem *)n2)->command);
      FreeMem((MenuItem *)n2);
    }

    FreeMem(((MenuTopic *)n1)->ident);
    FreeMem(((MenuTopic *)n1)->nname.name);
    FreeMem((MenuTopic *)n1);
  }
}


MenuTopic *
topic_new(char *name, int has_sub, NodeList *list)
{
  MenuTopic *topic;

  topic = MallocMem(MenuTopic);
  topic->nname.name = name;
  topic->ident = create_ident(name);
  topic->has_sub = has_sub;
  hsh_nodelist_init(&topic->items);

  hsh_node_name_insert(&topic->nname, list, 0);
  return topic;
}


static char *
create_ident(const char *s)
{
  char *p;
  size_t siz, n;

  siz = strlen(s) + 1;
  p = MallocArr(char, siz);

  for(n = 0; *s; s++) {
    if(isalnum((int)*s)) {
      p[n++] = tolower((int)*s);
    }
  }

  p[n] = (char)0;
  return p;
}


static int
menuok(const char *name, MKInterp *interp)
{
  char *p, *str;
  MKVariable *tmpvar;

  if((tmpvar = mkvar_find(interp, "nomenu"))) {
    str = MallocArr(char, strlen(tmpvar->value.buf) + 1);
    strcpy(str, tmpvar->value.buf);

    p = strtok(str, ":");
      
    while(p) {
      if(!strcasecmp(p, name)) {
        FreeMem(str);
        return 0;
      }

      p = strtok(0, ":");
    }

    FreeMem(str);
  }

  return 1;
}
