/* HickMK 1.2.2 -- HICK Make Utility.
** Copyright (C) 1998 - 2001 Daniel Bergstrom <daniel@octocode.com>
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the Free
** Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <strings.h>

#include "hicksh.h"
#include "linker.h"
#include "mkinterp.h"

#define READ_BUF  1024

static int mkexpr_exec(MKExpr *expr, MKInterp *interp);
static MKLoop *mkloop_find(MKInterp *interp, const char *name);
static MKVariable *mkvar_add(MKInterp *interp, const char *name, 
                             const char *value);


int
mkinterp_exec(const char *function, enum MKClass fclass, NodeList *funcs, 
              MKInterp *interp)
{
  Node *node;

  for(node = funcs->head; node; node = node->next) {
    if((fclass == ((MKFunction *)node)->fclass) && 
       !strcasecmp(function, ((MKFunction *)node)->name)) {

      mkinterp_exec_stmt(&((MKFunction *)node)->stmt, interp);
      return 1;
    }
  }

  return 0;
}


void
mkinterp_exec_stmts(NodeList *stmts, MKInterp *interp)
{
  Node *node;

  for(node = stmts->head; node; node = node->next) {
    mkinterp_exec_stmt((MKStmt *)node, interp);
  }
}


void
mkinterp_exec_stmt(MKStmt *stmt, MKInterp *interp)
{
  int n, first, last;
  size_t siz, len;
  FILE *file;
  Node *node;
  SBuf tmpsb;
  MKLoop *ldata;
  MKVariable *var;
  char *p, buf[READ_BUF+1], *oldpath;

  switch(stmt->token) {
    case MKT_DIR: {
      interp->buf.siz = 0;
      p = mkparse_string(stmt->value, interp);
      oldpath = NULL;
      len = 0;

      if(!((p[0] == '.') && (p[1] == (char)0))) {
        if((var = mkvar_find(interp, "PATH"))) {
          oldpath = MallocArr(char, (len = var->value.siz + 1));
          strcpy(oldpath, var->value.buf);
        }

        var = mkvar_add(interp, "PATH", p);
      } else {
        var = mkvar_find(interp, "PATH");
      }

      if(!var || (hsh_path_create(var->value.buf, NULL) == IOW_FAIL)) {
        mklang_error(MKTInterp, MKError, "failed to create directory '%s'.\n",
                     p);
        return;
      }

      hsh_chmod(var->value.buf, stmt->u.mode, NULL);
      mkinterp_exec_stmts(&stmt->stmt1, interp);

      if(oldpath) {
        FreeMem(var->value.buf);
        var->value.buf = oldpath;
        var->value.alloc = len;
        var->value.siz = len - 1;
      }
      break;
    }
    case MKT_FILE: {
      interp->buf.siz = 0;
      p = mkparse_string(stmt->value, interp);

      if(!(var = mkvar_find(interp, "PATH"))) {
        return;
      }

      sbuf_init(&tmpsb);
      sbuf_putsb(&var->value, &tmpsb);
      sbuf_put('/', &tmpsb);
      sbuf_putsb(&interp->buf, &tmpsb);
      sbuf_put((char)0, &tmpsb);

      if(!(var = mkvar_find(interp, "FILE"))) {
        mkvar_set(interp, "FILE", "", NULL, NULL, 1, 0, 1);

        if(!(var = mkvar_find(interp, "FILE"))) {
          return;
        }
      }

      sbuf_clear(&var->value);
      sbuf_move(&tmpsb, &var->value);

      if(!(interp->out = fopen(var->value.buf, "w"))) {
        mklang_error(MKTInterp, MKError, "failed to create file '%s'.\n", 
                     var->value.buf);
        return;
      }

      hsh_chmod(var->value.buf, stmt->u.mode, NULL);

      mkinterp_exec_stmts(&stmt->stmt1, interp);
      fclose(interp->out);
      break;
    }
    case MKT_FOREACH: {
      p = mkparse_string(stmt->value, interp);

      if(!(ldata = mkloop_find(interp, p))) {
        mklang_error(MKTInterp, MKError, "loop variable '%s' not found.\n", p);
        return;
      }

      mkvar_fl_store(interp, &first, &last);
      (ldata->lfunc)(ldata, interp, &stmt->stmt1);
      mkvar_fl_restore(interp, first, last);
      break;
    }
    case MKT_READ: {
      for(node = stmt->stmt1.head; node; node = node->next) {
        p = mkparse_string(((MKStmt *)node)->value, interp);

        _expandPath(buf, READ_BUF, p, interp->userhome, interp->stdpath);

        if((file = fopen(buf, "r"))) {
          while((siz = fread(buf, 1, READ_BUF, file))) {
            fwrite(buf, 1, siz, interp->out);
          }

          fclose(file);
          break;
        }
      }

      break;
    }
    case MKT_WRITE: {
      mkparse_string(stmt->value, interp);

      if(interp->buf.siz) {
        fwrite(interp->buf.buf, 1, interp->buf.siz - 1, interp->out);
      }
      break;
    }
    case MKT_IF: {
      if(mkexpr_exec(stmt->u.expr, interp)) {
        mkinterp_exec_stmts(&stmt->stmt1, interp);
      } else {
        mkinterp_exec_stmts(&stmt->stmt2, interp);
      }
      break;
    }
    case MKT_UNSET: {
      p = mkparse_string(stmt->value, interp);

      if((n = mkvar_unset(interp, p, 0))) {
        if(n == -1) { /* not found */
          mklang_error(MKTInterp, MKWarning, "variable '%s' was not found.\n", 
                       p);
        } else { /* read-only */
          mklang_error(MKTInterp, MKWarning, "variable '%s' is read-only.\n",
                       p);
        }
      }
      break;
    }
    case MKT_EXPR: {
      mkexpr_exec(stmt->u.expr, interp);
      break;
    }
  }
}


static int
mkexpr_exec(MKExpr *expr, MKInterp *interp)
{
  int ret = 0;
  SBuf tmpsb;
  const char *lid, *rid;

  sbuf_move(&interp->buf, &tmpsb);

  if(expr->op == MKO_ID) { /* TRUE: ID exist, FALSE: ID doesn't exist */
    lid = mkparse_string(expr->value, interp);

    ret = (lid[0] ? 1 : 0);
  } else {
    lid = mkparse_string(((MKExpr *)expr->node.next)->value, interp);
    sbuf_init(&interp->buf);

    rid = mkparse_string(((MKExpr *)expr->node.prev)->value, interp);  

    switch(expr->op) {
      case MKO_EQUAL: { /* TRUE: IDs equal, FALSE: IDs not equal */
        ret = (strcmp(lid, rid) ? 0 : 1);
        break;
      }
      case MKO_NOT_EQUAL: { /* TRUE: IDs not equal, FALSE: IDs equal */
        ret = (strcmp(lid, rid) ? 1 : 0);
        break;
      }
      case MKO_SET: { /* TRUE: successfull, FALSE: fauled */
        ret = (mkvar_set(interp, lid, rid, NULL, NULL, 0, 1, 0) ? 0 : 1);
        break;
      }
    }

    FreeMem(rid);
  }

  FreeMem(lid);
  sbuf_move(&tmpsb, &interp->buf);

  return ret;
}


int
mkloop_set(MKInterp *interp, const char *name, MKLoopFunc lfunc, void *data)
{
  MKLoop *ldata;

  if(!(ldata = mkloop_find(interp, name))) {
    if(!(ldata = MallocMem(MKLoop))) {
      return 0;
    }

    if(!(ldata->name = MallocArr(char, strlen(name) + 1))) {
      FreeMem(ldata);
      return 0;
    }

    strcpy(ldata->name, name);
    hsh_node_append(&ldata->node, &interp->loops);
  }

  ldata->lfunc = lfunc;
  ldata->data = data;
  return 1;
}


static MKLoop *
mkloop_find(MKInterp *interp, const char *name)
{
  Node *node;

  for(node = interp->loops.head; node; node = node->next) {
    if(!strcasecmp(((MKLoop *)node)->name, name)) {
      return (MKLoop *)node;
    }
  }

  return NULL;
}


static MKVariable *
mkvar_add(MKInterp *interp, const char *name, const char *value)
{
  MKVariable *var;

  if(!(var = mkvar_find(interp, name))) {
    mkvar_set(interp, name, value, NULL, NULL, 1, 0, 1);
    return mkvar_find(interp, name);
  } else {
    var->vfunc = NULL;

    if(var->value.siz) {
      var->value.siz--;
    }

    sbuf_put('/', &var->value);
    sbuf_puts(value, &var->value);
    sbuf_put((char)0, &var->value);
  }

  return var;
}


/* mkvar_set(): 0 - successfull, -1 - out of memory, -2 - read-only */

int
mkvar_set(MKInterp *interp, const char *name, const char *value, 
          MKValueFunc vfunc, void *data, int readonly, int temp_level, 
          int override)
{
  MKVariable *var;

  if(!(var = mkvar_find(interp, name))) {
    if(!(var = MallocMem(MKVariable))) {
      return -1;
    }

    if(!(var->name = MallocArr(char, strlen(name) + 1))) {
      FreeMem(var);
      return -1;
    }

    strcpy(var->name, name);
    sbuf_init(&var->value);
  } else {
    if(!override && var->read_only) {
      return -2;
    }

    hsh_node_remove(&var->node, &interp->variables);
  }

  var->data = data;
  var->vfunc = vfunc;

  if(vfunc) {
    sbuf_clear(&var->value);
  } else {
    var->value.siz = 0;
    sbuf_puts(value, &var->value);
    sbuf_put((char)0, &var->value);
  }

  var->temp_level = temp_level;
  var->read_only = readonly;
  hsh_node_append(&var->node, &interp->variables);
  return 0;
}


MKVariable *
mkvar_find(MKInterp *interp, const char *name)
{
  Node *node;

  for(node = interp->variables.head; node; node = node->next) {
    if(!strcasecmp(((MKVariable *)node)->name, name)) {
      return (MKVariable *)node;
    }
  }

  return NULL;
}

/* mkvar_unset(): 0 - successfull, -1 - not found, -2 - read-only */

int
mkvar_unset(MKInterp *interp, const char *name, int override)
{
  MKVariable *var;

  if(!(var = mkvar_find(interp, name))) {
    return -1;
  }

  if(!override && var->read_only) {
    return -2;
  }

  hsh_node_remove(&var->node, &interp->variables);

  if(var->value.buf) {
    FreeMem(var->value.buf);
  }

  FreeMem(var->name);
  FreeMem(var);
  return 0;
}


void 
mkvar_temp_unset(MKInterp *interp, int temp_level)
{
  Node *node, *tnode;

  for(node = interp->variables.head; node; node = tnode) {
    tnode = node->next;

    if(((MKVariable *)node)->temp_level == temp_level) {
      hsh_node_remove(node, &interp->variables);

      sbuf_clear(&((MKVariable *)node)->value);    
      FreeMem(((MKVariable *)node)->name);
      FreeMem(node);
    }
  }
}


void
mkvar_fl_store(MKInterp *interp, int *first, int *last)
{
  MKVariable *var;

  *first = *last = 0;

  if((var = mkvar_find(interp, "FIRST"))) {
    if(var->value.buf[0] == '1') {
      *first = 1;
    }
  }

  if((var = mkvar_find(interp, "LAST"))) {
    if(var->value.buf[0] == '1') {
      *last = 1;
    }
  }
}

void
mkvar_fl_restore(MKInterp *interp, int first, int last)
{
  mkvar_set(interp, "FIRST", (first ? "1" : ""), NULL, NULL, 1, 0, 1);
  mkvar_set(interp, "LAST", (last ? "1" : ""), NULL, NULL, 1, 0, 1);
}


void
mkvar_fl_set(MKInterp *interp, NodeList *list, Node *node)
{
  mkvar_set(interp, "FIRST", (list->head == node ? "1" : ""), NULL, NULL, 
            1, 0, 1);
  mkvar_set(interp, "LAST", (list->tail == node ? "1" : ""), NULL, NULL, 
            1, 0, 1);
}


void 
mkinterp_clear(MKInterp *interp)
{
  Node *node;

  for(node = interp->variables.head; node; node = interp->variables.head) {
    interp->variables.head = node->next;

    sbuf_clear(&((MKVariable *)node)->value);    
    FreeMem(((MKVariable *)node)->name);
    FreeMem(node);
  }

  for(node = interp->loops.head; node; node = interp->loops.head) {
    interp->loops.head = node->next;

    FreeMem(((MKLoop *)node)->name);
    FreeMem(node);
  }

  sbuf_clear(&interp->buf);

  mkinterp_init(interp);
}
