/* HickMK 1.2.2 -- HICK Make Utility.
** Copyright (C) 1998 - 2001 Daniel Bergstrom <daniel@octocode.com>
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the Free
** Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <unistd.h>

#include "common.h"
#include "hickmk.h"

#include "utils.h"
#include "mkinterp.h"

/* ScriptValue ************************************************************* */

typedef struct {
  NodeNum nnum; /* sorted by priority number */

  char *value;
  size_t len;   /* only used when loading inline scripts */
} ScriptValue;

/* ScriptType ************************************************************** */

enum ScriptType {
  SETENV_NONE    = 0, /* indicate no valid type is selected */
  SETENV_PREPEND = 1, /* only used when loading */
  SETENV_APPEND  = 2, /* only used when loading */
  SETENV_PATH    = 3, /* contains prepend and append */
  SETENV_SET     = 4, /* sets value */
  SETENV_UNSET   = 5  /* unsets value */
};

/* ScriptVar *************************************************************** */

typedef struct {
  NodeName nname;       /* sorted by name */

  enum ScriptType type; /* what kind of type: path, set or unset */

  NodeList prepend;     /* prepend / set / unset paths */
  NodeList append;      /* append paths */
} ScriptVar;

/* InlineInfo ************************************************************** */

typedef struct {
  const char *path; /* path to where inline scripts are placed */
  const char *name;
} InlineInfo;


static int var_prepend(MKVariable *var, MKInterp *interp, int is_not, 
                       int test);
static int var_append(MKVariable *var, MKInterp *interp, int is_not, int test);
static int var_script(MKVariable *var, MKInterp *interp, int is_not, int test);

static void loop_env(MKLoop *ldata, MKInterp *interp, NodeList *stmts);
static void loop_script(MKLoop *ldata, MKInterp *interp, NodeList *stmts);

static void environ_load(const char *filename, NodeList *list);
static void environ_sort_set(NodeList *list, unsigned int n);
static void environ_free(NodeList *list);

static void script_value_free_list(NodeList *list);

static void scripts_load(const char *path, const char *name, int have_all, 
                         NodeList *list);


char **
hickmk_scripts_get(const char *scriptpath, size_t *siz) 
{
  return class_get(scriptpath, SCRIPT_FILE, MKC_SCRIPT, siz);
}


void
hickmk_scripts_remove(const char *homepath, const char *hicktype)
{
  char fname[FNAME_LEN+1];

  hsh_create_str(fname, FNAME_LEN, homepath, "/", SCRIPTS_DIR, "/",hicktype,
                 NULL);

  _unlinkRecursive(fname, NULL, NULL, NULL);
}


int
hickmk_scripts_create(const char **names, unsigned int nameSize,
                      const char *scriptpath, const char *homepath,
                      const char *hicktype)
{
  int n;
  Node *node;
  MKInterp interp;
  InlineInfo info;
  char fname[FNAME_LEN+1], path[FNAME_LEN+1];
  NodeList environ, funcs;

  hsh_nodelist_init(&funcs);
  hsh_nodelist_init(&environ);
  mkinterp_init(&interp);

  hsh_create_str(fname, FNAME_LEN, scriptpath, "/", SCRIPT_FILE, NULL);

  if((n = mklang_functions_load(fname, &funcs))) {
    return n;
  }

  hsh_create_str(path, FNAME_LEN, homepath, "/", SCRIPTS_DIR, "/", hicktype,
                 NULL);

  if(hsh_path_create(path, NULL) == IOW_FAIL) {
    return -1;
  }

  hsh_create_str(fname, FNAME_LEN, homepath, "/.", hicktype, "/", 
                 ENVIRON_FILE, NULL);

  environ_load(fname, &environ);

  hsh_create_str(fname, FNAME_LEN, scriptpath, "/", hicktype, NULL);

  interp.userhome = homepath;
  interp.stdpath  = fname;

  hsh_create_str(fname, FNAME_LEN, homepath, "/.", hicktype, "/", SCRIPTS_DIR,
                 NULL);

  info.path = fname;

  for(node = funcs.head; node; node = node->next) {
    if(((MKFunction *)node)->fclass == MKC_SCRIPT) {
      info.name = ((MKFunction *)node)->name;

      if(names) {
        for(n = 0; n < nameSize; n++) {
          if(!strcasecmp(names[n], info.name)) {
            break;
          }
        }

        if(n >= nameSize) {
          continue;
        }
      }

      mkloop_set(&interp, "env",    loop_env, &environ);
      mkloop_set(&interp, "script", loop_script, &info);

      mkvar_set(&interp, "separator",   ":",        NULL, NULL, 0, 0, 0);
      mkvar_set(&interp, "include_all", "",         NULL, NULL, 0, 0, 0);
      mkvar_set(&interp, "NAME",        "",         NULL, NULL, 1, 0, 0);
      mkvar_set(&interp, "TYPE",        "",         NULL, NULL, 1, 0, 0);
      mkvar_set(&interp, HICKUPTYPE,    hicktype,   NULL, NULL, 1, 0, 0);
      mkvar_set(&interp, "USER",        hmk_getlogin(), NULL, NULL, 1, 0, 0);
      mkvar_set(&interp, "PATH",        path,       NULL, NULL, 1, 0, 0);
      mkvar_set(&interp, "FILE",        "",         NULL, NULL, 1, 0, 0);
      mkvar_set(&interp, "PREPEND",     NULL, var_prepend, NULL, 1, 0, 0);
      mkvar_set(&interp, "APPEND",      NULL, var_append,  NULL, 1, 0, 0);
      mkvar_set(&interp, "SCRIPT",      NULL, var_script,  NULL, 1, 0, 0);

      mkinterp_exec_stmt(&((MKFunction *)node)->stmt, &interp);
      mkinterp_clear(&interp);
    }
  }

  environ_free(&environ);
  mklang_functions_free(&funcs);
  return 0;
}


static int
var_prepend(MKVariable *var, MKInterp *interp, int is_not, int test)
{
  Node *node;
  ScriptVar *scvar;
  MKVariable *tmpvar;
  const char *separator = ":";
  
  if(!(scvar = (ScriptVar *)var->data)) {
    return 0;
  }

  if(test) {
    return (!is_not && scvar->prepend.head ? 1 : 
            (is_not && !scvar->prepend.head ? 1 : 0));
  }

  if(!scvar->prepend.head) {
    return 0;
  }

  if(scvar->type == SETENV_PATH) {
    if((tmpvar = mkvar_find(interp, "separator"))) {
      separator = tmpvar->value.buf;
    }

    for(node = scvar->prepend.head; node; node = node->next) {
      sbuf_puts(((ScriptValue *)node)->value, &interp->buf);

      if(node->next) {
        sbuf_puts(separator, &interp->buf);
      }
    }

    return 0;
  }

  if(scvar->type == SETENV_SET) {
    sbuf_puts(((ScriptValue *)scvar->prepend.head)->value, &interp->buf);
    return 0;
  }

  if(scvar->type == SETENV_UNSET) {
    sbuf_puts(((ScriptValue *)scvar->prepend.head)->value, &interp->buf);
    return 0;
  }

  return 0;
}


static int
var_append(MKVariable *var, MKInterp *interp, int is_not, int test)
{
  Node *node;
  ScriptVar *scvar;
  MKVariable *tmpvar;
  const char *separator = ":";

  if(!var->data) {
    return (is_not ? 1 : 0);
  }

  scvar = (ScriptVar *)var->data;

  if(test) {
    return (!is_not && scvar->append.head ? 1 : 
            (is_not && !scvar->append.head ? 1 : 0));
  }

  if(scvar->type == SETENV_PATH) {
    if((tmpvar = mkvar_find(interp, "separator"))) {
      separator = tmpvar->value.buf;
    }

    for(node = scvar->append.head; node; node = node->next) {
      sbuf_puts(((ScriptValue *)node)->value, &interp->buf);

      if(node->next) {
        sbuf_puts(separator, &interp->buf);
      }
    }

    return 0;
  }

  return 0;
}


static int 
var_script(MKVariable *var, MKInterp *interp, int is_not, int test)
{
  if(!var->data) {
    return (is_not ? 1 : 0);
  }

  if(test) {
    return (is_not ? 0 : 1);
  }

  sbuf_puts(((ScriptValue *)var->data)->value, &interp->buf);
  return 0;
}


static void
loop_env(MKLoop *ldata, MKInterp *interp, NodeList *stmts)
{
  Node *node;
  ScriptVar *scvar;
  int state = SETENV_SET;

  do {
    for(node = ((NodeList *)ldata->data)->head; node; node = node->next) {
      scvar = (ScriptVar *)node;

      if(scvar->type != state) {
        continue;
      }

      mkvar_set(interp, "NAME", scvar->nname.name, NULL, NULL, 1, 0, 1);
      mkvar_set(interp, "TYPE", (scvar->type == SETENV_PATH ? "path" :
                                 (scvar->type == SETENV_SET ? "set": "unset")),
                NULL, NULL, 1, 0, 1);
      mkvar_set(interp, "PREPEND", NULL, var_prepend, scvar, 1, 0, 1);
      mkvar_set(interp, "APPEND",  NULL, var_append,  scvar, 1, 0, 1);

      mkinterp_exec_stmts(stmts, interp);
    }

    state = (state == SETENV_SET ? SETENV_PATH : 
             (state == SETENV_PATH ? SETENV_UNSET : SETENV_NONE));
  } while(state);

  mkvar_set(interp, "NAME",    "", NULL, NULL, 1, 0, 1);
  mkvar_set(interp, "TYPE",    "", NULL, NULL, 1, 0, 1);
  mkvar_set(interp, "PREPEND", NULL, var_prepend, NULL, 1, 0, 1);
  mkvar_set(interp, "APPEND",  NULL, var_append,  NULL, 1, 0, 1);
}


static void
loop_script(MKLoop *ldata, MKInterp *interp, NodeList *stmts)
{
  Node *node;
  int have_all;
  NodeList list;
  const char *name;
  MKVariable *tmpvar;
  InlineInfo *info = (InlineInfo *)ldata->data;

  hsh_nodelist_init(&list);    

  if((tmpvar = mkvar_find(interp, "script_name"))) {
    name = tmpvar->value.buf;
  } else {
    name = info->name;
  }

  have_all = (mkvar_find(interp, "include_all") ? 1 : 0);

  scripts_load(info->path, name, have_all, &list);

  for(node = list.head; node; node = node->next) {
    mkvar_set(interp, "SCRIPT", NULL, var_script, (ScriptValue *)node, 1, 0,1);

    mkinterp_exec_stmts(stmts, interp);
  }

  mkvar_set(interp, "SCRIPT",    "", NULL, NULL, 1, 0, 1);

  script_value_free_list(&list);
}


static void
environ_load(const char *filename, NodeList *list)
{
  FILE *file;
  Node *node;
  ScriptVar *var;
  ScriptValue *value;
  char *s1, *s2, *s3;
  enum ScriptType type;
  unsigned int priority, n;

  if(!(file = fopen(filename, "r"))) {
    return;
  }

  while((s1 = _getCString(file))) { /* Priority level */
    priority = atol(s1);
    FreeMem(s1);

    if(!(s1 = _getCString(file))) { /* Flag: prepend, append or set */
      continue;
    }

    type = (!strcasecmp(s1, "prepend") ? SETENV_PREPEND :
            (!strcasecmp(s1, "append") ? SETENV_APPEND :
             (!strcasecmp(s1, "set") ? SETENV_SET :
              (!strcasecmp(s1, "unset") ? SETENV_UNSET : SETENV_NONE))));

    FreeMem(s1);

    if(!(s2 = _getCString(file))) { /* Variable name */
      continue;
    }

    if(!(s3 = _getCString(file))) { /* Variable value */
      FreeMem(s2);
      continue;
    }

    if(type == SETENV_NONE) {
      FreeMem(s2);
      FreeMem(s3);
      continue;
    }

    if(!(var = (ScriptVar *)hsh_node_name_find(s2, list))) {
      var = MallocMem(ScriptVar);
      var->type = (type == SETENV_SET ? SETENV_SET : 
                   (type == SETENV_UNSET ? SETENV_UNSET : SETENV_PATH));
      var->nname.name = s2;
      hsh_nodelist_init(&var->prepend);
      hsh_nodelist_init(&var->append);

      hsh_node_name_insert(&var->nname, list, 0);
    } else {
      FreeMem(s2);

      if(((var->type != SETENV_PATH) && 
          ((type == SETENV_APPEND) || (type == SETENV_APPEND))) ||
         ((var->type != SETENV_SET) && (type == SETENV_SET)) ||
         ((var->type != SETENV_UNSET) && (type == SETENV_UNSET))) {
        FreeMem(s3);
        continue;
      }
    }

    if((type == SETENV_APPEND) || (type == SETENV_PREPEND)) {
      for(node = var->append.head; node; node = node->next) {
        if(!strcmp(((ScriptValue *)node)->value, s3)) {
          break;
        }
      }

      if(node) {
        FreeMem(s3);
        continue;
      }
    }

    value = MallocMem(ScriptValue);
    value->nnum.num = priority;
    value->value = s3;

    if(type == SETENV_APPEND) {
      hsh_node_num_insert(&value->nnum, &var->append, 0);
    } else {
      hsh_node_num_insert(&value->nnum, &var->prepend, 1);
    }
  }

  for(n = 0, node = list->head; node; n++, node = node->next);

  environ_sort_set(list, n);

  fclose(file);
}

static void /* yes, this isn't perhaps the best sort routine... */
environ_sort_set(NodeList *list, unsigned int n)
{
  Node *node, *nnode;
  unsigned int i, npri, nnpri;

  if(!list->head || !list->head->next) {
    return;
  }

  for(i = 0; i < n; i++) {
    node = list->head;
    nnode = node->next;

    while(nnode) {
      npri = (((ScriptVar *)node)->type == SETENV_SET ? 
              ((ScriptValue *)((ScriptVar *)node)->prepend.head)->nnum.num : 
              (unsigned int)-1);
      nnpri = (((ScriptVar *)nnode)->type == SETENV_SET ? 
               ((ScriptValue *)((ScriptVar *)nnode)->prepend.head)->nnum.num :
               (unsigned int)-1);

      if(npri <= nnpri) {
        node = node->next;
        nnode = nnode->next;
        continue;
      }

      if(!node->prev) {         /* node is the first node in the list? */
        list->head = nnode;
      } else {
        node->prev->next = nnode;
      }

      nnode->prev = node->prev;

      if(!nnode->next) {        /* nnode is the last node in the list? */
        list->tail = node;
      } else {
        nnode->next->prev = node;
      }

      node->next = nnode->next;

      nnode->next = node;
      node->prev  = nnode;
      
      nnode = node->next;       /* get ready to check the next pair */
    }
  }
}


static void
environ_free(NodeList *list)
{
  Node *node;

  for(node = list->head; node; node = list->head) {
    list->head = node->next;

    script_value_free_list(&((ScriptVar *)node)->prepend);
    script_value_free_list(&((ScriptVar *)node)->append);

    FreeMem(((NodeName *)node)->name);
    FreeMem(node);
  }
}


static void
script_value_free_list(NodeList *list)
{
  Node *node;

  for(node = list->head; node; node = list->head) {
    list->head = node->next;

    FreeMem(((ScriptValue *)node)->value);
    FreeMem(node);
  }
}


static void
scripts_load(const char *path, const char *name, int have_all, NodeList *list)
{
  DIR *dirp;
  FILE *file;
  struct stat sbuf;
  ScriptValue *value;
  struct dirent *entry;
  unsigned int nlen, elen, i;
  char fname[FNAME_LEN+1], nbuf[FNAME_LEN+1], priStr[11];

  if(!(dirp = opendir(path))) {
    return;
  }

  nlen = strlen(name); /* length of the shell name */

  while((entry = readdir(dirp))) {
    hsh_create_str(fname, FNAME_LEN, path, "/", entry->d_name, NULL);

    if((hsh_stat(fname, &sbuf, NULL) == IOW_FAIL) || !S_ISREG(sbuf.st_mode) ||
       !sbuf.st_size) {
      continue;
    }

    elen = NAMLEN(entry);

    for(nlen = 0; (nlen < elen) && (entry->d_name[nlen] != '-'); ++nlen);

    if(entry->d_name[nlen] != '-') {
      continue;
    }

    strncpy(nbuf, entry->d_name, nlen);
    nbuf[nlen] = (char)0;

    if(strcasecmp(nbuf, name) &&
       (!have_all || (have_all && strcasecmp(nbuf, "all")))) {
      continue;
    }

    for(i = nlen; (i < elen) && (entry->d_name[i] != '.'); ++i);

    if(i >= elen) { /* didn't find any dot..  */
      continue;
    }

    strncpy(priStr, entry->d_name + i + 1, 11);
    priStr[11] = (char)0;

    if(!(file = fopen(fname, "r"))) { /* open the file for reading */
      continue;
    }

    value = MallocMem(ScriptValue);
    value->nnum.num = atoi(priStr);
    value->value = MallocArr(char, sbuf.st_size + 1);
    value->len = sbuf.st_size;

    fread(value->value, sbuf.st_size, 1, file);
    value->value[value->len] = (char)0;

    fclose(file);

    hsh_node_num_insert(&value->nnum, list, 1);
  }

  closedir(dirp);
}
