/* HickSH 1.1.2 -- Shared routines between hick parts.
** Copyright (C) 1999 - 2001 Daniel Bergstrom <daniel@octocode.com>
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the Free
** Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef STDC_HEADERS
# include <string.h>
#endif

#include <errno.h>

#include "hicksh.h"

#define NEW_DIR_MODE  (S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH)

static int write_error(int err, const char *name, const char *path,
                       FILE *logfile);


int
hsh_mkdir(const char *path, mode_t mode, FILE *logfile)
{
  int serrno;
  struct stat sbuf;

  while(mkdir(path, mode)) {
    if(errno == EEXIST) {
      serrno = errno;

      if(hsh_stat(path, &sbuf, logfile) || !S_ISDIR(sbuf.st_mode)) {
        write_error(serrno, "hsh_mkdir", path, logfile);
        return IOW_FAIL;
      }

      return IOW_OK;
    }

    if(write_error(errno, "hsh_mkdir", path, logfile)) {
      return IOW_FAIL;
    }
  }

  return IOW_OK;
}


int
hsh_rmdir(const char *path, FILE *logfile)
{
  while(rmdir(path) && (errno != ENOENT)) {
    if(write_error(errno, "hsh_rmdir", path, logfile)) {
      return IOW_FAIL;
    }
  }

  return IOW_OK;
}


int
hsh_unlink(const char *path, FILE *logfile)
{
  while(unlink(path) && (errno != ENOENT)) {
    if(write_error(errno, "hsh_unlink", path, logfile)) {
      return IOW_FAIL;
    }
  }

  return IOW_OK;
}


int
hsh_symlink(const char *name1, const char *name2, FILE *logfile)
{
  while(symlink(name1, name2)) {
    if(write_error(errno, "hsh_symlink", name2, logfile)) {
      return IOW_FAIL;
    }
  }

  return IOW_OK;
}


int
hsh_rename(const char *oname, const char *nname, FILE *logfile)
{
  while(rename(oname, nname)) {
    if(write_error(errno, "hsh_rename", oname, logfile)) {
      return IOW_FAIL;
    }
  }

  return IOW_OK;
}


char *
hsh_getcwd(char *buf, size_t size, FILE *logfile)
{
  char *p;

  while((p = getcwd(buf, size))) {
    if(write_error(errno, "hsh_getcwd", buf, logfile)) {
      return NULL;
    }
  }

  return p;
}


int
hsh_chdir(const char *path, FILE *logfile)
{
  while(chdir(path)) {
    if(write_error(errno, "hsh_chdir", path, logfile)) {
      return IOW_FAIL;
    }
  }

  return IOW_OK;
}


int
hsh_stat(const char *path, struct stat *sbuf, FILE *logfile)
{
  while(stat(path, sbuf)) {
    if(write_error(errno, "hsh_stat", path, logfile)) {
      return IOW_FAIL;
    }
  }

  return IOW_OK;
}


int
hsh_lstat(const char *path, struct stat *sbuf, FILE *logfile)
{
  while(lstat(path, sbuf)) {
    if(write_error(errno, "hsh_lstat", path, logfile)) {
      return IOW_FAIL;
    }
  }

  return IOW_OK;
}


int
hsh_chmod(const char *path, mode_t mode, FILE *logfile)
{
  while(chmod(path, mode)) {
    if(write_error(errno, "hsh_chmod", path, logfile)) {
      return IOW_FAIL;
    }
  }

  return IOW_OK;
}


DIR *
hsh_opendir(const char *path, FILE *logfile)
{
  DIR *dirp;

  while(!(dirp = opendir(path)) && errno) {
    if(write_error(errno, "hsh_opendir", path, logfile)) {
      return NULL;
    }
  }

  return dirp;
}


int
hsh_path_create(const char *path, FILE *logfile)
{
  size_t n, delim;
  struct stat sbuf;
  char fname[FNAME_LEN+1];

  if(path && *path) {
    if((hsh_stat(path, &sbuf, NULL) == IOW_OK) && S_ISDIR(sbuf.st_mode)) {
      return IOW_OK;
    }

    for(n = 0, delim = 0; (n < FNAME_LEN) && (fname[n] = path[n]); n++) {
      if((path[n] == '/') && (!n || (n && (path[n - 1] != '/')))) {
        delim = n;
      }
    }

    fname[delim] = (char)0;

    if(hsh_path_create(fname, logfile) == IOW_OK) {
      return hsh_mkdir(path, NEW_DIR_MODE, logfile);
    }
  }

  return IOW_FAIL;
}


static int
write_error(int err, const char *name, const char *path, FILE *logfile)
{
  const char *p;

  if(logfile && (p = strerror(err))) {
    if(path) {
      fprintf(logfile, "%s: %s: %s (%s).\n", name,
              (err == ESTALE ? "warning" : "error"), p, path);
    } else {
      fprintf(logfile, "%s: %s: %s.\n", name,
              (err == ESTALE ? "warning" : "error"), p);
    }
  }

  return (err != ESTALE);
}
