/* Memory Debugger 1.2.1 - Runtime memory debugger.
** Copyright (C) 1999 - 2001 Daniel Bergstrom <daniel@octocode.com>
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the Free
** Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <errno.h>

#if STDC_HEADERS
# include <string.h>
#endif

#include "memd.h"

#ifdef MEMDEBUG

#define FILE_LEN 32

typedef struct _memd_node {
  struct _memd_node *next; /* Next block in list. */

  char show;               /* show this block in ShowMem? */

  char file[FILE_LEN + 1]; /* File that allocated us. */
  long line;               /* Line we were allocated on. */

  size_t size;             /* Size of memory block. */
  void *ptr;               /* Address pointer to the block. */
} MemDNode;


static MemDNode *remove_ptr(void *ptr);
static void *add_node(MemDNode *node, size_t size, const char *file,
                      long line);
static void out_of_memory(const char *file, long line, const char *func,
                          int err);


static MemDNode *__first = NULL;
static unsigned long __free_calls = 0;
static unsigned long __malloc_calls = 0;
static unsigned long __realloc_calls = 0;


void *
_memd_alloc(size_t size, const char *file, long line)
{
  MemDNode *pNode;

  __malloc_calls++;

  if(!size) {
    fprintf(stderr, "%s:%ld: _memd_alloc: zero size.\n", file, line);
    return NULL;
  }

  pNode = (MemDNode *)malloc(sizeof(MemDNode) + size);

  if(!pNode) {
    out_of_memory(file, line, "_memd_alloc", errno);
  }

  return add_node(pNode, size, file, line);
}


void *
_memd_realloc(void *ptr, size_t size, const char *file, long line)
{
  MemDNode *pNode;

  __realloc_calls++;

  if(ptr) {
    pNode = remove_ptr(ptr);

    if(!pNode) {
      fprintf(stderr, "%s:%ld: _memd_realloc: 0x%x not allocated before?\n",
              file, line, (unsigned int)((size_t)ptr));
      return NULL;
    }

    if(!size) {
      fprintf(stderr, "%s:%ld: _memd_realloc: zero size, freeing.\n", file,
              line);
      free(pNode);
      return NULL;
    }

    pNode = (MemDNode *)realloc(pNode, sizeof(MemDNode) + size);
  } else {
    fprintf(stderr, "%s:%ld: _memd_realloc: realloc NULL pointer.\n", file,
            line);

    pNode = (MemDNode *)malloc(sizeof(MemDNode) + size);
  }

  if(!pNode) {
    out_of_memory(file, line, "_memd_realloc", errno);
  }

  return add_node(pNode, size, file, line);
}


void
_memd_free(void *ptr, const char *file, long line)
{
  MemDNode *pNode;

  __free_calls++;

  if(!ptr) {
    fprintf(stderr, "%s:%ld: _memd_free: freeing NULL.\n", file, line);
  } else {
    pNode = remove_ptr(ptr);

    if(pNode) {
      free(pNode);
    } else {
      fprintf(stderr, "%s:%ld: _memd_free: freeing 0x%x twice?\n", file, line,
              (unsigned int)((size_t)ptr));
    }
  }
}


extern void *
_memd_setline(void *ptr, const char *file, long line)
{
  MemDNode *pNode;

  if(ptr) {
    for(pNode = __first; pNode && (pNode->ptr != ptr); pNode = pNode->next);

    if(pNode) {
      strncpy(pNode->file, file, FILE_LEN);
      pNode->file[FILE_LEN] = (char)0;
      pNode->line = line;
    } else {
      fprintf(stderr, "%s:%ld: _memd_setline: pointer 0x%x not found.\n", file,
              line, (unsigned int)((size_t)ptr));
    }
  }

  return ptr;
}


extern void *
_memd_setshow(void *ptr, char value, const char *file, long line)
{
  MemDNode *pNode;

  if(ptr) {
    for(pNode = __first; pNode && (pNode->ptr != ptr); pNode = pNode->next);

    if(pNode) {
      pNode->show = value;
    } else {
      fprintf(stderr, "%s:%ld: _memd_setshow: pointer 0x%x not found.\n", file,
              line, (unsigned int)((size_t)ptr));
    }
  }

  return ptr;
}


void
_memd_showmem(const char *file, long line)
{
  MemDNode *pNode;
  unsigned long size, nr, pheader = 0;

  fprintf(stderr, "%s:%ld: _memd_showmem: memory list.\n", file, line);

  if(__first) {
    for(pNode = __first, nr = 0, size = 0; pNode; pNode = pNode->next) {
      if(pNode->show) {
        if(!pheader) {
          fprintf(stderr, "address       size         file             line\n"
                  "------------------------------------------------\n");
          pheader = 1;
        }

        nr++;
        size += pNode->size;

        fprintf(stderr, "0x%-10x  %-12u %-16s %4ld\n", 
                (unsigned int)((size_t)pNode->ptr),
                (unsigned int)pNode->size, pNode->file, pNode->line);
      }
    }

    if(pheader) {
      fprintf(stderr, "\nallocated: %ld byte%sin %ld block%s.\n", size,
              (size == 1 ? " " : "s "), nr, (nr == 1 ? "" : "s"));
    }
  }

  fprintf(stderr, "malloc calls: %6ld  realloc calls: %6ld  free calls: "
          "%6ld\n", __malloc_calls, __realloc_calls, __free_calls);
}


void
_memd_clearmem(const char *file, long line)
{
  MemDNode *pNode;

  fprintf(stderr, "%s:%ld: _memd_clearmem: empty memory list.\n", file, line);

  for(pNode = __first; pNode; pNode = pNode->next) {
    pNode->show = 0;
  }
}


static MemDNode *
remove_ptr(void *ptr)
{
  MemDNode *pNode = __first, *pOld = NULL;

  while(pNode && (pNode->ptr != ptr)) {
    pOld = pNode;
    pNode = pNode->next;
  }

  if(pNode) {
    if(pOld) {
      pOld->next = pNode->next;
    } else {
      __first = pNode->next;
    }
  }

  return pNode;
}


static void *
add_node(MemDNode *pNode, size_t size, const char *file, long line)
{
  pNode->next = __first;
  __first = pNode;

  strncpy(pNode->file, file, FILE_LEN);
  pNode->file[FILE_LEN] = (char)0;
  pNode->line = line;
  pNode->size = size;
  pNode->show = 1;
  pNode->ptr  = (void *)((unsigned long)pNode + sizeof(MemDNode));

  return pNode->ptr;
}


static void
out_of_memory(const char *file, long line, const char *func, int err)
{
  fprintf(stderr, "%s:%ld: %s: out of memory, %s.\n", file, line, func,
          (err == ENOMEM ? "exceeding the physical limits" :
           (err == EAGAIN ? "not enough memory at the time" : "unknown")));
  exit(1);
}

#endif /* MEMDEBUG */
